
Installation on 64-bit Windows
------------------------------

For installing the *mmgroup* package on a 64-bit Windows system you simply 
type::

   pip install mmgroup

32-bit Windows systems are not supported.

Installation on Linux
---------------------

There is experimantal support for installing the *mmgroup* package on a 
Linux system from a source distribution. Details are given in the 
*guide for developers*.



