rem Activate the MSVC compiler for a Windows 64-bit command line 
rem The path may be different on another PC or after reinstallation!!
rem The current MSVC compiler version has been downloaded at 2020-06-14 from
rem "https://visualstudio.microsoft.com/de/thank-you-downloading-visual-studio/?sku=BuildTools&rel=16"
rem instructions see:
rem "https://www.scivision.dev/python-windows-visual-c-14-required/"
rem But nobody knows where to get it next year or for the next python version!!
"C:\Program Files (x86)\Microsoft Visual Studio\2019\BuildTools\VC\Auxiliary\Build\vcvars64.bat"

