/* Generated by Cython 0.29.21 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__mmgroup__mat24
#define __PYX_HAVE_API__mmgroup__mat24
/* Early includes */
#include <stdint.h>
#include "mat24_functions.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src\\mmgroup\\dev\\pxd_files\\mat24fast.pyx",
};

/*--- Type declarations ---*/

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":306
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *      ploop_solve_size = 64
 * 
 */
enum  {
  __pyx_e_7mmgroup_5mat24_ploop_solve_size = 64
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.stdint' */

/* Module declarations from 'mmgroup.mat24' */
static int __pyx_7genexpr__pyx_v_7mmgroup_5mat24__i;
static int __pyx_8genexpr1__pyx_v_7mmgroup_5mat24__i;
static uint32_t __pyx_f_7mmgroup_5mat24_Mat24Sub_vtype(uint32_t, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "mmgroup.mat24"
extern int __pyx_module_is_main_mmgroup__mat24;
int __pyx_module_is_main_mmgroup__mat24 = 0;

/* Implementation of 'mmgroup.mat24' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_c1[] = "c1";
static const char __pyx_k_cn[] = "cn";
static const char __pyx_k_d1[] = "d1";
static const char __pyx_k_d2[] = "d2";
static const char __pyx_k_h1[] = "h1";
static const char __pyx_k_h2[] = "h2";
static const char __pyx_k_m1[] = "m1";
static const char __pyx_k_m2[] = "m2";
static const char __pyx_k_p1[] = "p1";
static const char __pyx_k_p2[] = "p2";
static const char __pyx_k_pa[] = "pa";
static const char __pyx_k_pi[] = "pi";
static const char __pyx_k_pm[] = "pm";
static const char __pyx_k_v1[] = "v1";
static const char __pyx_k_v2[] = "v2";
static const char __pyx_k_v3[] = "v3";
static const char __pyx_k_acc[] = "acc";
static const char __pyx_k_c1i[] = "c1i";
static const char __pyx_k_d1a[] = "d1a";
static const char __pyx_k_d2a[] = "d2a";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_h1a[] = "h1a";
static const char __pyx_k_h2a[] = "h2a";
static const char __pyx_k_m1a[] = "m1a";
static const char __pyx_k_m2a[] = "m2a";
static const char __pyx_k_m3a[] = "m3a";
static const char __pyx_k_p1a[] = "p1a";
static const char __pyx_k_p_p[] = "p_p";
static const char __pyx_k_pp1[] = "pp1";
static const char __pyx_k_pp2[] = "pp2";
static const char __pyx_k_pp3[] = "pp3";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_bw24[] = "bw24";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_m_io[] = "m_io";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_p_io[] = "p_io";
static const char __pyx_k_p_m1[] = "p_m1";
static const char __pyx_k_p_p1[] = "p_p1";
static const char __pyx_k_p_pi[] = "p_pi";
static const char __pyx_k_p_pm[] = "p_pm";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_basis[] = "basis";
static const char __pyx_k_entry[] = "entry";
static const char __pyx_k_m_ioa[] = "m_ioa";
static const char __pyx_k_p_d1a[] = "p_d1a";
static const char __pyx_k_p_d2a[] = "p_d2a";
static const char __pyx_k_p_h1a[] = "p_h1a";
static const char __pyx_k_p_h2a[] = "p_h2a";
static const char __pyx_k_p_m1a[] = "p_m1a";
static const char __pyx_k_p_m2a[] = "p_m2a";
static const char __pyx_k_p_m3a[] = "p_m3a";
static const char __pyx_k_p_p1a[] = "p_p1a";
static const char __pyx_k_p_pp1[] = "p_pp1";
static const char __pyx_k_p_pp2[] = "p_pp2";
static const char __pyx_k_p_pp3[] = "p_pp3";
static const char __pyx_k_p_res[] = "p_res";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_u_exp[] = "u_exp";
static const char __pyx_k_u_len[] = "u_len";
static const char __pyx_k_u_m24[] = "u_m24";
static const char __pyx_k_u_sub[] = "u_sub";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_u_mask[] = "u_mask";
static const char __pyx_k_u_sub1[] = "u_sub1";
static const char __pyx_k_u_sub2[] = "u_sub2";
static const char __pyx_k_lsbit24[] = "lsbit24";
static const char __pyx_k_p_m_ioa[] = "p_m_ioa";
static const char __pyx_k_u_octad[] = "u_octad";
static const char __pyx_k_inv_perm[] = "inv_perm";
static const char __pyx_k_mul_perm[] = "mul_perm";
static const char __pyx_k_syndrome[] = "syndrome";
static const char __pyx_k_u_single[] = "u_single";
static const char __pyx_k_u_strict[] = "u_strict";
static const char __pyx_k_u_tetrad[] = "u_tetrad";
static const char __pyx_k_inv_autpl[] = "inv_autpl";
static const char __pyx_k_mul_autpl[] = "mul_autpl";
static const char __pyx_k_mul_ploop[] = "mul_ploop";
static const char __pyx_k_ploop_cap[] = "ploop_cap";
static const char __pyx_k_pow_ploop[] = "pow_ploop";
static const char __pyx_k_vect_type[] = "vect_type";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_perm_check[] = "perm_check";
static const char __pyx_k_ploop_comm[] = "ploop_comm";
static const char __pyx_k_spread_b24[] = "spread_b24";
static const char __pyx_k_MAT24_ORDER[] = "MAT24_ORDER";
static const char __pyx_k_extract_b24[] = "extract_b24";
static const char __pyx_k_perm_to_net[] = "perm_to_net";
static const char __pyx_k_ploop_assoc[] = "ploop_assoc";
static const char __pyx_k_ploop_solve[] = "ploop_solve";
static const char __pyx_k_ploop_theta[] = "ploop_theta";
static const char __pyx_k_recip_basis[] = "recip_basis";
static const char __pyx_k_scalar_prod[] = "scalar_prod";
static const char __pyx_k_gcode_weight[] = "gcode_weight";
static const char __pyx_k_op_all_autpl[] = "op_all_autpl";
static const char __pyx_k_op_vect_perm[] = "op_vect_perm";
static const char __pyx_k_vect_to_list[] = "vect_to_list";
static const char __pyx_k_autpl_to_perm[] = "autpl_to_perm";
static const char __pyx_k_cocode_weight[] = "cocode_weight";
static const char __pyx_k_gcode_to_vect[] = "gcode_to_vect";
static const char __pyx_k_mmgroup_mat24[] = "mmgroup.mat24";
static const char __pyx_k_octad_to_vect[] = "octad_to_vect";
static const char __pyx_k_op_all_cocode[] = "op_all_cocode";
static const char __pyx_k_op_gcode_perm[] = "op_gcode_perm";
static const char __pyx_k_perm_from_map[] = "perm_from_map";
static const char __pyx_k_perm_to_autpl[] = "perm_to_autpl";
static const char __pyx_k_ploop_cocycle[] = "ploop_cocycle";
static const char __pyx_k_result_length[] = "result_length";
static const char __pyx_k_vect_to_gcode[] = "vect_to_gcode";
static const char __pyx_k_vect_to_octad[] = "vect_to_octad";
static const char __pyx_k_cocode_to_vect[] = "cocode_to_vect";
static const char __pyx_k_gcode_to_octad[] = "gcode_to_octad";
static const char __pyx_k_m24num_to_perm[] = "m24num_to_perm";
static const char __pyx_k_matrix_to_perm[] = "matrix_to_perm";
static const char __pyx_k_octad_to_gcode[] = "octad_to_gcode";
static const char __pyx_k_op_cocode_perm[] = "op_cocode_perm";
static const char __pyx_k_op_ploop_autpl[] = "op_ploop_autpl";
static const char __pyx_k_perm_to_iautpl[] = "perm_to_iautpl";
static const char __pyx_k_perm_to_m24num[] = "perm_to_m24num";
static const char __pyx_k_perm_to_matrix[] = "perm_to_matrix";
static const char __pyx_k_syndrome_table[] = "syndrome_table";
static const char __pyx_k_vect_to_cocode[] = "vect_to_cocode";
static const char __pyx_k_autpl_set_qform[] = "autpl_set_qform";
static const char __pyx_k_autpl_to_cocode[] = "autpl_to_cocode";
static const char __pyx_k_cocode_syndrome[] = "cocode_syndrome";
static const char __pyx_k_cocode_to_autpl[] = "cocode_to_autpl";
static const char __pyx_k_op_gcode_matrix[] = "op_gcode_matrix";
static const char __pyx_k_suboctad_weight[] = "suboctad_weight";
static const char __pyx_k_vect_to_vintern[] = "vect_to_vintern";
static const char __pyx_k_vintern_to_vect[] = "vintern_to_vect";
static const char __pyx_k_cocode_to_sextet[] = "cocode_to_sextet";
static const char __pyx_k_vect_to_bit_list[] = "vect_to_bit_list";
static const char __pyx_k_gcode_to_bit_list[] = "gcode_to_bit_list";
static const char __pyx_k_perm_from_heptads[] = "perm_from_heptads";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_cocode_to_bit_list[] = "cocode_to_bit_list";
static const char __pyx_k_cocode_to_suboctad[] = "cocode_to_suboctad";
static const char __pyx_k_perm_from_dodecads[] = "perm_from_dodecads";
static const char __pyx_k_suboctad_to_cocode[] = "suboctad_to_cocode";
static const char __pyx_k_perm_complete_octad[] = "perm_complete_octad";
static const char __pyx_k_Illegal_octad_number[] = "Illegal octad number";
static const char __pyx_k_cocode_as_subdodecad[] = "cocode_as_subdodecad";
static const char __pyx_k_perm_complete_heptad[] = "perm_complete_heptad";
static const char __pyx_k_suboctad_scalar_prod[] = "suboctad_scalar_prod";
static const char __pyx_k_Mat24Sub_count_vtypes[] = "Mat24Sub_count_vtypes";
static const char __pyx_k_matrix_from_mod_omega[] = "matrix_from_mod_omega";
static const char __pyx_k_Octad_suboctad_mismatch[] = "Octad/suboctad mismatch";
static const char __pyx_k_Bit_vector_is_not_an_octad[] = "Bit vector is not an octad";
static const char __pyx_k_Vector_is_not_an_umbral_heptad[] = "Vector is not an umbral heptad";
static const char __pyx_k_Array_in_function_perm_from_map[] = "Array in function perm_from_map() is too long";
static const char __pyx_k_Cannot_construct_permutation_in[] = "Cannot construct permutation in Mat24 from heptads";
static const char __pyx_k_Golay_code_word_is_not_an_octad[] = "Golay code word is not an octad";
static const char __pyx_k_Permutation_in_group_Mat24_must[] = "Permutation in group Mat24 must have length 24";
static const char __pyx_k_Arrays_in_function_perm_from_map[] = "Arrays in function perm_from_map() have different lengths";
static const char __pyx_k_Attempt_to_compute_suboctad_of_a[] = "Attempt to compute suboctad of a non-octad";
static const char __pyx_k_Bit_vector_is_not_a_Golay_code_w[] = "Bit vector is not a Golay code word";
static const char __pyx_k_Cannot_complete_a_vector_to_an_o[] = "Cannot complete a vector to an octad";
static const char __pyx_k_Cannot_correct_signs_of_all_Park[] = "Cannot correct signs of all Parker loop elements";
static const char __pyx_k_Cocode_element_is_not_a_subset_o[] = "Cocode element is not a subset of the docecad";
static const char __pyx_k_Golay_cocode_word_is_not_a_sexte[] = "Golay cocode word is not a sextet";
static const char __pyx_k_Golay_code_syndrome_is_not_uniqu[] = "Golay code syndrome is not unique";
static const char __pyx_k_Illegal_entry_in_array_in_functi[] = "Illegal entry in array in function perm_from_map()";
static const char __pyx_k_Illegal_number_for_a_permutation[] = "Illegal number for a permutation in Mat24";
static const char __pyx_k_Permutation_is_not_in_group_Mat2[] = "Permutation is not in group Mat24";
static const char __pyx_k_src_mmgroup_dev_pxd_files_mat24f[] = "src\\mmgroup\\dev\\pxd_files\\mat24fast.pyx";
static const char __pyx_k_Cannot_construct_permutation_in_2[] = "Cannot construct permutation in Mat24 from dodecads";
static PyObject *__pyx_kp_u_Array_in_function_perm_from_map;
static PyObject *__pyx_kp_u_Arrays_in_function_perm_from_map;
static PyObject *__pyx_kp_u_Attempt_to_compute_suboctad_of_a;
static PyObject *__pyx_kp_u_Bit_vector_is_not_a_Golay_code_w;
static PyObject *__pyx_kp_u_Bit_vector_is_not_an_octad;
static PyObject *__pyx_kp_u_Cannot_complete_a_vector_to_an_o;
static PyObject *__pyx_kp_u_Cannot_construct_permutation_in;
static PyObject *__pyx_kp_u_Cannot_construct_permutation_in_2;
static PyObject *__pyx_kp_u_Cannot_correct_signs_of_all_Park;
static PyObject *__pyx_kp_u_Cocode_element_is_not_a_subset_o;
static PyObject *__pyx_kp_u_Golay_cocode_word_is_not_a_sexte;
static PyObject *__pyx_kp_u_Golay_code_syndrome_is_not_uniqu;
static PyObject *__pyx_kp_u_Golay_code_word_is_not_an_octad;
static PyObject *__pyx_kp_u_Illegal_entry_in_array_in_functi;
static PyObject *__pyx_kp_u_Illegal_number_for_a_permutation;
static PyObject *__pyx_kp_u_Illegal_octad_number;
static PyObject *__pyx_n_s_MAT24_ORDER;
static PyObject *__pyx_n_s_Mat24Sub_count_vtypes;
static PyObject *__pyx_kp_u_Octad_suboctad_mismatch;
static PyObject *__pyx_kp_u_Permutation_in_group_Mat24_must;
static PyObject *__pyx_kp_u_Permutation_is_not_in_group_Mat2;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u_Vector_is_not_an_umbral_heptad;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_acc;
static PyObject *__pyx_n_s_autpl_set_qform;
static PyObject *__pyx_n_s_autpl_to_cocode;
static PyObject *__pyx_n_s_autpl_to_perm;
static PyObject *__pyx_n_s_basis;
static PyObject *__pyx_n_s_bw24;
static PyObject *__pyx_n_s_c1;
static PyObject *__pyx_n_s_c1i;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cn;
static PyObject *__pyx_n_s_cocode_as_subdodecad;
static PyObject *__pyx_n_s_cocode_syndrome;
static PyObject *__pyx_n_s_cocode_to_autpl;
static PyObject *__pyx_n_s_cocode_to_bit_list;
static PyObject *__pyx_n_s_cocode_to_sextet;
static PyObject *__pyx_n_s_cocode_to_suboctad;
static PyObject *__pyx_n_s_cocode_to_vect;
static PyObject *__pyx_n_s_cocode_weight;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_d1;
static PyObject *__pyx_n_s_d1a;
static PyObject *__pyx_n_s_d2;
static PyObject *__pyx_n_s_d2a;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_entry;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_extract_b24;
static PyObject *__pyx_n_s_gcode_to_bit_list;
static PyObject *__pyx_n_s_gcode_to_octad;
static PyObject *__pyx_n_s_gcode_to_vect;
static PyObject *__pyx_n_s_gcode_weight;
static PyObject *__pyx_n_s_h1;
static PyObject *__pyx_n_s_h1a;
static PyObject *__pyx_n_s_h2;
static PyObject *__pyx_n_s_h2a;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_inv_autpl;
static PyObject *__pyx_n_s_inv_perm;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_lsbit24;
static PyObject *__pyx_n_s_m1;
static PyObject *__pyx_n_s_m1a;
static PyObject *__pyx_n_s_m2;
static PyObject *__pyx_n_s_m24num_to_perm;
static PyObject *__pyx_n_s_m2a;
static PyObject *__pyx_n_s_m3a;
static PyObject *__pyx_n_s_m_io;
static PyObject *__pyx_n_s_m_ioa;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_matrix_from_mod_omega;
static PyObject *__pyx_n_s_matrix_to_perm;
static PyObject *__pyx_n_s_mmgroup_mat24;
static PyObject *__pyx_n_s_mul_autpl;
static PyObject *__pyx_n_s_mul_perm;
static PyObject *__pyx_n_s_mul_ploop;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_octad_to_gcode;
static PyObject *__pyx_n_s_octad_to_vect;
static PyObject *__pyx_n_s_op_all_autpl;
static PyObject *__pyx_n_s_op_all_cocode;
static PyObject *__pyx_n_s_op_cocode_perm;
static PyObject *__pyx_n_s_op_gcode_matrix;
static PyObject *__pyx_n_s_op_gcode_perm;
static PyObject *__pyx_n_s_op_ploop_autpl;
static PyObject *__pyx_n_s_op_vect_perm;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_p1;
static PyObject *__pyx_n_s_p1a;
static PyObject *__pyx_n_s_p2;
static PyObject *__pyx_n_s_p_d1a;
static PyObject *__pyx_n_s_p_d2a;
static PyObject *__pyx_n_s_p_h1a;
static PyObject *__pyx_n_s_p_h2a;
static PyObject *__pyx_n_s_p_io;
static PyObject *__pyx_n_s_p_m1;
static PyObject *__pyx_n_s_p_m1a;
static PyObject *__pyx_n_s_p_m2a;
static PyObject *__pyx_n_s_p_m3a;
static PyObject *__pyx_n_s_p_m_ioa;
static PyObject *__pyx_n_s_p_p;
static PyObject *__pyx_n_s_p_p1;
static PyObject *__pyx_n_s_p_p1a;
static PyObject *__pyx_n_s_p_pi;
static PyObject *__pyx_n_s_p_pm;
static PyObject *__pyx_n_s_p_pp1;
static PyObject *__pyx_n_s_p_pp2;
static PyObject *__pyx_n_s_p_pp3;
static PyObject *__pyx_n_s_p_res;
static PyObject *__pyx_n_s_pa;
static PyObject *__pyx_n_s_perm_check;
static PyObject *__pyx_n_s_perm_complete_heptad;
static PyObject *__pyx_n_s_perm_complete_octad;
static PyObject *__pyx_n_s_perm_from_dodecads;
static PyObject *__pyx_n_s_perm_from_heptads;
static PyObject *__pyx_n_s_perm_from_map;
static PyObject *__pyx_n_s_perm_to_autpl;
static PyObject *__pyx_n_s_perm_to_iautpl;
static PyObject *__pyx_n_s_perm_to_m24num;
static PyObject *__pyx_n_s_perm_to_matrix;
static PyObject *__pyx_n_s_perm_to_net;
static PyObject *__pyx_n_s_pi;
static PyObject *__pyx_n_s_ploop_assoc;
static PyObject *__pyx_n_s_ploop_cap;
static PyObject *__pyx_n_s_ploop_cocycle;
static PyObject *__pyx_n_s_ploop_comm;
static PyObject *__pyx_n_s_ploop_solve;
static PyObject *__pyx_n_s_ploop_theta;
static PyObject *__pyx_n_s_pm;
static PyObject *__pyx_n_s_pow_ploop;
static PyObject *__pyx_n_s_pp1;
static PyObject *__pyx_n_s_pp2;
static PyObject *__pyx_n_s_pp3;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_recip_basis;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_result_length;
static PyObject *__pyx_n_s_scalar_prod;
static PyObject *__pyx_n_s_spread_b24;
static PyObject *__pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_suboctad_scalar_prod;
static PyObject *__pyx_n_s_suboctad_to_cocode;
static PyObject *__pyx_n_s_suboctad_weight;
static PyObject *__pyx_n_s_syndrome;
static PyObject *__pyx_n_s_syndrome_table;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_u_exp;
static PyObject *__pyx_n_s_u_len;
static PyObject *__pyx_n_s_u_m24;
static PyObject *__pyx_n_s_u_mask;
static PyObject *__pyx_n_s_u_octad;
static PyObject *__pyx_n_s_u_single;
static PyObject *__pyx_n_s_u_strict;
static PyObject *__pyx_n_s_u_sub;
static PyObject *__pyx_n_s_u_sub1;
static PyObject *__pyx_n_s_u_sub2;
static PyObject *__pyx_n_s_u_tetrad;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_v1;
static PyObject *__pyx_n_s_v2;
static PyObject *__pyx_n_s_v3;
static PyObject *__pyx_n_s_vect_to_bit_list;
static PyObject *__pyx_n_s_vect_to_cocode;
static PyObject *__pyx_n_s_vect_to_gcode;
static PyObject *__pyx_n_s_vect_to_list;
static PyObject *__pyx_n_s_vect_to_octad;
static PyObject *__pyx_n_s_vect_to_vintern;
static PyObject *__pyx_n_s_vect_type;
static PyObject *__pyx_n_s_vintern_to_vect;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_pf_7mmgroup_5mat24_lsbit24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_2bw24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_4vect_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_6vect_to_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_len); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_8extract_b24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_mask); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_10spread_b24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_mask); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_12vect_to_vintern(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_14vintern_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_16vect_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_18gcode_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_20cocode_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_22vect_to_gcode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_24gcode_to_octad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_strict); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_26vect_to_octad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_strict); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_28octad_to_gcode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_octad); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_30octad_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_octad); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_32syndrome_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_34cocode_syndrome(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_u_tetrad); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_36syndrome(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_tetrad); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_38vect_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_40gcode_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_42gcode_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_44cocode_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_46cocode_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_u_tetrad); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_48cocode_to_sextet(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_50scalar_prod(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_52suboctad_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_54cocode_to_suboctad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_56suboctad_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_58suboctad_scalar_prod(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub1, uint32_t __pyx_v_u_sub2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_60cocode_as_subdodecad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_single); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_62ploop_theta(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_64ploop_cocycle(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_66mul_ploop(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_68pow_ploop(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_exp); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_70ploop_comm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_72ploop_cap(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_74ploop_assoc(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2, uint32_t __pyx_v_v3); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_76ploop_solve(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_78perm_complete_heptad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p_io); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_80perm_check(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_82perm_complete_octad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p_io); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_84perm_from_heptads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_h1, PyObject *__pyx_v_h2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_86perm_from_map(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_h1, PyObject *__pyx_v_h2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_88m24num_to_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_m24); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_90perm_to_m24num(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_92perm_to_matrix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_94matrix_to_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_96matrix_from_mod_omega(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_98perm_from_dodecads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_d1, PyObject *__pyx_v_d2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_100op_vect_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_102op_gcode_matrix(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_104op_gcode_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_106op_cocode_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_108mul_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_110inv_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_112autpl_set_qform(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m_io); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_114perm_to_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_116cocode_to_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_118autpl_to_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_120autpl_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_122op_ploop_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v1, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_124mul_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1, PyObject *__pyx_v_m2); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_126inv_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_128perm_to_iautpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_130perm_to_net(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_132op_all_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_134op_all_cocode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_136Mat24Sub_vtype(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_7mmgroup_5mat24_138Mat24Sub_count_vtypes(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_int_2047;
static PyObject *__pyx_int_244823040;
static PyObject *__pyx_int_4278190080;
static PyObject *__pyx_int_4294963200;
static PyObject *__pyx_int_4294967295;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__73;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__79;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__83;
static PyObject *__pyx_codeobj__85;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__89;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__93;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__101;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__105;
static PyObject *__pyx_codeobj__107;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__111;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__117;
static PyObject *__pyx_codeobj__119;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__123;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__129;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__135;
static PyObject *__pyx_codeobj__137;
static PyObject *__pyx_codeobj__139;
static PyObject *__pyx_codeobj__141;
static PyObject *__pyx_codeobj__143;
/* Late includes */

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":30
 * 
 * 
 * def lsbit24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_lsbit24(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_1lsbit24(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_1lsbit24 = {"lsbit24", (PyCFunction)__pyx_pw_7mmgroup_5mat24_1lsbit24, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_1lsbit24(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lsbit24 (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.lsbit24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_lsbit24(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_lsbit24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lsbit24", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":31
 * 
 * def lsbit24(uint32_t v1):
 *     return mat24_lsbit24(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_lsbit24(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":30
 * 
 * 
 * def lsbit24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_lsbit24(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.lsbit24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":34
 * 
 * 
 * def bw24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_bw24(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_3bw24(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_3bw24 = {"bw24", (PyCFunction)__pyx_pw_7mmgroup_5mat24_3bw24, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_3bw24(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bw24 (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.bw24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_2bw24(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_2bw24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bw24", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":35
 * 
 * def bw24(uint32_t v1):
 *     return mat24_bw24(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_bw24(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":34
 * 
 * 
 * def bw24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_bw24(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.bw24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":53
 * 
 * 
 * def vect_to_bit_list(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_5vect_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_5vect_to_bit_list = {"vect_to_bit_list", (PyCFunction)__pyx_pw_7mmgroup_5mat24_5vect_to_bit_list, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_5vect_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_bit_list (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_4vect_to_bit_list(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_4vect_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  uint8_t __pyx_v_a[24];
  uint32_t __pyx_v_w;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_bit_list", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":56
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 *     w = mat24_vect_to_bit_list(v1, &a[0])             # <<<<<<<<<<<<<<
 *     return w, list(a[:24])
 * 
 */
  __pyx_v_w = mat24_vect_to_bit_list(__pyx_v_v1, (&(__pyx_v_a[0])));

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":57
 *     cdef uint32_t w
 *     w = mat24_vect_to_bit_list(v1, &a[0])
 *     return w, list(a[:24])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_w); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_a) + 0, 24 - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":53
 * 
 * 
 * def vect_to_bit_list(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":61
 * 
 * 
 * def vect_to_list(uint32_t v1, uint32_t u_len):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_7vect_to_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_7vect_to_list = {"vect_to_list", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_7vect_to_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_7vect_to_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_len;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_len,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_len)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("vect_to_list", 1, 2, 2, 1); __PYX_ERR(0, 61, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "vect_to_list") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    __pyx_v_u_len = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_len == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("vect_to_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_6vect_to_list(__pyx_self, __pyx_v_v1, __pyx_v_u_len);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_6vect_to_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_len) {
  uint8_t __pyx_v_a[24];
  uint32_t __pyx_v_w;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_list", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":64
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 *     w = mat24_vect_to_list(v1, u_len, &a[0])             # <<<<<<<<<<<<<<
 *     return list(a[:w])
 * 
 */
  __pyx_v_w = mat24_vect_to_list(__pyx_v_v1, __pyx_v_u_len, (&(__pyx_v_a[0])));

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":65
 *     cdef uint32_t w
 *     w = mat24_vect_to_list(v1, u_len, &a[0])
 *     return list(a[:w])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_a) + 0, __pyx_v_w - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":61
 * 
 * 
 * def vect_to_list(uint32_t v1, uint32_t u_len):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":69
 * 
 * 
 * def extract_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_extract_b24(v1, u_mask)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_9extract_b24(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_9extract_b24 = {"extract_b24", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_9extract_b24, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_9extract_b24(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_mask;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extract_b24 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_mask,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("extract_b24", 1, 2, 2, 1); __PYX_ERR(0, 69, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "extract_b24") < 0)) __PYX_ERR(0, 69, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L3_error)
    __pyx_v_u_mask = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_mask == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("extract_b24", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 69, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.extract_b24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_8extract_b24(__pyx_self, __pyx_v_v1, __pyx_v_u_mask);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_8extract_b24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_mask) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extract_b24", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":70
 * 
 * def extract_b24(uint32_t v1, uint32_t u_mask):
 *     return mat24_extract_b24(v1, u_mask)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_extract_b24(__pyx_v_v1, __pyx_v_u_mask)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":69
 * 
 * 
 * def extract_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_extract_b24(v1, u_mask)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.extract_b24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":73
 * 
 * 
 * def spread_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_spread_b24(v1, u_mask)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_11spread_b24(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_11spread_b24 = {"spread_b24", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_11spread_b24, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_11spread_b24(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_mask;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spread_b24 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_mask,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spread_b24", 1, 2, 2, 1); __PYX_ERR(0, 73, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spread_b24") < 0)) __PYX_ERR(0, 73, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
    __pyx_v_u_mask = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_mask == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spread_b24", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 73, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.spread_b24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_10spread_b24(__pyx_self, __pyx_v_v1, __pyx_v_u_mask);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_10spread_b24(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_mask) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("spread_b24", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":74
 * 
 * def spread_b24(uint32_t v1, uint32_t u_mask):
 *     return mat24_spread_b24(v1, u_mask)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_spread_b24(__pyx_v_v1, __pyx_v_u_mask)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":73
 * 
 * 
 * def spread_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_spread_b24(v1, u_mask)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.spread_b24", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":84
 * 
 * 
 * def vect_to_vintern(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_vintern(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_13vect_to_vintern(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_13vect_to_vintern = {"vect_to_vintern", (PyCFunction)__pyx_pw_7mmgroup_5mat24_13vect_to_vintern, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_13vect_to_vintern(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_vintern (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 84, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_vintern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_12vect_to_vintern(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_12vect_to_vintern(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_vintern", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":85
 * 
 * def vect_to_vintern(uint32_t v1):
 *     return mat24_vect_to_vintern(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_vect_to_vintern(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":84
 * 
 * 
 * def vect_to_vintern(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_vintern(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_vintern", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":89
 * 
 * 
 * def vintern_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return  mat24_vintern_to_vect(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_15vintern_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_15vintern_to_vect = {"vintern_to_vect", (PyCFunction)__pyx_pw_7mmgroup_5mat24_15vintern_to_vect, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_15vintern_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vintern_to_vect (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vintern_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_14vintern_to_vect(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_14vintern_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vintern_to_vect", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":90
 * 
 * def vintern_to_vect(uint32_t v1):
 *     return  mat24_vintern_to_vect(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_vintern_to_vect(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":89
 * 
 * 
 * def vintern_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return  mat24_vintern_to_vect(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.vintern_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":93
 * 
 * 
 * def vect_to_cocode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_cocode(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_17vect_to_cocode(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_17vect_to_cocode = {"vect_to_cocode", (PyCFunction)__pyx_pw_7mmgroup_5mat24_17vect_to_cocode, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_17vect_to_cocode(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_cocode (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 93, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_16vect_to_cocode(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_16vect_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_cocode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":94
 * 
 * def vect_to_cocode(uint32_t v1):
 *     return mat24_vect_to_cocode(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_vect_to_cocode(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":93
 * 
 * 
 * def vect_to_cocode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_cocode(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":97
 * 
 * 
 * def gcode_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_to_vect(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_19gcode_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_19gcode_to_vect = {"gcode_to_vect", (PyCFunction)__pyx_pw_7mmgroup_5mat24_19gcode_to_vect, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_19gcode_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gcode_to_vect (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_18gcode_to_vect(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_18gcode_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gcode_to_vect", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":98
 * 
 * def gcode_to_vect(uint32_t v1):
 *     return mat24_gcode_to_vect(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_gcode_to_vect(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":97
 * 
 * 
 * def gcode_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_to_vect(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":101
 * 
 * 
 * def cocode_to_vect(uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_to_vect(c1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_21cocode_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_21cocode_to_vect = {"cocode_to_vect", (PyCFunction)__pyx_pw_7mmgroup_5mat24_21cocode_to_vect, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_21cocode_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_c1) {
  uint32_t __pyx_v_c1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_to_vect (wrapper)", 0);
  assert(__pyx_arg_c1); {
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_c1); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_20cocode_to_vect(__pyx_self, ((uint32_t)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_20cocode_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_to_vect", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":102
 * 
 * def cocode_to_vect(uint32_t c1):
 *     return mat24_cocode_to_vect(c1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_cocode_to_vect(__pyx_v_c1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":101
 * 
 * 
 * def cocode_to_vect(uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_to_vect(c1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":105
 * 
 * 
 * def vect_to_gcode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_23vect_to_gcode(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_23vect_to_gcode = {"vect_to_gcode", (PyCFunction)__pyx_pw_7mmgroup_5mat24_23vect_to_gcode, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_23vect_to_gcode(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_gcode (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 105, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_gcode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_22vect_to_gcode(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_22vect_to_gcode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  uint32_t __pyx_v_cn;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_gcode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":107
 * def vect_to_gcode(uint32_t v1):
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)             # <<<<<<<<<<<<<<
 *     if cn & 0xff000000:
 *         err = "Bit vector is not a Golay code word"
 */
  __pyx_v_cn = mat24_vect_to_gcode(__pyx_v_v1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":108
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Bit vector is not a Golay code word"
 *         raise ValueError, err
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":109
 *     cn = mat24_vect_to_gcode(v1)
 *     if cn & 0xff000000:
 *         err = "Bit vector is not a Golay code word"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return cn
 */
    __Pyx_INCREF(__pyx_kp_u_Bit_vector_is_not_a_Golay_code_w);
    __pyx_v_err = __pyx_kp_u_Bit_vector_is_not_a_Golay_code_w;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":110
 *     if cn & 0xff000000:
 *         err = "Bit vector is not a Golay code word"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return cn
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 110, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":108
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Bit vector is not a Golay code word"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":111
 *         err = "Bit vector is not a Golay code word"
 *         raise ValueError, err
 *     return cn             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":105
 * 
 * 
 * def vect_to_gcode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_gcode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":114
 * 
 * 
 * def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_25gcode_to_octad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_25gcode_to_octad = {"gcode_to_octad", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_25gcode_to_octad, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_25gcode_to_octad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_strict;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gcode_to_octad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_strict,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_strict);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "gcode_to_octad") < 0)) __PYX_ERR(0, 114, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_u_strict = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_strict == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L3_error)
    } else {
      __pyx_v_u_strict = ((uint32_t)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("gcode_to_octad", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 114, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_octad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_24gcode_to_octad(__pyx_self, __pyx_v_v1, __pyx_v_u_strict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_24gcode_to_octad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_strict) {
  uint32_t __pyx_v_cn;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gcode_to_octad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":116
 * def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)             # <<<<<<<<<<<<<<
 *     if cn & 0xff000000:
 *         err = "Golay code word is not an octad"
 */
  __pyx_v_cn = mat24_gcode_to_octad(__pyx_v_v1, __pyx_v_u_strict);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":117
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Golay code word is not an octad"
 *         raise ValueError, err
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":118
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 *     if cn & 0xff000000:
 *         err = "Golay code word is not an octad"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return cn
 */
    __Pyx_INCREF(__pyx_kp_u_Golay_code_word_is_not_an_octad);
    __pyx_v_err = __pyx_kp_u_Golay_code_word_is_not_an_octad;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":119
 *     if cn & 0xff000000:
 *         err = "Golay code word is not an octad"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return cn
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 119, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":117
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Golay code word is not an octad"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":120
 *         err = "Golay code word is not an octad"
 *         raise ValueError, err
 *     return cn             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":114
 * 
 * 
 * def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_octad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":123
 * 
 * 
 * def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_27vect_to_octad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_27vect_to_octad = {"vect_to_octad", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_27vect_to_octad, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_27vect_to_octad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_strict;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_to_octad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_strict,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_strict);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "vect_to_octad") < 0)) __PYX_ERR(0, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_u_strict = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_strict == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L3_error)
    } else {
      __pyx_v_u_strict = ((uint32_t)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("vect_to_octad", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_octad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_26vect_to_octad(__pyx_self, __pyx_v_v1, __pyx_v_u_strict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_26vect_to_octad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_strict) {
  uint32_t __pyx_v_v;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_to_octad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":125
 * def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)             # <<<<<<<<<<<<<<
 *     if v & 0xff000000:
 *         err = "Bit vector is not an octad"
 */
  __pyx_v_v = mat24_vect_to_octad(__pyx_v_v1, __pyx_v_u_strict);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":126
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 *     if v & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Bit vector is not an octad"
 *         raise ValueError, err
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_v); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":127
 *     v = mat24_vect_to_octad(v1, u_strict)
 *     if v & 0xff000000:
 *         err = "Bit vector is not an octad"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return v
 */
    __Pyx_INCREF(__pyx_kp_u_Bit_vector_is_not_an_octad);
    __pyx_v_err = __pyx_kp_u_Bit_vector_is_not_an_octad;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":128
 *     if v & 0xff000000:
 *         err = "Bit vector is not an octad"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 128, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":126
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 *     if v & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Bit vector is not an octad"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":129
 *         err = "Bit vector is not an octad"
 *         raise ValueError, err
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_v); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":123
 * 
 * 
 * def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.vect_to_octad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":134
 * 
 * 
 * def octad_to_gcode(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_29octad_to_gcode(PyObject *__pyx_self, PyObject *__pyx_arg_u_octad); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_29octad_to_gcode = {"octad_to_gcode", (PyCFunction)__pyx_pw_7mmgroup_5mat24_29octad_to_gcode, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_29octad_to_gcode(PyObject *__pyx_self, PyObject *__pyx_arg_u_octad) {
  uint32_t __pyx_v_u_octad;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("octad_to_gcode (wrapper)", 0);
  assert(__pyx_arg_u_octad); {
    __pyx_v_u_octad = __Pyx_PyInt_As_uint32_t(__pyx_arg_u_octad); if (unlikely((__pyx_v_u_octad == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.octad_to_gcode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_28octad_to_gcode(__pyx_self, ((uint32_t)__pyx_v_u_octad));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_28octad_to_gcode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_octad) {
  uint32_t __pyx_v_cn;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("octad_to_gcode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":136
 * def octad_to_gcode(uint32_t u_octad):
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)             # <<<<<<<<<<<<<<
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"
 */
  __pyx_v_cn = mat24_octad_to_gcode(__pyx_v_u_octad);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":137
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Illegal octad number"
 *         raise ValueError, err
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":138
 *     cn = mat24_octad_to_gcode(u_octad)
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return cn
 */
    __Pyx_INCREF(__pyx_kp_u_Illegal_octad_number);
    __pyx_v_err = __pyx_kp_u_Illegal_octad_number;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":139
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return cn
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 139, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":137
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Illegal octad number"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":140
 *         err = "Illegal octad number"
 *         raise ValueError, err
 *     return cn             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":134
 * 
 * 
 * def octad_to_gcode(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.octad_to_gcode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":143
 * 
 * 
 * def octad_to_vect(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_31octad_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_u_octad); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_31octad_to_vect = {"octad_to_vect", (PyCFunction)__pyx_pw_7mmgroup_5mat24_31octad_to_vect, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_31octad_to_vect(PyObject *__pyx_self, PyObject *__pyx_arg_u_octad) {
  uint32_t __pyx_v_u_octad;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("octad_to_vect (wrapper)", 0);
  assert(__pyx_arg_u_octad); {
    __pyx_v_u_octad = __Pyx_PyInt_As_uint32_t(__pyx_arg_u_octad); if (unlikely((__pyx_v_u_octad == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 143, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.octad_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_30octad_to_vect(__pyx_self, ((uint32_t)__pyx_v_u_octad));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_30octad_to_vect(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_octad) {
  uint32_t __pyx_v_cn;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("octad_to_vect", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":145
 * def octad_to_vect(uint32_t u_octad):
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)             # <<<<<<<<<<<<<<
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"
 */
  __pyx_v_cn = mat24_octad_to_vect(__pyx_v_u_octad);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":146
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Illegal octad number"
 *         raise ValueError, err
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":147
 *     cn = mat24_octad_to_vect(u_octad)
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return cn
 */
    __Pyx_INCREF(__pyx_kp_u_Illegal_octad_number);
    __pyx_v_err = __pyx_kp_u_Illegal_octad_number;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":148
 *     if cn & 0xff000000:
 *         err = "Illegal octad number"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return cn
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 148, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":146
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 *     if cn & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Illegal octad number"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":149
 *         err = "Illegal octad number"
 *         raise ValueError, err
 *     return cn             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_cn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":143
 * 
 * 
 * def octad_to_vect(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.octad_to_vect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":159
 * 
 * 
 * def syndrome_table(c1):             # <<<<<<<<<<<<<<
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_33syndrome_table(PyObject *__pyx_self, PyObject *__pyx_v_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_33syndrome_table = {"syndrome_table", (PyCFunction)__pyx_pw_7mmgroup_5mat24_33syndrome_table, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_33syndrome_table(PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("syndrome_table (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_32syndrome_table(__pyx_self, ((PyObject *)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_32syndrome_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("syndrome_table", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":160
 * 
 * def syndrome_table(c1):
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]             # <<<<<<<<<<<<<<
 * 
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_v_c1, __pyx_int_2047, 0x7ff, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t((MAT24_SYNDROME_TABLE[__pyx_t_2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":159
 * 
 * 
 * def syndrome_table(c1):             # <<<<<<<<<<<<<<
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.syndrome_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":162
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_35cocode_syndrome(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_35cocode_syndrome = {"cocode_syndrome", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_35cocode_syndrome, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_35cocode_syndrome(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_c1;
  uint32_t __pyx_v_u_tetrad;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_syndrome (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_u_tetrad,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_tetrad);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cocode_syndrome") < 0)) __PYX_ERR(0, 162, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_u_tetrad = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_tetrad == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L3_error)
    } else {
      __pyx_v_u_tetrad = ((uint32_t)24);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cocode_syndrome", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 162, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_syndrome", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_34cocode_syndrome(__pyx_self, __pyx_v_c1, __pyx_v_u_tetrad);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_34cocode_syndrome(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_u_tetrad) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_syndrome", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":164
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)             # <<<<<<<<<<<<<<
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 */
  __pyx_v_res = mat24_cocode_syndrome(__pyx_v_c1, __pyx_v_u_tetrad);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":165
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 *     if res == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":166
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 166, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":165
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 *     if res == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":167
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":162
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_syndrome", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":170
 * 
 * 
 * def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_37syndrome(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_37syndrome = {"syndrome", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_37syndrome, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_37syndrome(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_tetrad;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("syndrome (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_tetrad,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_tetrad);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "syndrome") < 0)) __PYX_ERR(0, 170, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_u_tetrad = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_tetrad == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L3_error)
    } else {
      __pyx_v_u_tetrad = ((uint32_t)24);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("syndrome", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 170, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.syndrome", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_36syndrome(__pyx_self, __pyx_v_v1, __pyx_v_u_tetrad);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_36syndrome(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_tetrad) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("syndrome", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":172
 * def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)             # <<<<<<<<<<<<<<
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 */
  __pyx_v_res = mat24_syndrome(__pyx_v_v1, __pyx_v_u_tetrad);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":173
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 *     if res == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":174
 *     res =  mat24_syndrome(v1, u_tetrad)
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 174, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":173
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 *     if res == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":175
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":170
 * 
 * 
 * def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.syndrome", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":179
 * 
 * 
 * def vect_type(v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_39vect_type(PyObject *__pyx_self, PyObject *__pyx_v_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_39vect_type = {"vect_type", (PyCFunction)__pyx_pw_7mmgroup_5mat24_39vect_type, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_39vect_type(PyObject *__pyx_self, PyObject *__pyx_v_v1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vect_type (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_38vect_type(__pyx_self, ((PyObject *)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_38vect_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v1) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("vect_type", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":181
 * def vect_type(v1):
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)             # <<<<<<<<<<<<<<
 *     return ( res >> 5, (res >> 2) & 7, res  & 3 )
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_v1); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_v_res = __pyx_f_7mmgroup_5mat24_Mat24Sub_vtype(__pyx_t_1, 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":182
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)
 *     return ( res >> 5, (res >> 2) & 7, res  & 3 )             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_res >> 5)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_long(((__pyx_v_res >> 2) & 7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_res & 3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":179
 * 
 * 
 * def vect_type(v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mmgroup.mat24.vect_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":187
 * 
 * 
 * def gcode_weight(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_weight(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_41gcode_weight(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_41gcode_weight = {"gcode_weight", (PyCFunction)__pyx_pw_7mmgroup_5mat24_41gcode_weight, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_41gcode_weight(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gcode_weight (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 187, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.gcode_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_40gcode_weight(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_40gcode_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gcode_weight", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":188
 * 
 * def gcode_weight(uint32_t v1):
 *     return mat24_gcode_weight(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_gcode_weight(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":187
 * 
 * 
 * def gcode_weight(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_weight(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.gcode_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":191
 * 
 * 
 * def gcode_to_bit_list(uint32_t  v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t length
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_43gcode_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_43gcode_to_bit_list = {"gcode_to_bit_list", (PyCFunction)__pyx_pw_7mmgroup_5mat24_43gcode_to_bit_list, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_43gcode_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gcode_to_bit_list (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_42gcode_to_bit_list(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_42gcode_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  uint8_t __pyx_v_data[24];
  uint32_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("gcode_to_bit_list", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":194
 *     cdef uint8_t data[24]
 *     cdef uint32_t length
 *     length = mat24_gcode_to_bit_list(v1, &data[0])             # <<<<<<<<<<<<<<
 *     return list(data[:length])
 * 
 */
  __pyx_v_length = mat24_gcode_to_bit_list(__pyx_v_v1, (&(__pyx_v_data[0])));

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":195
 *     cdef uint32_t length
 *     length = mat24_gcode_to_bit_list(v1, &data[0])
 *     return list(data[:length])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_data) + 0, __pyx_v_length - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":191
 * 
 * 
 * def gcode_to_bit_list(uint32_t  v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t length
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.gcode_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":199
 * 
 * 
 * def cocode_weight(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_weight(c1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_45cocode_weight(PyObject *__pyx_self, PyObject *__pyx_arg_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_45cocode_weight = {"cocode_weight", (PyCFunction)__pyx_pw_7mmgroup_5mat24_45cocode_weight, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_45cocode_weight(PyObject *__pyx_self, PyObject *__pyx_arg_c1) {
  uint32_t __pyx_v_c1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_weight (wrapper)", 0);
  assert(__pyx_arg_c1); {
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_c1); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_44cocode_weight(__pyx_self, ((uint32_t)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_44cocode_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_weight", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":200
 * 
 * def cocode_weight(uint32_t  c1):
 *     return mat24_cocode_weight(c1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_cocode_weight(__pyx_v_c1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":199
 * 
 * 
 * def cocode_weight(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_weight(c1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":203
 * 
 * 
 * def cocode_to_bit_list(uint32_t  c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[4]
 *     cdef uint32_t length
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_47cocode_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_47cocode_to_bit_list = {"cocode_to_bit_list", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_47cocode_to_bit_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_47cocode_to_bit_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_c1;
  uint32_t __pyx_v_u_tetrad;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_to_bit_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_u_tetrad,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_tetrad);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cocode_to_bit_list") < 0)) __PYX_ERR(0, 203, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_u_tetrad = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_tetrad == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
    } else {
      __pyx_v_u_tetrad = ((uint32_t)24);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cocode_to_bit_list", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 203, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_46cocode_to_bit_list(__pyx_self, __pyx_v_c1, __pyx_v_u_tetrad);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_46cocode_to_bit_list(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_u_tetrad) {
  uint8_t __pyx_v_data[4];
  uint32_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_to_bit_list", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":206
 *     cdef uint8_t data[4]
 *     cdef uint32_t length
 *     length = mat24_cocode_to_bit_list(c1, u_tetrad, &data[0])             # <<<<<<<<<<<<<<
 *     if length == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 */
  __pyx_v_length = mat24_cocode_to_bit_list(__pyx_v_c1, __pyx_v_u_tetrad, (&(__pyx_v_data[0])));

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":207
 *     cdef uint32_t length
 *     length = mat24_cocode_to_bit_list(c1, u_tetrad, &data[0])
 *     if length == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return list(data[:length])
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":208
 *     length = mat24_cocode_to_bit_list(c1, u_tetrad, &data[0])
 *     if length == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")             # <<<<<<<<<<<<<<
 *     return list(data[:length])
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 208, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":207
 *     cdef uint32_t length
 *     length = mat24_cocode_to_bit_list(c1, u_tetrad, &data[0])
 *     if length == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay code syndrome is not unique")
 *     return list(data[:length])
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":209
 *     if length == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")
 *     return list(data[:length])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_data) + 0, __pyx_v_length - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":203
 * 
 * 
 * def cocode_to_bit_list(uint32_t  c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[4]
 *     cdef uint32_t length
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_bit_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":212
 * 
 * 
 * def cocode_to_sextet(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_49cocode_to_sextet(PyObject *__pyx_self, PyObject *__pyx_arg_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_49cocode_to_sextet = {"cocode_to_sextet", (PyCFunction)__pyx_pw_7mmgroup_5mat24_49cocode_to_sextet, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_49cocode_to_sextet(PyObject *__pyx_self, PyObject *__pyx_arg_c1) {
  uint32_t __pyx_v_c1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_to_sextet (wrapper)", 0);
  assert(__pyx_arg_c1); {
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_c1); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_sextet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_48cocode_to_sextet(__pyx_self, ((uint32_t)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_48cocode_to_sextet(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1) {
  uint8_t __pyx_v_data[24];
  uint32_t __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_to_sextet", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":215
 *     cdef uint8_t data[24]
 *     cdef uint32_t result
 *     result = mat24_cocode_to_sextet(c1, &data[0])             # <<<<<<<<<<<<<<
 *     if result == 0xffffffff:
 *         raise ValueError("Golay cocode word is not a sextet")
 */
  __pyx_v_result = mat24_cocode_to_sextet(__pyx_v_c1, (&(__pyx_v_data[0])));

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":216
 *     cdef uint32_t result
 *     result = mat24_cocode_to_sextet(c1, &data[0])
 *     if result == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay cocode word is not a sextet")
 *     return list(data[:24])
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":217
 *     result = mat24_cocode_to_sextet(c1, &data[0])
 *     if result == 0xffffffff:
 *         raise ValueError("Golay cocode word is not a sextet")             # <<<<<<<<<<<<<<
 *     return list(data[:24])
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 217, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":216
 *     cdef uint32_t result
 *     result = mat24_cocode_to_sextet(c1, &data[0])
 *     if result == 0xffffffff:             # <<<<<<<<<<<<<<
 *         raise ValueError("Golay cocode word is not a sextet")
 *     return list(data[:24])
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":218
 *     if result == 0xffffffff:
 *         raise ValueError("Golay cocode word is not a sextet")
 *     return list(data[:24])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_data) + 0, 24 - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":212
 * 
 * 
 * def cocode_to_sextet(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_sextet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":227
 * 
 * 
 * def scalar_prod(uint32_t v1, uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_scalar_prod(v1, c1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_51scalar_prod(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_51scalar_prod = {"scalar_prod", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_51scalar_prod, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_51scalar_prod(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_c1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scalar_prod (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_c1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scalar_prod", 1, 2, 2, 1); __PYX_ERR(0, 227, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scalar_prod") < 0)) __PYX_ERR(0, 227, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scalar_prod", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 227, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.scalar_prod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_50scalar_prod(__pyx_self, __pyx_v_v1, __pyx_v_c1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_50scalar_prod(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_c1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("scalar_prod", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":228
 * 
 * def scalar_prod(uint32_t v1, uint32_t c1):
 *     return mat24_scalar_prod(v1, c1)             # <<<<<<<<<<<<<<
 * 
 * ############################################################################
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_scalar_prod(__pyx_v_v1, __pyx_v_c1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":227
 * 
 * 
 * def scalar_prod(uint32_t v1, uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_scalar_prod(v1, c1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.scalar_prod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":235
 * 
 * 
 * def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_53suboctad_to_cocode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_53suboctad_to_cocode = {"suboctad_to_cocode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_53suboctad_to_cocode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_53suboctad_to_cocode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_u_sub;
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suboctad_to_cocode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_u_sub,&__pyx_n_s_v1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_sub)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("suboctad_to_cocode", 1, 2, 2, 1); __PYX_ERR(0, 235, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "suboctad_to_cocode") < 0)) __PYX_ERR(0, 235, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_u_sub = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_u_sub == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L3_error)
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("suboctad_to_cocode", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 235, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_to_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_52suboctad_to_cocode(__pyx_self, __pyx_v_u_sub, __pyx_v_v1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_52suboctad_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub, uint32_t __pyx_v_v1) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("suboctad_to_cocode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":237
 * def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)             # <<<<<<<<<<<<<<
 *     if res & 0xfffff000:
 *         raise ValueError("Attempt to compute suboctad of a non-octad")
 */
  __pyx_v_res = mat24_suboctad_to_cocode(__pyx_v_u_sub, __pyx_v_v1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":238
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 *     if res & 0xfffff000:             # <<<<<<<<<<<<<<
 *         raise ValueError("Attempt to compute suboctad of a non-octad")
 *     return res
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4294963200); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":239
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 *     if res & 0xfffff000:
 *         raise ValueError("Attempt to compute suboctad of a non-octad")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 239, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":238
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 *     if res & 0xfffff000:             # <<<<<<<<<<<<<<
 *         raise ValueError("Attempt to compute suboctad of a non-octad")
 *     return res
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":240
 *     if res & 0xfffff000:
 *         raise ValueError("Attempt to compute suboctad of a non-octad")
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":235
 * 
 * 
 * def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_to_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":243
 * 
 * 
 * def cocode_to_suboctad(uint32_t c1, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_55cocode_to_suboctad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_55cocode_to_suboctad = {"cocode_to_suboctad", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_55cocode_to_suboctad, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_55cocode_to_suboctad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_c1;
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_to_suboctad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_v1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cocode_to_suboctad", 1, 2, 2, 1); __PYX_ERR(0, 243, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cocode_to_suboctad") < 0)) __PYX_ERR(0, 243, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L3_error)
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cocode_to_suboctad", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 243, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_suboctad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_54cocode_to_suboctad(__pyx_self, __pyx_v_c1, __pyx_v_v1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_54cocode_to_suboctad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_v1) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_to_suboctad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":245
 * def cocode_to_suboctad(uint32_t c1, uint32_t v1):
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)             # <<<<<<<<<<<<<<
 *     if res & 0xfffff000:
 *         raise ValueError("Octad/suboctad mismatch")
 */
  __pyx_v_res = mat24_cocode_to_suboctad(__pyx_v_c1, __pyx_v_v1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":246
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 *     if res & 0xfffff000:             # <<<<<<<<<<<<<<
 *         raise ValueError("Octad/suboctad mismatch")
 *     return res
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4294963200); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":247
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 *     if res & 0xfffff000:
 *         raise ValueError("Octad/suboctad mismatch")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 247, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":246
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 *     if res & 0xfffff000:             # <<<<<<<<<<<<<<
 *         raise ValueError("Octad/suboctad mismatch")
 *     return res
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":248
 *     if res & 0xfffff000:
 *         raise ValueError("Octad/suboctad mismatch")
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":243
 * 
 * 
 * def cocode_to_suboctad(uint32_t c1, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_suboctad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":251
 * 
 * 
 * def suboctad_weight(uint32_t u_sub):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_weight(u_sub)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_57suboctad_weight(PyObject *__pyx_self, PyObject *__pyx_arg_u_sub); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_57suboctad_weight = {"suboctad_weight", (PyCFunction)__pyx_pw_7mmgroup_5mat24_57suboctad_weight, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_57suboctad_weight(PyObject *__pyx_self, PyObject *__pyx_arg_u_sub) {
  uint32_t __pyx_v_u_sub;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suboctad_weight (wrapper)", 0);
  assert(__pyx_arg_u_sub); {
    __pyx_v_u_sub = __Pyx_PyInt_As_uint32_t(__pyx_arg_u_sub); if (unlikely((__pyx_v_u_sub == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_56suboctad_weight(__pyx_self, ((uint32_t)__pyx_v_u_sub));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_56suboctad_weight(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("suboctad_weight", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":252
 * 
 * def suboctad_weight(uint32_t u_sub):
 *     return mat24_suboctad_weight(u_sub)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_suboctad_weight(__pyx_v_u_sub)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":251
 * 
 * 
 * def suboctad_weight(uint32_t u_sub):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_weight(u_sub)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_weight", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":255
 * 
 * 
 * def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_scalar_prod(u_sub1, u_sub2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_59suboctad_scalar_prod(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_59suboctad_scalar_prod = {"suboctad_scalar_prod", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_59suboctad_scalar_prod, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_59suboctad_scalar_prod(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_u_sub1;
  uint32_t __pyx_v_u_sub2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suboctad_scalar_prod (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_u_sub1,&__pyx_n_s_u_sub2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_sub1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_sub2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("suboctad_scalar_prod", 1, 2, 2, 1); __PYX_ERR(0, 255, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "suboctad_scalar_prod") < 0)) __PYX_ERR(0, 255, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_u_sub1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_u_sub1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
    __pyx_v_u_sub2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_sub2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("suboctad_scalar_prod", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 255, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_scalar_prod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_58suboctad_scalar_prod(__pyx_self, __pyx_v_u_sub1, __pyx_v_u_sub2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_58suboctad_scalar_prod(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_sub1, uint32_t __pyx_v_u_sub2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("suboctad_scalar_prod", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":256
 * 
 * def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):
 *     return mat24_suboctad_scalar_prod(u_sub1, u_sub2)             # <<<<<<<<<<<<<<
 * 
 * ###########################################################################
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_suboctad_scalar_prod(__pyx_v_u_sub1, __pyx_v_u_sub2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":255
 * 
 * 
 * def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_scalar_prod(u_sub1, u_sub2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.suboctad_scalar_prod", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":262
 * ###########################################################################
 * 
 * def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_61cocode_as_subdodecad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_61cocode_as_subdodecad = {"cocode_as_subdodecad", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_61cocode_as_subdodecad, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_61cocode_as_subdodecad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_c1;
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_single;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_as_subdodecad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_v1,&__pyx_n_s_u_single,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cocode_as_subdodecad", 0, 2, 3, 1); __PYX_ERR(0, 262, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_single);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cocode_as_subdodecad") < 0)) __PYX_ERR(0, 262, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 262, __pyx_L3_error)
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 262, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_u_single = __Pyx_PyInt_As_uint32_t(values[2]); if (unlikely((__pyx_v_u_single == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 262, __pyx_L3_error)
    } else {
      __pyx_v_u_single = ((uint32_t)24);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cocode_as_subdodecad", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.cocode_as_subdodecad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_60cocode_as_subdodecad(__pyx_self, __pyx_v_c1, __pyx_v_v1, __pyx_v_u_single);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_60cocode_as_subdodecad(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_single) {
  uint32_t __pyx_v_res;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_as_subdodecad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":264
 * def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)             # <<<<<<<<<<<<<<
 *     if res & 0xff000000:
 *         err = "Cocode element is not a subset of the docecad"
 */
  __pyx_v_res = mat24_cocode_as_subdodecad(__pyx_v_c1, __pyx_v_v1, __pyx_v_u_single);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":265
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 *     if res & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Cocode element is not a subset of the docecad"
 *         raise ValueError(err)
 */
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_4278190080); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":266
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 *     if res & 0xff000000:
 *         err = "Cocode element is not a subset of the docecad"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     return res
 */
    __Pyx_INCREF(__pyx_kp_u_Cocode_element_is_not_a_subset_o);
    __pyx_v_err = __pyx_kp_u_Cocode_element_is_not_a_subset_o;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":267
 *     if res & 0xff000000:
 *         err = "Cocode element is not a subset of the docecad"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 267, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":265
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 *     if res & 0xff000000:             # <<<<<<<<<<<<<<
 *         err = "Cocode element is not a subset of the docecad"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":268
 *         err = "Cocode element is not a subset of the docecad"
 *         raise ValueError(err)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":262
 * ###########################################################################
 * 
 * def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_as_subdodecad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":276
 * 
 * 
 * def ploop_theta(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_theta(v1)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_63ploop_theta(PyObject *__pyx_self, PyObject *__pyx_arg_v1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_63ploop_theta = {"ploop_theta", (PyCFunction)__pyx_pw_7mmgroup_5mat24_63ploop_theta, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_63ploop_theta(PyObject *__pyx_self, PyObject *__pyx_arg_v1) {
  uint32_t __pyx_v_v1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_theta (wrapper)", 0);
  assert(__pyx_arg_v1); {
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(__pyx_arg_v1); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 276, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.ploop_theta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_62ploop_theta(__pyx_self, ((uint32_t)__pyx_v_v1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_62ploop_theta(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_theta", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":277
 * 
 * def ploop_theta(uint32_t v1):
 *     return mat24_ploop_theta(v1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_ploop_theta(__pyx_v_v1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":276
 * 
 * 
 * def ploop_theta(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_theta(v1)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_theta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":280
 * 
 * 
 * def ploop_cocycle(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cocycle(v1, v2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_65ploop_cocycle(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_65ploop_cocycle = {"ploop_cocycle", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_65ploop_cocycle, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_65ploop_cocycle(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_v2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_cocycle (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_v2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ploop_cocycle", 1, 2, 2, 1); __PYX_ERR(0, 280, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ploop_cocycle") < 0)) __PYX_ERR(0, 280, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L3_error)
    __pyx_v_v2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ploop_cocycle", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 280, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.ploop_cocycle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_64ploop_cocycle(__pyx_self, __pyx_v_v1, __pyx_v_v2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_64ploop_cocycle(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_cocycle", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":281
 * 
 * def ploop_cocycle(uint32_t v1, uint32_t v2):
 *     return mat24_ploop_cocycle(v1, v2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_ploop_cocycle(__pyx_v_v1, __pyx_v_v2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":280
 * 
 * 
 * def ploop_cocycle(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cocycle(v1, v2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_cocycle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":285
 * 
 * 
 * def mul_ploop(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_mul_ploop(v1, v2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_67mul_ploop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_67mul_ploop = {"mul_ploop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_67mul_ploop, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_67mul_ploop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_v2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul_ploop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_v2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mul_ploop", 1, 2, 2, 1); __PYX_ERR(0, 285, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mul_ploop") < 0)) __PYX_ERR(0, 285, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L3_error)
    __pyx_v_v2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mul_ploop", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 285, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.mul_ploop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_66mul_ploop(__pyx_self, __pyx_v_v1, __pyx_v_v2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_66mul_ploop(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul_ploop", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":286
 * 
 * def mul_ploop(uint32_t v1, uint32_t v2):
 *     return mat24_mul_ploop(v1, v2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_mul_ploop(__pyx_v_v1, __pyx_v_v2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":285
 * 
 * 
 * def mul_ploop(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_mul_ploop(v1, v2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.mul_ploop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":289
 * 
 * 
 * def pow_ploop(uint32_t v1, uint32_t u_exp):             # <<<<<<<<<<<<<<
 *     return mat24_pow_ploop(v1, u_exp)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_69pow_ploop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_69pow_ploop = {"pow_ploop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_69pow_ploop, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_69pow_ploop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_u_exp;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pow_ploop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_u_exp,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_u_exp)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pow_ploop", 1, 2, 2, 1); __PYX_ERR(0, 289, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pow_ploop") < 0)) __PYX_ERR(0, 289, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L3_error)
    __pyx_v_u_exp = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_u_exp == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pow_ploop", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 289, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.pow_ploop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_68pow_ploop(__pyx_self, __pyx_v_v1, __pyx_v_u_exp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_68pow_ploop(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_u_exp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pow_ploop", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":290
 * 
 * def pow_ploop(uint32_t v1, uint32_t u_exp):
 *     return mat24_pow_ploop(v1, u_exp)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_pow_ploop(__pyx_v_v1, __pyx_v_u_exp)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":289
 * 
 * 
 * def pow_ploop(uint32_t v1, uint32_t u_exp):             # <<<<<<<<<<<<<<
 *     return mat24_pow_ploop(v1, u_exp)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.pow_ploop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":294
 * 
 * 
 * def ploop_comm(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_comm(v1, v2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_71ploop_comm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_71ploop_comm = {"ploop_comm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_71ploop_comm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_71ploop_comm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_v2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_comm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_v2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ploop_comm", 1, 2, 2, 1); __PYX_ERR(0, 294, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ploop_comm") < 0)) __PYX_ERR(0, 294, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
    __pyx_v_v2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ploop_comm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 294, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.ploop_comm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_70ploop_comm(__pyx_self, __pyx_v_v1, __pyx_v_v2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_70ploop_comm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_comm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":295
 * 
 * def ploop_comm(uint32_t v1, uint32_t v2):
 *     return mat24_ploop_comm(v1, v2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_ploop_comm(__pyx_v_v1, __pyx_v_v2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":294
 * 
 * 
 * def ploop_comm(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_comm(v1, v2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_comm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":298
 * 
 * 
 * def ploop_cap(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cap(v1, v2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_73ploop_cap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_73ploop_cap = {"ploop_cap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_73ploop_cap, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_73ploop_cap(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_v2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_cap (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_v2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ploop_cap", 1, 2, 2, 1); __PYX_ERR(0, 298, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ploop_cap") < 0)) __PYX_ERR(0, 298, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
    __pyx_v_v2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ploop_cap", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 298, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.ploop_cap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_72ploop_cap(__pyx_self, __pyx_v_v1, __pyx_v_v2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_72ploop_cap(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_cap", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":299
 * 
 * def ploop_cap(uint32_t v1, uint32_t v2):
 *     return mat24_ploop_cap(v1, v2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_ploop_cap(__pyx_v_v1, __pyx_v_v2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":298
 * 
 * 
 * def ploop_cap(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cap(v1, v2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_cap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":302
 * 
 * 
 * def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_assoc(v1, v2, v3)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_75ploop_assoc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_75ploop_assoc = {"ploop_assoc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_75ploop_assoc, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_75ploop_assoc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  uint32_t __pyx_v_v2;
  uint32_t __pyx_v_v3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_assoc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_v2,&__pyx_n_s_v3,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ploop_assoc", 1, 3, 3, 1); __PYX_ERR(0, 302, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v3)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ploop_assoc", 1, 3, 3, 2); __PYX_ERR(0, 302, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ploop_assoc") < 0)) __PYX_ERR(0, 302, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
    __pyx_v_v2 = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_v2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
    __pyx_v_v3 = __Pyx_PyInt_As_uint32_t(values[2]); if (unlikely((__pyx_v_v3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ploop_assoc", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 302, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.ploop_assoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_74ploop_assoc(__pyx_self, __pyx_v_v1, __pyx_v_v2, __pyx_v_v3);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_74ploop_assoc(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, uint32_t __pyx_v_v2, uint32_t __pyx_v_v3) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_assoc", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":303
 * 
 * def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):
 *     return mat24_ploop_assoc(v1, v2, v3)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(mat24_ploop_assoc(__pyx_v_v1, __pyx_v_v2, __pyx_v_v3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":302
 * 
 * 
 * def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_assoc(v1, v2, v3)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_assoc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":309
 *      ploop_solve_size = 64
 * 
 * def ploop_solve(a):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pa[ploop_solve_size + 13]
 *     cdef unsigned int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_77ploop_solve(PyObject *__pyx_self, PyObject *__pyx_v_a); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_77ploop_solve = {"ploop_solve", (PyCFunction)__pyx_pw_7mmgroup_5mat24_77ploop_solve, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_77ploop_solve(PyObject *__pyx_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ploop_solve (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_76ploop_solve(__pyx_self, ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_76ploop_solve(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a) {
  uint32_t __pyx_v_pa[(__pyx_e_7mmgroup_5mat24_ploop_solve_size + 13)];
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_res;
  unsigned int __pyx_v_length;
  unsigned int __pyx_v_start;
  unsigned int __pyx_v_d;
  unsigned int __pyx_v_result_length;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  uint32_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ploop_solve", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":312
 *     cdef uint32_t pa[ploop_solve_size + 13]
 *     cdef unsigned int i
 *     cdef unsigned int res = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int length = len(a)
 *     cdef unsigned int start = 0
 */
  __pyx_v_res = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":313
 *     cdef unsigned int i
 *     cdef unsigned int res = 0
 *     cdef unsigned int length = len(a)             # <<<<<<<<<<<<<<
 *     cdef unsigned int start = 0
 *     cdef unsigned int d
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_a); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":314
 *     cdef unsigned int res = 0
 *     cdef unsigned int length = len(a)
 *     cdef unsigned int start = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int d
 *     cdef unsigned int result_length = 0
 */
  __pyx_v_start = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":316
 *     cdef unsigned int start = 0
 *     cdef unsigned int d
 *     cdef unsigned int result_length = 0             # <<<<<<<<<<<<<<
 *     while start < length:
 *         d = min(length - start, ploop_solve_size)
 */
  __pyx_v_result_length = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":317
 *     cdef unsigned int d
 *     cdef unsigned int result_length = 0
 *     while start < length:             # <<<<<<<<<<<<<<
 *         d = min(length - start, ploop_solve_size)
 *         for i in range(d):
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_start < __pyx_v_length) != 0);
    if (!__pyx_t_2) break;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":318
 *     cdef unsigned int result_length = 0
 *     while start < length:
 *         d = min(length - start, ploop_solve_size)             # <<<<<<<<<<<<<<
 *         for i in range(d):
 *             pa[result_length + i] = a[start + i]
 */
    __pyx_t_3 = __pyx_e_7mmgroup_5mat24_ploop_solve_size;
    __pyx_t_4 = (__pyx_v_length - __pyx_v_start);
    if (((__pyx_t_3 < __pyx_t_4) != 0)) {
      __pyx_t_5 = __pyx_t_3;
    } else {
      __pyx_t_5 = __pyx_t_4;
    }
    __pyx_v_d = __pyx_t_5;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":319
 *     while start < length:
 *         d = min(length - start, ploop_solve_size)
 *         for i in range(d):             # <<<<<<<<<<<<<<
 *             pa[result_length + i] = a[start + i]
 *         res = mat24_ploop_solve(pa, result_length + d)
 */
    __pyx_t_5 = __pyx_v_d;
    __pyx_t_4 = __pyx_t_5;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_4; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":320
 *         d = min(length - start, ploop_solve_size)
 *         for i in range(d):
 *             pa[result_length + i] = a[start + i]             # <<<<<<<<<<<<<<
 *         res = mat24_ploop_solve(pa, result_length + d)
 *         result_length = res >> 16
 */
      __pyx_t_7 = (__pyx_v_start + __pyx_v_i);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_a, __pyx_t_7, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_As_uint32_t(__pyx_t_8); if (unlikely((__pyx_t_9 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      (__pyx_v_pa[(__pyx_v_result_length + __pyx_v_i)]) = __pyx_t_9;
    }

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":321
 *         for i in range(d):
 *             pa[result_length + i] = a[start + i]
 *         res = mat24_ploop_solve(pa, result_length + d)             # <<<<<<<<<<<<<<
 *         result_length = res >> 16
 *         start += d
 */
    __pyx_v_res = mat24_ploop_solve(__pyx_v_pa, (__pyx_v_result_length + __pyx_v_d));

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":322
 *             pa[result_length + i] = a[start + i]
 *         res = mat24_ploop_solve(pa, result_length + d)
 *         result_length = res >> 16             # <<<<<<<<<<<<<<
 *         start += d
 *     if res & 0x1000:
 */
    __pyx_v_result_length = (__pyx_v_res >> 16);

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":323
 *         res = mat24_ploop_solve(pa, result_length + d)
 *         result_length = res >> 16
 *         start += d             # <<<<<<<<<<<<<<
 *     if res & 0x1000:
 *         err = "Cannot correct signs of all Parker loop elements"
 */
    __pyx_v_start = (__pyx_v_start + __pyx_v_d);
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":324
 *         result_length = res >> 16
 *         start += d
 *     if res & 0x1000:             # <<<<<<<<<<<<<<
 *         err = "Cannot correct signs of all Parker loop elements"
 *         raise ValueError(err)
 */
  __pyx_t_2 = ((__pyx_v_res & 0x1000) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":325
 *         start += d
 *     if res & 0x1000:
 *         err = "Cannot correct signs of all Parker loop elements"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     return res
 */
    __Pyx_INCREF(__pyx_kp_u_Cannot_correct_signs_of_all_Park);
    __pyx_v_err = __pyx_kp_u_Cannot_correct_signs_of_all_Park;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":326
 *     if res & 0x1000:
 *         err = "Cannot correct signs of all Parker loop elements"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 326, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":324
 *         result_length = res >> 16
 *         start += d
 *     if res & 0x1000:             # <<<<<<<<<<<<<<
 *         err = "Cannot correct signs of all Parker loop elements"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":327
 *         err = "Cannot correct signs of all Parker loop elements"
 *         raise ValueError(err)
 *     return res             # <<<<<<<<<<<<<<
 * 
 * ###########################################################################
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_int(__pyx_v_res); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":309
 *      ploop_solve_size = 64
 * 
 * def ploop_solve(a):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pa[ploop_solve_size + 13]
 *     cdef unsigned int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mmgroup.mat24.ploop_solve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":335
 * 
 * 
 * def perm_complete_heptad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_79perm_complete_heptad(PyObject *__pyx_self, PyObject *__pyx_v_p_io); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_79perm_complete_heptad = {"perm_complete_heptad", (PyCFunction)__pyx_pw_7mmgroup_5mat24_79perm_complete_heptad, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_79perm_complete_heptad(PyObject *__pyx_self, PyObject *__pyx_v_p_io) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_complete_heptad (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_78perm_complete_heptad(__pyx_self, ((PyObject *)__pyx_v_p_io));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_78perm_complete_heptad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p_io) {
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  uint32_t __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  uint8_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  uint8_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_complete_heptad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":337
 * def perm_complete_heptad(p_io):
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i])
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":339
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i])             # <<<<<<<<<<<<<<
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
 *     if res != 0:
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_p_io, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (__pyx_t_3 == Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if ((__pyx_t_4 != 0)) {
      __pyx_t_2 = 24;
    } else {
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_p_io, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_uint8_t(__pyx_t_5); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_2 = __pyx_t_6;
    }
    (__pyx_v_p1[__pyx_v_i]) = __pyx_t_2;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":340
 *     cdef int i
 *     for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i])
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)             # <<<<<<<<<<<<<<
 *     if res != 0:
 *         raise ValueError("Vector is not an umbral heptad")
 */
  __pyx_v_res = mat24_perm_complete_heptad(__pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":341
 *     for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i])
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
 *     if res != 0:             # <<<<<<<<<<<<<<
 *         raise ValueError("Vector is not an umbral heptad")
 *     for i in range(24): p_io[i]  = p1[i]
 */
  __pyx_t_4 = ((__pyx_v_res != 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":342
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
 *     if res != 0:
 *         raise ValueError("Vector is not an umbral heptad")             # <<<<<<<<<<<<<<
 *     for i in range(24): p_io[i]  = p1[i]
 *     return res
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 342, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":341
 *     for i in range(24): p1[i] = 24 if p_io[i] is None else int(p_io[i])
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
 *     if res != 0:             # <<<<<<<<<<<<<<
 *         raise ValueError("Vector is not an umbral heptad")
 *     for i in range(24): p_io[i]  = p1[i]
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":343
 *     if res != 0:
 *         raise ValueError("Vector is not an umbral heptad")
 *     for i in range(24): p_io[i]  = p1[i]             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_5 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_p_io, __pyx_v_i, __pyx_t_5, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":344
 *         raise ValueError("Vector is not an umbral heptad")
 *     for i in range(24): p_io[i]  = p1[i]
 *     return res             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":335
 * 
 * 
 * def perm_complete_heptad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mmgroup.mat24.perm_complete_heptad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":348
 * 
 * 
 * def perm_check(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_81perm_check(PyObject *__pyx_self, PyObject *__pyx_v_p1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_81perm_check = {"perm_check", (PyCFunction)__pyx_pw_7mmgroup_5mat24_81perm_check, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_81perm_check(PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_check (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_80perm_check(__pyx_self, ((PyObject *)__pyx_v_p1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_80perm_check(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  int __pyx_v_i;
  uint32_t __pyx_v_acc;
  uint32_t __pyx_v_entry;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  uint32_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_check", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":350
 * def perm_check(p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef uint32_t acc = 0
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":352
 *     cdef uint8_t *p_p1a = p1a
 *     cdef int i
 *     cdef uint32_t acc = 0             # <<<<<<<<<<<<<<
 *     cdef uint32_t entry
 *     if len(p1) != 24:
 */
  __pyx_v_acc = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":354
 *     cdef uint32_t acc = 0
 *     cdef uint32_t entry
 *     if len(p1) != 24:             # <<<<<<<<<<<<<<
 *         err = "Permutation in group Mat24 must have length 24"
 *         raise ValueError(err)
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_p1); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 354, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 != 24) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":355
 *     cdef uint32_t entry
 *     if len(p1) != 24:
 *         err = "Permutation in group Mat24 must have length 24"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     for i in range(24):
 */
    __Pyx_INCREF(__pyx_kp_u_Permutation_in_group_Mat24_must);
    __pyx_v_err = __pyx_kp_u_Permutation_in_group_Mat24_must;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":356
 *     if len(p1) != 24:
 *         err = "Permutation in group Mat24 must have length 24"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     for i in range(24):
 *         entry = int(p1[i])
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 356, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":354
 *     cdef uint32_t acc = 0
 *     cdef uint32_t entry
 *     if len(p1) != 24:             # <<<<<<<<<<<<<<
 *         err = "Permutation in group Mat24 must have length 24"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":357
 *         err = "Permutation in group Mat24 must have length 24"
 *         raise ValueError(err)
 *     for i in range(24):             # <<<<<<<<<<<<<<
 *         entry = int(p1[i])
 *         p1a[i] = entry
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 24; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":358
 *         raise ValueError(err)
 *     for i in range(24):
 *         entry = int(p1[i])             # <<<<<<<<<<<<<<
 *         p1a[i] = entry
 *         acc |= entry
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_uint32_t(__pyx_t_5); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_entry = __pyx_t_6;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":359
 *     for i in range(24):
 *         entry = int(p1[i])
 *         p1a[i] = entry             # <<<<<<<<<<<<<<
 *         acc |= entry
 *     if acc  & -32 or mat24_perm_check(p_p1a):
 */
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_v_entry;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":360
 *         entry = int(p1[i])
 *         p1a[i] = entry
 *         acc |= entry             # <<<<<<<<<<<<<<
 *     if acc  & -32 or mat24_perm_check(p_p1a):
 *         err =   "Permutation is not in group Mat24"
 */
    __pyx_v_acc = (__pyx_v_acc | __pyx_v_entry);
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":361
 *         p1a[i] = entry
 *         acc |= entry
 *     if acc  & -32 or mat24_perm_check(p_p1a):             # <<<<<<<<<<<<<<
 *         err =   "Permutation is not in group Mat24"
 *         raise ValueError(err)
 */
  __pyx_t_7 = ((__pyx_v_acc & -32L) != 0);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_7 = (mat24_perm_check(__pyx_v_p_p1a) != 0);
  __pyx_t_2 = __pyx_t_7;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":362
 *         acc |= entry
 *     if acc  & -32 or mat24_perm_check(p_p1a):
 *         err =   "Permutation is not in group Mat24"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 * 
 */
    __Pyx_INCREF(__pyx_kp_u_Permutation_is_not_in_group_Mat2);
    __pyx_v_err = __pyx_kp_u_Permutation_is_not_in_group_Mat2;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":363
 *     if acc  & -32 or mat24_perm_check(p_p1a):
 *         err =   "Permutation is not in group Mat24"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 363, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":361
 *         p1a[i] = entry
 *         acc |= entry
 *     if acc  & -32 or mat24_perm_check(p_p1a):             # <<<<<<<<<<<<<<
 *         err =   "Permutation is not in group Mat24"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":348
 * 
 * 
 * def perm_check(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("mmgroup.mat24.perm_check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":366
 * 
 * 
 * def perm_complete_octad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[8]
 *     cdef uint8_t *p_p1 = p1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_83perm_complete_octad(PyObject *__pyx_self, PyObject *__pyx_v_p_io); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_83perm_complete_octad = {"perm_complete_octad", (PyCFunction)__pyx_pw_7mmgroup_5mat24_83perm_complete_octad, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_83perm_complete_octad(PyObject *__pyx_self, PyObject *__pyx_v_p_io) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_complete_octad (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_82perm_complete_octad(__pyx_self, ((PyObject *)__pyx_v_p_io));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_82perm_complete_octad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p_io) {
  uint8_t __pyx_v_p1[8];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  uint32_t __pyx_v_res;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  uint8_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_complete_octad", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":368
 * def perm_complete_octad(p_io):
 *     cdef uint8_t p1[8]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     p1[5] = 24
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":370
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     p1[5] = 24             # <<<<<<<<<<<<<<
 *     for i in range(5): p1[i] = p_io[i]
 *     if p_io[5] is not None: p1[5] = p_io[5]
 */
  (__pyx_v_p1[5]) = 24;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":371
 *     cdef int i
 *     p1[5] = 24
 *     for i in range(5): p1[i] = p_io[i]             # <<<<<<<<<<<<<<
 *     if p_io[5] is not None: p1[5] = p_io[5]
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 5; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p_io, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_p1[__pyx_v_i]) = __pyx_t_3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":372
 *     p1[5] = 24
 *     for i in range(5): p1[i] = p_io[i]
 *     if p_io[5] is not None: p1[5] = p_io[5]             # <<<<<<<<<<<<<<
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)
 *     if res:
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p_io, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p_io, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_p1[5]) = __pyx_t_3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":373
 *     for i in range(5): p1[i] = p_io[i]
 *     if p_io[5] is not None: p1[5] = p_io[5]
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)             # <<<<<<<<<<<<<<
 *     if res:
 *          err = "Cannot complete a vector to an octad"
 */
  __pyx_v_res = mat24_perm_complete_octad(__pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":374
 *     if p_io[5] is not None: p1[5] = p_io[5]
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *          err = "Cannot complete a vector to an octad"
 *          raise ValueError(err)
 */
  __pyx_t_5 = (__pyx_v_res != 0);
  if (unlikely(__pyx_t_5)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":375
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)
 *     if res:
 *          err = "Cannot complete a vector to an octad"             # <<<<<<<<<<<<<<
 *          raise ValueError(err)
 *     for i in range(8): p_io[i]  = p1[i]
 */
    __Pyx_INCREF(__pyx_kp_u_Cannot_complete_a_vector_to_an_o);
    __pyx_v_err = __pyx_kp_u_Cannot_complete_a_vector_to_an_o;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":376
 *     if res:
 *          err = "Cannot complete a vector to an octad"
 *          raise ValueError(err)             # <<<<<<<<<<<<<<
 *     for i in range(8): p_io[i]  = p1[i]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 376, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":374
 *     if p_io[5] is not None: p1[5] = p_io[5]
 *     cdef uint32_t res = mat24_perm_complete_octad(p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *          err = "Cannot complete a vector to an octad"
 *          raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":377
 *          err = "Cannot complete a vector to an octad"
 *          raise ValueError(err)
 *     for i in range(8): p_io[i]  = p1[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 8; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_SetItemInt(__pyx_v_p_io, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":366
 * 
 * 
 * def perm_complete_octad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[8]
 *     cdef uint8_t *p_p1 = p1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("mmgroup.mat24.perm_complete_octad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":380
 * 
 * 
 * def perm_from_heptads(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[7]
 *     cdef uint8_t *p_h1a = h1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_85perm_from_heptads(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_85perm_from_heptads = {"perm_from_heptads", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_85perm_from_heptads, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_85perm_from_heptads(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_h1 = 0;
  PyObject *__pyx_v_h2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_from_heptads (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_h1,&__pyx_n_s_h2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perm_from_heptads", 1, 2, 2, 1); __PYX_ERR(0, 380, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perm_from_heptads") < 0)) __PYX_ERR(0, 380, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_h1 = values[0];
    __pyx_v_h2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perm_from_heptads", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 380, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_heptads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_84perm_from_heptads(__pyx_self, __pyx_v_h1, __pyx_v_h2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_84perm_from_heptads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_h1, PyObject *__pyx_v_h2) {
  uint8_t __pyx_v_h1a[7];
  uint8_t *__pyx_v_p_h1a;
  uint8_t __pyx_v_h2a[7];
  uint8_t *__pyx_v_p_h2a;
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  uint32_t __pyx_v_res;
  PyObject *__pyx_v_err = NULL;
  int __pyx_8genexpr2__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_from_heptads", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":382
 * def perm_from_heptads(h1, h2):
 *     cdef uint8_t h1a[7]
 *     cdef uint8_t *p_h1a = h1a             # <<<<<<<<<<<<<<
 *     cdef uint8_t h2a[7]
 *     cdef uint8_t *p_h2a = h2a
 */
  __pyx_v_p_h1a = __pyx_v_h1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":384
 *     cdef uint8_t *p_h1a = h1a
 *     cdef uint8_t h2a[7]
 *     cdef uint8_t *p_h2a = h2a             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_v_p_h2a = __pyx_v_h2a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":386
 *     cdef uint8_t *p_h2a = h2a
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(7):
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":388
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     for i in range(7):             # <<<<<<<<<<<<<<
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 7; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":389
 *     cdef int i
 *     for i in range(7):
 *         h1a[i] = int(h1[i])             # <<<<<<<<<<<<<<
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_h1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_h1a[__pyx_v_i]) = __pyx_t_4;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":390
 *     for i in range(7):
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])             # <<<<<<<<<<<<<<
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
 *     if res:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_h2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_h2a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":391
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)             # <<<<<<<<<<<<<<
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from heptads"
 */
  __pyx_v_res = mat24_perm_from_heptads(__pyx_v_p_h1a, __pyx_v_p_h2a, __pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":392
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *         err = "Cannot construct permutation in Mat24 from heptads"
 *         raise ValueError(err)
 */
  __pyx_t_5 = (__pyx_v_res != 0);
  if (unlikely(__pyx_t_5)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":393
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from heptads"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     return [p1[i] for i in range(24)]
 */
    __Pyx_INCREF(__pyx_kp_u_Cannot_construct_permutation_in);
    __pyx_v_err = __pyx_kp_u_Cannot_construct_permutation_in;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":394
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from heptads"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     return [p1[i] for i in range(24)]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 394, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":392
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_heptads(p_h1a, p_h2a, p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *         err = "Cannot construct permutation in Mat24 from heptads"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":395
 *         err = "Cannot construct permutation in Mat24 from heptads"
 *         raise ValueError(err)
 *     return [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_8genexpr2__pyx_v_i = __pyx_t_1;
      __pyx_t_3 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_8genexpr2__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":380
 * 
 * 
 * def perm_from_heptads(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[7]
 *     cdef uint8_t *p_h1a = h1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_heptads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":398
 * 
 * 
 * def perm_from_map(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[24]
 *     cdef uint8_t *p_h1a = h1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_87perm_from_map(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_87perm_from_map = {"perm_from_map", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_87perm_from_map, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_87perm_from_map(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_h1 = 0;
  PyObject *__pyx_v_h2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_from_map (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_h1,&__pyx_n_s_h2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perm_from_map", 1, 2, 2, 1); __PYX_ERR(0, 398, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perm_from_map") < 0)) __PYX_ERR(0, 398, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_h1 = values[0];
    __pyx_v_h2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perm_from_map", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 398, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_86perm_from_map(__pyx_self, __pyx_v_h1, __pyx_v_h2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_86perm_from_map(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_h1, PyObject *__pyx_v_h2) {
  uint8_t __pyx_v_h1a[24];
  uint8_t *__pyx_v_p_h1a;
  uint8_t __pyx_v_h2a[24];
  uint8_t *__pyx_v_p_h2a;
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  uint32_t __pyx_v_length;
  PyObject *__pyx_v_err = NULL;
  uint32_t __pyx_v_i;
  uint32_t __pyx_v_res;
  long __pyx_8genexpr3__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  uint32_t __pyx_t_5;
  uint32_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  uint8_t __pyx_t_8;
  long __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_from_map", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":400
 * def perm_from_map(h1, h2):
 *     cdef uint8_t h1a[24]
 *     cdef uint8_t *p_h1a = h1a             # <<<<<<<<<<<<<<
 *     cdef uint8_t h2a[24]
 *     cdef uint8_t *p_h2a = h2a
 */
  __pyx_v_p_h1a = __pyx_v_h1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":402
 *     cdef uint8_t *p_h1a = h1a
 *     cdef uint8_t h2a[24]
 *     cdef uint8_t *p_h2a = h2a             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_v_p_h2a = __pyx_v_h2a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":404
 *     cdef uint8_t *p_h2a = h2a
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef uint32_t length = len(h1)
 *     if len(h2) != length:
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":405
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 *     cdef uint32_t length = len(h1)             # <<<<<<<<<<<<<<
 *     if len(h2) != length:
 *         err = "Arrays in function perm_from_map() have different lengths"
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_h1); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":406
 *     cdef uint8_t *p_p1 = p1
 *     cdef uint32_t length = len(h1)
 *     if len(h2) != length:             # <<<<<<<<<<<<<<
 *         err = "Arrays in function perm_from_map() have different lengths"
 *         raise ValueError(err)
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_h2); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 406, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 != __pyx_v_length) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":407
 *     cdef uint32_t length = len(h1)
 *     if len(h2) != length:
 *         err = "Arrays in function perm_from_map() have different lengths"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     if length > 24:
 */
    __Pyx_INCREF(__pyx_kp_u_Arrays_in_function_perm_from_map);
    __pyx_v_err = __pyx_kp_u_Arrays_in_function_perm_from_map;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":408
 *     if len(h2) != length:
 *         err = "Arrays in function perm_from_map() have different lengths"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     if length > 24:
 *         err = "Array in function perm_from_map() is too long"
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 408, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":406
 *     cdef uint8_t *p_p1 = p1
 *     cdef uint32_t length = len(h1)
 *     if len(h2) != length:             # <<<<<<<<<<<<<<
 *         err = "Arrays in function perm_from_map() have different lengths"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":409
 *         err = "Arrays in function perm_from_map() have different lengths"
 *         raise ValueError(err)
 *     if length > 24:             # <<<<<<<<<<<<<<
 *         err = "Array in function perm_from_map() is too long"
 *         raise ValueError(err)
 */
  __pyx_t_2 = ((__pyx_v_length > 24) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":410
 *         raise ValueError(err)
 *     if length > 24:
 *         err = "Array in function perm_from_map() is too long"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     for i in range(length):
 */
    __Pyx_INCREF(__pyx_kp_u_Array_in_function_perm_from_map);
    __pyx_v_err = __pyx_kp_u_Array_in_function_perm_from_map;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":411
 *     if length > 24:
 *         err = "Array in function perm_from_map() is too long"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     for i in range(length):
 *         h1a[i] = int(h1[i])
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 411, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":409
 *         err = "Arrays in function perm_from_map() have different lengths"
 *         raise ValueError(err)
 *     if length > 24:             # <<<<<<<<<<<<<<
 *         err = "Array in function perm_from_map() is too long"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":412
 *         err = "Array in function perm_from_map() is too long"
 *         raise ValueError(err)
 *     for i in range(length):             # <<<<<<<<<<<<<<
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])
 */
  __pyx_t_4 = __pyx_v_length;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":413
 *         raise ValueError(err)
 *     for i in range(length):
 *         h1a[i] = int(h1[i])             # <<<<<<<<<<<<<<
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_h1, __pyx_v_i, uint32_t, 0, __Pyx_PyInt_From_uint32_t, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_uint8_t(__pyx_t_7); if (unlikely((__pyx_t_8 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    (__pyx_v_h1a[__pyx_v_i]) = __pyx_t_8;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":414
 *     for i in range(length):
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])             # <<<<<<<<<<<<<<
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
 *     if res < 0:
 */
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_h2, __pyx_v_i, uint32_t, 0, __Pyx_PyInt_From_uint32_t, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_8 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_h2a[__pyx_v_i]) = __pyx_t_8;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":415
 *         h1a[i] = int(h1[i])
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)             # <<<<<<<<<<<<<<
 *     if res < 0:
 *         err = "Illegal entry in array in function perm_from_map()"
 */
  __pyx_v_res = mat24_perm_from_map(__pyx_v_p_h1a, __pyx_v_p_h2a, __pyx_v_length, __pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":416
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
 *     if res < 0:             # <<<<<<<<<<<<<<
 *         err = "Illegal entry in array in function perm_from_map()"
 *         raise ValueError(err)
 */
  __pyx_t_2 = ((__pyx_v_res < 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":417
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
 *     if res < 0:
 *         err = "Illegal entry in array in function perm_from_map()"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     return res, [p1[i] for i in range(24)]
 */
    __Pyx_INCREF(__pyx_kp_u_Illegal_entry_in_array_in_functi);
    __pyx_v_err = __pyx_kp_u_Illegal_entry_in_array_in_functi;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":418
 *     if res < 0:
 *         err = "Illegal entry in array in function perm_from_map()"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     return res, [p1[i] for i in range(24)]
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 418, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":416
 *         h2a[i] = int(h2[i])
 *     cdef uint32_t res = mat24_perm_from_map(p_h1a, p_h2a, length, p_p1)
 *     if res < 0:             # <<<<<<<<<<<<<<
 *         err = "Illegal entry in array in function perm_from_map()"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":419
 *         err = "Illegal entry in array in function perm_from_map()"
 *         raise ValueError(err)
 *     return res, [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(__pyx_v_res); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { /* enter inner scope */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    for (__pyx_t_9 = 0; __pyx_t_9 < 24; __pyx_t_9+=1) {
      __pyx_8genexpr3__pyx_v_i = __pyx_t_9;
      __pyx_t_10 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_8genexpr3__pyx_v_i])); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 419, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 419, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
  } /* exit inner scope */
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
  __pyx_t_3 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":398
 * 
 * 
 * def perm_from_map(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[24]
 *     cdef uint8_t *p_h1a = h1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":423
 * 
 * 
 * def m24num_to_perm(uint32_t u_m24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_89m24num_to_perm(PyObject *__pyx_self, PyObject *__pyx_arg_u_m24); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_89m24num_to_perm = {"m24num_to_perm", (PyCFunction)__pyx_pw_7mmgroup_5mat24_89m24num_to_perm, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_89m24num_to_perm(PyObject *__pyx_self, PyObject *__pyx_arg_u_m24) {
  uint32_t __pyx_v_u_m24;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("m24num_to_perm (wrapper)", 0);
  assert(__pyx_arg_u_m24); {
    __pyx_v_u_m24 = __Pyx_PyInt_As_uint32_t(__pyx_arg_u_m24); if (unlikely((__pyx_v_u_m24 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 423, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.m24num_to_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_88m24num_to_perm(__pyx_self, ((uint32_t)__pyx_v_u_m24));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_88m24num_to_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_u_m24) {
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  PyObject *__pyx_v_err = NULL;
  int __pyx_8genexpr4__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("m24num_to_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":425
 * def m24num_to_perm(uint32_t u_m24):
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     if mat24_m24num_to_perm(u_m24, p_p1):
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":427
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     if mat24_m24num_to_perm(u_m24, p_p1):             # <<<<<<<<<<<<<<
 *         err = "Illegal number for a permutation in Mat24"
 *         raise ValueError, err
 */
  __pyx_t_1 = (mat24_m24num_to_perm(__pyx_v_u_m24, __pyx_v_p_p1) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":428
 *     cdef int i
 *     if mat24_m24num_to_perm(u_m24, p_p1):
 *         err = "Illegal number for a permutation in Mat24"             # <<<<<<<<<<<<<<
 *         raise ValueError, err
 *     return [p1[i] for i in range(24)]
 */
    __Pyx_INCREF(__pyx_kp_u_Illegal_number_for_a_permutation);
    __pyx_v_err = __pyx_kp_u_Illegal_number_for_a_permutation;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":429
 *     if mat24_m24num_to_perm(u_m24, p_p1):
 *         err = "Illegal number for a permutation in Mat24"
 *         raise ValueError, err             # <<<<<<<<<<<<<<
 *     return [p1[i] for i in range(24)]
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_v_err, 0, 0);
    __PYX_ERR(0, 429, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":427
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     if mat24_m24num_to_perm(u_m24, p_p1):             # <<<<<<<<<<<<<<
 *         err = "Illegal number for a permutation in Mat24"
 *         raise ValueError, err
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":430
 *         err = "Illegal number for a permutation in Mat24"
 *         raise ValueError, err
 *     return [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_3 = 0; __pyx_t_3 < 24; __pyx_t_3+=1) {
      __pyx_8genexpr4__pyx_v_i = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_8genexpr4__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 430, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":423
 * 
 * 
 * def m24num_to_perm(uint32_t u_m24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mmgroup.mat24.m24num_to_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":435
 * 
 * 
 * def perm_to_m24num(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_91perm_to_m24num(PyObject *__pyx_self, PyObject *__pyx_v_p1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_91perm_to_m24num = {"perm_to_m24num", (PyCFunction)__pyx_pw_7mmgroup_5mat24_91perm_to_m24num, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_91perm_to_m24num(PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_to_m24num (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_90perm_to_m24num(__pyx_self, ((PyObject *)__pyx_v_p1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_90perm_to_m24num(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_to_m24num", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":437
 * def perm_to_m24num(p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":439
 *     cdef uint8_t *p_p1a = p1a
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     return  mat24_perm_to_m24num(p_p1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":440
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     return  mat24_perm_to_m24num(p_p1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_perm_to_m24num(__pyx_v_p_p1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":435
 * 
 * 
 * def perm_to_m24num(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_m24num", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":443
 * 
 * 
 * def perm_to_matrix(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_93perm_to_matrix(PyObject *__pyx_self, PyObject *__pyx_v_p1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_93perm_to_matrix = {"perm_to_matrix", (PyCFunction)__pyx_pw_7mmgroup_5mat24_93perm_to_matrix, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_93perm_to_matrix(PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_to_matrix (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_92perm_to_matrix(__pyx_self, ((PyObject *)__pyx_v_p1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_92perm_to_matrix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  uint32_t __pyx_v_m1[12];
  uint32_t *__pyx_v_p_m1;
  int __pyx_v_i;
  int __pyx_8genexpr5__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_to_matrix", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":445
 * def perm_to_matrix(p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":447
 *     cdef uint8_t *p_p1a = p1a
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_m1 = __pyx_v_m1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":449
 *     cdef uint32_t *p_m1 = m1
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     mat24_perm_to_matrix(p_p1a, p_m1)
 *     return [m1[i] for i in range(12)]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":450
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_matrix(p_p1a, p_m1)             # <<<<<<<<<<<<<<
 *     return [m1[i] for i in range(12)]
 * 
 */
  mat24_perm_to_matrix(__pyx_v_p_p1a, __pyx_v_p_m1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":451
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_matrix(p_p1a, p_m1)
 *     return [m1[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
      __pyx_8genexpr5__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint32_t((__pyx_v_m1[__pyx_8genexpr5__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":443
 * 
 * 
 * def perm_to_matrix(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_matrix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":454
 * 
 * 
 * def matrix_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_95matrix_to_perm(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_95matrix_to_perm = {"matrix_to_perm", (PyCFunction)__pyx_pw_7mmgroup_5mat24_95matrix_to_perm, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_95matrix_to_perm(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("matrix_to_perm (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_94matrix_to_perm(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_94matrix_to_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  uint32_t *__pyx_v_p_m1a;
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  int __pyx_8genexpr6__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("matrix_to_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":456
 * def matrix_to_perm(m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":458
 *     cdef uint32_t *p_m1a = m1a
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":460
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     mat24_matrix_to_perm(p_m1a, p_p1)
 *     return [p1[i] for i in range(24)]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":461
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_matrix_to_perm(p_m1a, p_p1)             # <<<<<<<<<<<<<<
 *     return [p1[i] for i in range(24)]
 * 
 */
  mat24_matrix_to_perm(__pyx_v_p_m1a, __pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":462
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_matrix_to_perm(p_m1a, p_p1)
 *     return [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * def matrix_from_mod_omega(m1):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_8genexpr6__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_8genexpr6__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 462, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":454
 * 
 * 
 * def matrix_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.matrix_to_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":464
 *     return [p1[i] for i in range(24)]
 * 
 * def matrix_from_mod_omega(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_97matrix_from_mod_omega(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_97matrix_from_mod_omega = {"matrix_from_mod_omega", (PyCFunction)__pyx_pw_7mmgroup_5mat24_97matrix_from_mod_omega, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_97matrix_from_mod_omega(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("matrix_from_mod_omega (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_96matrix_from_mod_omega(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_96matrix_from_mod_omega(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("matrix_from_mod_omega", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":466
 * def matrix_from_mod_omega(m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":468
 *     cdef uint32_t *p_m1a = m1a
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     mat24_matrix_from_mod_omega(m1a)
 *     for i in range(12): m1[i] = int(m1a[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":469
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_matrix_from_mod_omega(m1a)             # <<<<<<<<<<<<<<
 *     for i in range(12): m1[i] = int(m1a[i])
 * 
 */
  mat24_matrix_from_mod_omega(__pyx_v_m1a);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":470
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_matrix_from_mod_omega(m1a)
 *     for i in range(12): m1[i] = int(m1a[i])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_3 = __Pyx_PyInt_From_uint32_t((__pyx_v_m1a[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__Pyx_SetItemInt(__pyx_v_m1, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":464
 *     return [p1[i] for i in range(24)]
 * 
 * def matrix_from_mod_omega(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.matrix_from_mod_omega", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":479
 * 
 * 
 * def perm_from_dodecads(d1, d2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t d1a[9]
 *     cdef uint8_t *p_d1a = d1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_99perm_from_dodecads(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_99perm_from_dodecads = {"perm_from_dodecads", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_99perm_from_dodecads, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_99perm_from_dodecads(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_d1 = 0;
  PyObject *__pyx_v_d2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_from_dodecads (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_d1,&__pyx_n_s_d2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perm_from_dodecads", 1, 2, 2, 1); __PYX_ERR(0, 479, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perm_from_dodecads") < 0)) __PYX_ERR(0, 479, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_d1 = values[0];
    __pyx_v_d2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perm_from_dodecads", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_dodecads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_98perm_from_dodecads(__pyx_self, __pyx_v_d1, __pyx_v_d2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_98perm_from_dodecads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_d1, PyObject *__pyx_v_d2) {
  uint8_t __pyx_v_d1a[9];
  uint8_t *__pyx_v_p_d1a;
  uint8_t __pyx_v_d2a[9];
  uint8_t *__pyx_v_p_d2a;
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  uint32_t __pyx_v_res;
  PyObject *__pyx_v_err = NULL;
  int __pyx_8genexpr7__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_from_dodecads", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":481
 * def perm_from_dodecads(d1, d2):
 *     cdef uint8_t d1a[9]
 *     cdef uint8_t *p_d1a = d1a             # <<<<<<<<<<<<<<
 *     cdef uint8_t d2a[9]
 *     cdef uint8_t *p_d2a = d2a
 */
  __pyx_v_p_d1a = __pyx_v_d1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":483
 *     cdef uint8_t *p_d1a = d1a
 *     cdef uint8_t d2a[9]
 *     cdef uint8_t *p_d2a = d2a             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_v_p_d2a = __pyx_v_d2a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":485
 *     cdef uint8_t *p_d2a = d2a
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(9):
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":487
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     for i in range(9):             # <<<<<<<<<<<<<<
 *         d1a[i] = int(d1[i])
 *         d2a[i] = int(d2[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 9; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":488
 *     cdef int i
 *     for i in range(9):
 *         d1a[i] = int(d1[i])             # <<<<<<<<<<<<<<
 *         d2a[i] = int(d2[i])
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_d1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_d1a[__pyx_v_i]) = __pyx_t_4;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":489
 *     for i in range(9):
 *         d1a[i] = int(d1[i])
 *         d2a[i] = int(d2[i])             # <<<<<<<<<<<<<<
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
 *     if res:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_d2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_d2a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":490
 *         d1a[i] = int(d1[i])
 *         d2a[i] = int(d2[i])
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)             # <<<<<<<<<<<<<<
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from dodecads"
 */
  __pyx_v_res = mat24_perm_from_dodecads(__pyx_v_p_d1a, __pyx_v_p_d2a, __pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":491
 *         d2a[i] = int(d2[i])
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *         err = "Cannot construct permutation in Mat24 from dodecads"
 *         raise ValueError(err)
 */
  __pyx_t_5 = (__pyx_v_res != 0);
  if (unlikely(__pyx_t_5)) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":492
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from dodecads"             # <<<<<<<<<<<<<<
 *         raise ValueError(err)
 *     return [p1[i] for i in range(24)]
 */
    __Pyx_INCREF(__pyx_kp_u_Cannot_construct_permutation_in_2);
    __pyx_v_err = __pyx_kp_u_Cannot_construct_permutation_in_2;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":493
 *     if res:
 *         err = "Cannot construct permutation in Mat24 from dodecads"
 *         raise ValueError(err)             # <<<<<<<<<<<<<<
 *     return [p1[i] for i in range(24)]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 493, __pyx_L1_error)

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":491
 *         d2a[i] = int(d2[i])
 *     cdef uint32_t res = mat24_perm_from_dodecads(p_d1a, p_d2a, p_p1)
 *     if res:             # <<<<<<<<<<<<<<
 *         err = "Cannot construct permutation in Mat24 from dodecads"
 *         raise ValueError(err)
 */
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":494
 *         err = "Cannot construct permutation in Mat24 from dodecads"
 *         raise ValueError(err)
 *     return [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_8genexpr7__pyx_v_i = __pyx_t_1;
      __pyx_t_3 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_8genexpr7__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":479
 * 
 * 
 * def perm_from_dodecads(d1, d2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t d1a[9]
 *     cdef uint8_t *p_d1a = d1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.perm_from_dodecads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":504
 * 
 * 
 * def op_vect_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_101op_vect_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_101op_vect_perm = {"op_vect_perm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_101op_vect_perm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_101op_vect_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  PyObject *__pyx_v_p1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_vect_perm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_p1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("op_vect_perm", 1, 2, 2, 1); __PYX_ERR(0, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "op_vect_perm") < 0)) __PYX_ERR(0, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
    __pyx_v_p1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("op_vect_perm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.op_vect_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_100op_vect_perm(__pyx_self, __pyx_v_v1, __pyx_v_p1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_100op_vect_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_vect_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":506
 * def op_vect_perm(uint32_t v1, p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":508
 *     cdef uint8_t *p_p1a = p1a
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     return mat24_op_vect_perm(v1, p_p1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":509
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     return mat24_op_vect_perm(v1, p_p1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_op_vect_perm(__pyx_v_v1, __pyx_v_p_p1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":504
 * 
 * 
 * def op_vect_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.op_vect_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":513
 * 
 * 
 * def op_gcode_matrix(uint32_t v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_103op_gcode_matrix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_103op_gcode_matrix = {"op_gcode_matrix", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_103op_gcode_matrix, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_103op_gcode_matrix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  PyObject *__pyx_v_m1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_gcode_matrix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_m1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("op_gcode_matrix", 1, 2, 2, 1); __PYX_ERR(0, 513, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "op_gcode_matrix") < 0)) __PYX_ERR(0, 513, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 513, __pyx_L3_error)
    __pyx_v_m1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("op_gcode_matrix", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 513, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.op_gcode_matrix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_102op_gcode_matrix(__pyx_self, __pyx_v_v1, __pyx_v_m1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_102op_gcode_matrix(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  uint32_t *__pyx_v_p_m1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_gcode_matrix", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":515
 * def op_gcode_matrix(uint32_t v1, m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":517
 *     cdef uint32_t *p_m1a = m1a
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     return mat24_op_gcode_matrix(v1,  p_m1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":518
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     return mat24_op_gcode_matrix(v1,  p_m1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_op_gcode_matrix(__pyx_v_v1, __pyx_v_p_m1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":513
 * 
 * 
 * def op_gcode_matrix(uint32_t v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.op_gcode_matrix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":521
 * 
 * 
 * def op_gcode_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_105op_gcode_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_105op_gcode_perm = {"op_gcode_perm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_105op_gcode_perm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_105op_gcode_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_v1;
  PyObject *__pyx_v_p1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_gcode_perm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_p1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("op_gcode_perm", 1, 2, 2, 1); __PYX_ERR(0, 521, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "op_gcode_perm") < 0)) __PYX_ERR(0, 521, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_v1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 521, __pyx_L3_error)
    __pyx_v_p1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("op_gcode_perm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 521, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.op_gcode_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_104op_gcode_perm(__pyx_self, __pyx_v_v1, __pyx_v_p1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_104op_gcode_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_v1, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_gcode_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":523
 * def op_gcode_perm(uint32_t v1, p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":525
 *     cdef uint8_t *p_p1a = p1a
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     return mat24_op_gcode_perm(v1, p_p1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":526
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     return mat24_op_gcode_perm(v1, p_p1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_op_gcode_perm(__pyx_v_v1, __pyx_v_p_p1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":521
 * 
 * 
 * def op_gcode_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.op_gcode_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":531
 * 
 * 
 * def op_cocode_perm(uint32_t c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_107op_cocode_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_107op_cocode_perm = {"op_cocode_perm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_107op_cocode_perm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_107op_cocode_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint32_t __pyx_v_c1;
  PyObject *__pyx_v_p1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_cocode_perm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_p1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("op_cocode_perm", 1, 2, 2, 1); __PYX_ERR(0, 531, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "op_cocode_perm") < 0)) __PYX_ERR(0, 531, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_c1 = __Pyx_PyInt_As_uint32_t(values[0]); if (unlikely((__pyx_v_c1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L3_error)
    __pyx_v_p1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("op_cocode_perm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 531, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.op_cocode_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_106op_cocode_perm(__pyx_self, __pyx_v_c1, __pyx_v_p1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_106op_cocode_perm(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_c1, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_cocode_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":533
 * def op_cocode_perm(uint32_t c1, p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":535
 *     cdef uint8_t *p_p1a = p1a
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     return mat24_op_cocode_perm(c1, p_p1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":536
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     return mat24_op_cocode_perm(c1, p_p1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_op_cocode_perm(__pyx_v_c1, __pyx_v_p_p1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":531
 * 
 * 
 * def op_cocode_perm(uint32_t c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.op_cocode_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":540
 * 
 * 
 * def mul_perm(p1, p2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_109mul_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_109mul_perm = {"mul_perm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_109mul_perm, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_109mul_perm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_p1 = 0;
  PyObject *__pyx_v_p2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul_perm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p1,&__pyx_n_s_p2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mul_perm", 1, 2, 2, 1); __PYX_ERR(0, 540, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mul_perm") < 0)) __PYX_ERR(0, 540, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_p1 = values[0];
    __pyx_v_p2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mul_perm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 540, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.mul_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_108mul_perm(__pyx_self, __pyx_v_p1, __pyx_v_p2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_108mul_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2) {
  uint8_t __pyx_v_pp1[24];
  uint8_t *__pyx_v_p_pp1;
  uint8_t __pyx_v_pp2[24];
  uint8_t *__pyx_v_p_pp2;
  uint8_t __pyx_v_pp3[24];
  uint8_t *__pyx_v_p_pp3;
  int __pyx_v_i;
  int __pyx_8genexpr8__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":542
 * def mul_perm(p1, p2):
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp2[24]
 *     cdef uint8_t *p_pp2 = pp2
 */
  __pyx_v_p_pp1 = __pyx_v_pp1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":544
 *     cdef uint8_t *p_pp1 = pp1
 *     cdef uint8_t pp2[24]
 *     cdef uint8_t *p_pp2 = pp2             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp3[24]
 *     cdef uint8_t *p_pp3 = pp3
 */
  __pyx_v_p_pp2 = __pyx_v_pp2;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":546
 *     cdef uint8_t *p_pp2 = pp2
 *     cdef uint8_t pp3[24]
 *     cdef uint8_t *p_pp3 = pp3             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24):
 */
  __pyx_v_p_pp3 = __pyx_v_pp3;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":548
 *     cdef uint8_t *p_pp3 = pp3
 *     cdef int i
 *     for i in range(24):             # <<<<<<<<<<<<<<
 *         pp1[i] = int(p1[i])
 *         pp2[i] = int(p2[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":549
 *     cdef int i
 *     for i in range(24):
 *         pp1[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *         pp2[i] = int(p2[i])
 *     mat24_mul_perm(p_pp1, p_pp2, p_pp3)
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_pp1[__pyx_v_i]) = __pyx_t_4;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":550
 *     for i in range(24):
 *         pp1[i] = int(p1[i])
 *         pp2[i] = int(p2[i])             # <<<<<<<<<<<<<<
 *     mat24_mul_perm(p_pp1, p_pp2, p_pp3)
 *     return [pp3[i] for i in range(24)]
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_p2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_pp2[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":551
 *         pp1[i] = int(p1[i])
 *         pp2[i] = int(p2[i])
 *     mat24_mul_perm(p_pp1, p_pp2, p_pp3)             # <<<<<<<<<<<<<<
 *     return [pp3[i] for i in range(24)]
 * 
 */
  mat24_mul_perm(__pyx_v_p_pp1, __pyx_v_p_pp2, __pyx_v_p_pp3);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":552
 *         pp2[i] = int(p2[i])
 *     mat24_mul_perm(p_pp1, p_pp2, p_pp3)
 *     return [pp3[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_8genexpr8__pyx_v_i = __pyx_t_1;
      __pyx_t_3 = __Pyx_PyInt_From_uint8_t((__pyx_v_pp3[__pyx_8genexpr8__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":540
 * 
 * 
 * def mul_perm(p1, p2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.mul_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":555
 * 
 * 
 * def inv_perm(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_111inv_perm(PyObject *__pyx_self, PyObject *__pyx_v_p1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_111inv_perm = {"inv_perm", (PyCFunction)__pyx_pw_7mmgroup_5mat24_111inv_perm, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_111inv_perm(PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inv_perm (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_110inv_perm(__pyx_self, ((PyObject *)__pyx_v_p1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_110inv_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_pp1[24];
  uint8_t *__pyx_v_p_pp1;
  uint8_t __pyx_v_pp2[24];
  uint8_t *__pyx_v_p_pp2;
  int __pyx_v_i;
  int __pyx_8genexpr9__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("inv_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":557
 * def inv_perm(p1):
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp2[24]
 *     cdef uint8_t *p_pp2 = pp2
 */
  __pyx_v_p_pp1 = __pyx_v_pp1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":559
 *     cdef uint8_t *p_pp1 = pp1
 *     cdef uint8_t pp2[24]
 *     cdef uint8_t *p_pp2 = pp2             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24):
 */
  __pyx_v_p_pp2 = __pyx_v_pp2;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":561
 *     cdef uint8_t *p_pp2 = pp2
 *     cdef int i
 *     for i in range(24):             # <<<<<<<<<<<<<<
 *         pp1[i] = int(p1[i])
 *     mat24_inv_perm(p_pp1, p_pp2)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":562
 *     cdef int i
 *     for i in range(24):
 *         pp1[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     mat24_inv_perm(p_pp1, p_pp2)
 *     return [pp2[i] for i in range(24)]
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_pp1[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":563
 *     for i in range(24):
 *         pp1[i] = int(p1[i])
 *     mat24_inv_perm(p_pp1, p_pp2)             # <<<<<<<<<<<<<<
 *     return [pp2[i] for i in range(24)]
 * 
 */
  mat24_inv_perm(__pyx_v_p_pp1, __pyx_v_p_pp2);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":564
 *         pp1[i] = int(p1[i])
 *     mat24_inv_perm(p_pp1, p_pp2)
 *     return [pp2[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_8genexpr9__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint8_t((__pyx_v_pp2[__pyx_8genexpr9__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 564, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":555
 * 
 * 
 * def inv_perm(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.inv_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":573
 * 
 * 
 * def autpl_set_qform(m_io):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m_ioa[12]
 *     cdef uint32_t *p_m_ioa = m_ioa
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_113autpl_set_qform(PyObject *__pyx_self, PyObject *__pyx_v_m_io); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_113autpl_set_qform = {"autpl_set_qform", (PyCFunction)__pyx_pw_7mmgroup_5mat24_113autpl_set_qform, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_113autpl_set_qform(PyObject *__pyx_self, PyObject *__pyx_v_m_io) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("autpl_set_qform (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_112autpl_set_qform(__pyx_self, ((PyObject *)__pyx_v_m_io));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_112autpl_set_qform(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m_io) {
  uint32_t __pyx_v_m_ioa[12];
  uint32_t *__pyx_v_p_m_ioa;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("autpl_set_qform", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":575
 * def autpl_set_qform(m_io):
 *     cdef uint32_t m_ioa[12]
 *     cdef uint32_t *p_m_ioa = m_ioa             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m_ioa[i] = int(m_io[i])
 */
  __pyx_v_p_m_ioa = __pyx_v_m_ioa;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":577
 *     cdef uint32_t *p_m_ioa = m_ioa
 *     cdef int i
 *     for i in range(12): m_ioa[i] = int(m_io[i])             # <<<<<<<<<<<<<<
 *     mat24_autpl_set_qform(p_m_ioa)
 *     for i in range(12): m_io[i] = int(m_ioa[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m_io, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m_ioa[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":578
 *     cdef int i
 *     for i in range(12): m_ioa[i] = int(m_io[i])
 *     mat24_autpl_set_qform(p_m_ioa)             # <<<<<<<<<<<<<<
 *     for i in range(12): m_io[i] = int(m_ioa[i])
 *     return m_io
 */
  mat24_autpl_set_qform(__pyx_v_p_m_ioa);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":579
 *     for i in range(12): m_ioa[i] = int(m_io[i])
 *     mat24_autpl_set_qform(p_m_ioa)
 *     for i in range(12): m_io[i] = int(m_ioa[i])             # <<<<<<<<<<<<<<
 *     return m_io
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_3 = __Pyx_PyInt_From_uint32_t((__pyx_v_m_ioa[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyInt_Type)), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__Pyx_SetItemInt(__pyx_v_m_io, __pyx_v_i, __pyx_t_2, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":580
 *     mat24_autpl_set_qform(p_m_ioa)
 *     for i in range(12): m_io[i] = int(m_ioa[i])
 *     return m_io             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_m_io);
  __pyx_r = __pyx_v_m_io;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":573
 * 
 * 
 * def autpl_set_qform(m_io):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m_ioa[12]
 *     cdef uint32_t *p_m_ioa = m_ioa
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.autpl_set_qform", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":584
 * 
 * 
 * def perm_to_autpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_115perm_to_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_115perm_to_autpl = {"perm_to_autpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_115perm_to_autpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_115perm_to_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_c1 = 0;
  PyObject *__pyx_v_p1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_to_autpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_p1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perm_to_autpl", 1, 2, 2, 1); __PYX_ERR(0, 584, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perm_to_autpl") < 0)) __PYX_ERR(0, 584, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_c1 = values[0];
    __pyx_v_p1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perm_to_autpl", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 584, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_114perm_to_autpl(__pyx_self, __pyx_v_c1, __pyx_v_p1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_114perm_to_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  uint32_t __pyx_v_m1[12];
  uint32_t *__pyx_v_p_m1;
  int __pyx_v_i;
  int __pyx_9genexpr10__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  uint32_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_to_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":586
 * def perm_to_autpl(c1, p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":588
 *     cdef uint8_t *p_p1a = p1a
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_m1 = __pyx_v_m1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":590
 *     cdef uint32_t *p_m1 = m1
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     mat24_perm_to_autpl(int(c1), p_p1a, p_m1)
 *     return [m1[i] for i in range(12)]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":591
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_autpl(int(c1), p_p1a, p_m1)             # <<<<<<<<<<<<<<
 *     return [m1[i] for i in range(12)]
 * 
 */
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_c1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_5 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  mat24_perm_to_autpl(__pyx_t_5, __pyx_v_p_p1a, __pyx_v_p_m1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":592
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_autpl(int(c1), p_p1a, p_m1)
 *     return [m1[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
      __pyx_9genexpr10__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint32_t((__pyx_v_m1[__pyx_9genexpr10__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 592, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":584
 * 
 * 
 * def perm_to_autpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":595
 * 
 * 
 * def cocode_to_autpl(c1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_117cocode_to_autpl(PyObject *__pyx_self, PyObject *__pyx_v_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_117cocode_to_autpl = {"cocode_to_autpl", (PyCFunction)__pyx_pw_7mmgroup_5mat24_117cocode_to_autpl, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_117cocode_to_autpl(PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cocode_to_autpl (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_116cocode_to_autpl(__pyx_self, ((PyObject *)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_116cocode_to_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  uint32_t __pyx_v_m1[12];
  uint32_t *__pyx_v_p_m1;
  long __pyx_9genexpr11__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint32_t __pyx_t_2;
  long __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cocode_to_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":597
 * def cocode_to_autpl(c1):
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1             # <<<<<<<<<<<<<<
 *     mat24_cocode_to_autpl(int(c1), p_m1)
 *     return [m1[i] for i in range(12)]
 */
  __pyx_v_p_m1 = __pyx_v_m1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":598
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 *     mat24_cocode_to_autpl(int(c1), p_m1)             # <<<<<<<<<<<<<<
 *     return [m1[i] for i in range(12)]
 * 
 */
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_c1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  mat24_cocode_to_autpl(__pyx_t_2, __pyx_v_p_m1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":599
 *     cdef uint32_t *p_m1 = m1
 *     mat24_cocode_to_autpl(int(c1), p_m1)
 *     return [m1[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_3 = 0; __pyx_t_3 < 12; __pyx_t_3+=1) {
      __pyx_9genexpr11__pyx_v_i = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_From_uint32_t((__pyx_v_m1[__pyx_9genexpr11__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 599, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":595
 * 
 * 
 * def cocode_to_autpl(c1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mmgroup.mat24.cocode_to_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":603
 * 
 * 
 * def autpl_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_119autpl_to_perm(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_119autpl_to_perm = {"autpl_to_perm", (PyCFunction)__pyx_pw_7mmgroup_5mat24_119autpl_to_perm, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_119autpl_to_perm(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("autpl_to_perm (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_118autpl_to_perm(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_118autpl_to_perm(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  uint32_t *__pyx_v_p_m1a;
  uint8_t __pyx_v_p1[24];
  uint8_t *__pyx_v_p_p1;
  int __pyx_v_i;
  int __pyx_9genexpr12__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("autpl_to_perm", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":605
 * def autpl_to_perm(m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":607
 *     cdef uint32_t *p_m1a = m1a
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_p1 = __pyx_v_p1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":609
 *     cdef uint8_t *p_p1 = p1
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     mat24_autpl_to_perm(p_m1a, p_p1)
 *     return [p1[i] for i in range(24)]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":610
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_autpl_to_perm(p_m1a, p_p1)             # <<<<<<<<<<<<<<
 *     return [p1[i] for i in range(24)]
 * 
 */
  mat24_autpl_to_perm(__pyx_v_p_m1a, __pyx_v_p_p1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":611
 *     for i in range(12): m1a[i] = int(m1[i])
 *     mat24_autpl_to_perm(p_m1a, p_p1)
 *     return [p1[i] for i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_9genexpr12__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint8_t((__pyx_v_p1[__pyx_9genexpr12__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 611, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":603
 * 
 * 
 * def autpl_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.autpl_to_perm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":614
 * 
 * 
 * def autpl_to_cocode(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_121autpl_to_cocode(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_121autpl_to_cocode = {"autpl_to_cocode", (PyCFunction)__pyx_pw_7mmgroup_5mat24_121autpl_to_cocode, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_121autpl_to_cocode(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("autpl_to_cocode (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_120autpl_to_cocode(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_120autpl_to_cocode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  uint32_t *__pyx_v_p_m1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("autpl_to_cocode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":616
 * def autpl_to_cocode(m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":618
 *     cdef uint32_t *p_m1a = m1a
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     return mat24_autpl_to_cocode(p_m1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":619
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     return mat24_autpl_to_cocode(p_m1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_autpl_to_cocode(__pyx_v_p_m1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":614
 * 
 * 
 * def autpl_to_cocode(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.autpl_to_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":623
 * 
 * 
 * def op_ploop_autpl(v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_123op_ploop_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_123op_ploop_autpl = {"op_ploop_autpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_123op_ploop_autpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_123op_ploop_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_v1 = 0;
  PyObject *__pyx_v_m1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_ploop_autpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_v1,&__pyx_n_s_m1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("op_ploop_autpl", 1, 2, 2, 1); __PYX_ERR(0, 623, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "op_ploop_autpl") < 0)) __PYX_ERR(0, 623, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_v1 = values[0];
    __pyx_v_m1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("op_ploop_autpl", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 623, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.op_ploop_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_122op_ploop_autpl(__pyx_self, __pyx_v_v1, __pyx_v_m1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_122op_ploop_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v1, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  uint32_t *__pyx_v_p_m1a;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_ploop_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":625
 * def op_ploop_autpl(v1, m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":627
 *     cdef uint32_t *p_m1a = m1a
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     return mat24_op_ploop_autpl(v1, p_m1a)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":628
 *     cdef int i
 *     for i in range(12): m1a[i] = int(m1[i])
 *     return mat24_op_ploop_autpl(v1, p_m1a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_v_v1); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 628, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t(mat24_op_ploop_autpl(__pyx_t_4, __pyx_v_p_m1a)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":623
 * 
 * 
 * def op_ploop_autpl(v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.op_ploop_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":631
 * 
 * 
 * def mul_autpl(m1, m2):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_125mul_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_125mul_autpl = {"mul_autpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_125mul_autpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_125mul_autpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_m1 = 0;
  PyObject *__pyx_v_m2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mul_autpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_m1,&__pyx_n_s_m2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mul_autpl", 1, 2, 2, 1); __PYX_ERR(0, 631, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mul_autpl") < 0)) __PYX_ERR(0, 631, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_m1 = values[0];
    __pyx_v_m2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mul_autpl", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 631, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.mul_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_124mul_autpl(__pyx_self, __pyx_v_m1, __pyx_v_m2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_124mul_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1, PyObject *__pyx_v_m2) {
  uint32_t __pyx_v_m1a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m1a;
  uint32_t __pyx_v_m2a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m2a;
  uint32_t __pyx_v_m3a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m3a;
  int __pyx_v_i;
  int __pyx_9genexpr13__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mul_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":633
 * def mul_autpl(m1, m2):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m2a[12]
 *     cdef uint32_t *p_m2a = m2a
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":635
 *     cdef uint32_t *p_m1a = m1a
 *     cdef uint32_t m2a[12]
 *     cdef uint32_t *p_m2a = m2a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m3a[12]
 *     cdef uint32_t *p_m3a = m3a
 */
  __pyx_v_p_m2a = __pyx_v_m2a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":637
 *     cdef uint32_t *p_m2a = m2a
 *     cdef uint32_t m3a[12]
 *     cdef uint32_t *p_m3a = m3a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12):
 */
  __pyx_v_p_m3a = __pyx_v_m3a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":639
 *     cdef uint32_t *p_m3a = m3a
 *     cdef int i
 *     for i in range(12):             # <<<<<<<<<<<<<<
 *         m1a[i] = int(m1[i])
 *         m2a[i] = int(m2[i])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":640
 *     cdef int i
 *     for i in range(12):
 *         m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *         m2a[i] = int(m2[i])
 *     mat24_mul_autpl(m1a, m2a, m3a)
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":641
 *     for i in range(12):
 *         m1a[i] = int(m1[i])
 *         m2a[i] = int(m2[i])             # <<<<<<<<<<<<<<
 *     mat24_mul_autpl(m1a, m2a, m3a)
 *     return [m3a[i] for i in range(12)]
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_m2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_m2a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":642
 *         m1a[i] = int(m1[i])
 *         m2a[i] = int(m2[i])
 *     mat24_mul_autpl(m1a, m2a, m3a)             # <<<<<<<<<<<<<<
 *     return [m3a[i] for i in range(12)]
 * 
 */
  mat24_mul_autpl(__pyx_v_m1a, __pyx_v_m2a, __pyx_v_m3a);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":643
 *         m2a[i] = int(m2[i])
 *     mat24_mul_autpl(m1a, m2a, m3a)
 *     return [m3a[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
      __pyx_9genexpr13__pyx_v_i = __pyx_t_1;
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t((__pyx_v_m3a[__pyx_9genexpr13__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":631
 * 
 * 
 * def mul_autpl(m1, m2):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.mul_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":646
 * 
 * 
 * def inv_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_127inv_autpl(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_127inv_autpl = {"inv_autpl", (PyCFunction)__pyx_pw_7mmgroup_5mat24_127inv_autpl, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_127inv_autpl(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inv_autpl (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_126inv_autpl(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_126inv_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_m1a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m1a;
  uint32_t __pyx_v_m2a[12];
  CYTHON_UNUSED uint32_t *__pyx_v_p_m2a;
  int __pyx_v_i;
  int __pyx_9genexpr14__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("inv_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":648
 * def inv_autpl(m1):
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m2a[12]
 *     cdef uint32_t *p_m2a = m2a
 */
  __pyx_v_p_m1a = __pyx_v_m1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":650
 *     cdef uint32_t *p_m1a = m1a
 *     cdef uint32_t m2a[12]
 *     cdef uint32_t *p_m2a = m2a             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12):
 */
  __pyx_v_p_m2a = __pyx_v_m2a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":652
 *     cdef uint32_t *p_m2a = m2a
 *     cdef int i
 *     for i in range(12):             # <<<<<<<<<<<<<<
 *         m1a[i] = int(m1[i])
 *     mat24_inv_autpl(m1a, m2a)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":653
 *     cdef int i
 *     for i in range(12):
 *         m1a[i] = int(m1[i])             # <<<<<<<<<<<<<<
 *     mat24_inv_autpl(m1a, m2a)
 *     return [m2a[i] for i in range(12)]
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_m1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":654
 *     for i in range(12):
 *         m1a[i] = int(m1[i])
 *     mat24_inv_autpl(m1a, m2a)             # <<<<<<<<<<<<<<
 *     return [m2a[i] for i in range(12)]
 * 
 */
  mat24_inv_autpl(__pyx_v_m1a, __pyx_v_m2a);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":655
 *         m1a[i] = int(m1[i])
 *     mat24_inv_autpl(m1a, m2a)
 *     return [m2a[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
      __pyx_9genexpr14__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint32_t((__pyx_v_m2a[__pyx_9genexpr14__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":646
 * 
 * 
 * def inv_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("mmgroup.mat24.inv_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":659
 * 
 * 
 * def perm_to_iautpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_129perm_to_iautpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_129perm_to_iautpl = {"perm_to_iautpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7mmgroup_5mat24_129perm_to_iautpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_129perm_to_iautpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_c1 = 0;
  PyObject *__pyx_v_p1 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_to_iautpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c1,&__pyx_n_s_p1,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perm_to_iautpl", 1, 2, 2, 1); __PYX_ERR(0, 659, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perm_to_iautpl") < 0)) __PYX_ERR(0, 659, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_c1 = values[0];
    __pyx_v_p1 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perm_to_iautpl", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 659, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_iautpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_128perm_to_iautpl(__pyx_self, __pyx_v_c1, __pyx_v_p1);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_128perm_to_iautpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p1a[24];
  uint8_t *__pyx_v_p_p1a;
  uint32_t __pyx_v_m1[12];
  uint32_t *__pyx_v_p_m1;
  uint8_t __pyx_v_pi[24];
  uint8_t *__pyx_v_p_pi;
  int __pyx_v_i;
  int __pyx_9genexpr15__pyx_v_i;
  int __pyx_9genexpr16__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  uint8_t __pyx_t_4;
  uint32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_to_iautpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":661
 * def perm_to_iautpl(c1, p1):
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */
  __pyx_v_p_p1a = __pyx_v_p1a;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":663
 *     cdef uint8_t *p_p1a = p1a
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1             # <<<<<<<<<<<<<<
 *     cdef uint8_t pi[24]
 *     cdef uint8_t *p_pi = pi
 */
  __pyx_v_p_m1 = __pyx_v_m1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":665
 *     cdef uint32_t *p_m1 = m1
 *     cdef uint8_t pi[24]
 *     cdef uint8_t *p_pi = pi             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 */
  __pyx_v_p_pi = __pyx_v_pi;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":667
 *     cdef uint8_t *p_pi = pi
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])             # <<<<<<<<<<<<<<
 *     mat24_perm_to_iautpl(int(c1), p_p1a, p_pi, p_m1)
 *     return [pi[i] for i in range(24)], [m1[i] for i in range(12)]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_3); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_p1a[__pyx_v_i]) = __pyx_t_4;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":668
 *     cdef int i
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_iautpl(int(c1), p_p1a, p_pi, p_m1)             # <<<<<<<<<<<<<<
 *     return [pi[i] for i in range(24)], [m1[i] for i in range(12)]
 * 
 */
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_c1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_uint32_t(__pyx_t_3); if (unlikely((__pyx_t_5 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  mat24_perm_to_iautpl(__pyx_t_5, __pyx_v_p_p1a, __pyx_v_p_pi, __pyx_v_p_m1);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":669
 *     for i in range(24): p1a[i] = int(p1[i])
 *     mat24_perm_to_iautpl(int(c1), p_p1a, p_pi, p_m1)
 *     return [pi[i] for i in range(24)], [m1[i] for i in range(12)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
      __pyx_9genexpr15__pyx_v_i = __pyx_t_1;
      __pyx_t_2 = __Pyx_PyInt_From_uint8_t((__pyx_v_pi[__pyx_9genexpr15__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
      __pyx_9genexpr16__pyx_v_i = __pyx_t_1;
      __pyx_t_6 = __Pyx_PyInt_From_uint32_t((__pyx_v_m1[__pyx_9genexpr16__pyx_v_i])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 669, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":659
 * 
 * 
 * def perm_to_iautpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_iautpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":679
 * 
 * 
 * def perm_to_net(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p[24]
 *     cdef uint8_t *p_p = p
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_131perm_to_net(PyObject *__pyx_self, PyObject *__pyx_v_p1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_131perm_to_net = {"perm_to_net", (PyCFunction)__pyx_pw_7mmgroup_5mat24_131perm_to_net, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_131perm_to_net(PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perm_to_net (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_130perm_to_net(__pyx_self, ((PyObject *)__pyx_v_p1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_130perm_to_net(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p1) {
  uint8_t __pyx_v_p[24];
  uint8_t *__pyx_v_p_p;
  uint32_t __pyx_v_res[9];
  uint32_t *__pyx_v_p_res;
  int __pyx_v_i;
  int __pyx_9genexpr17__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  uint8_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perm_to_net", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":681
 * def perm_to_net(p1):
 *     cdef uint8_t p[24]
 *     cdef uint8_t *p_p = p             # <<<<<<<<<<<<<<
 *     cdef uint32_t res[9]
 *     cdef uint32_t *p_res =res
 */
  __pyx_v_p_p = __pyx_v_p;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":683
 *     cdef uint8_t *p_p = p
 *     cdef uint32_t res[9]
 *     cdef uint32_t *p_res =res             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(24):
 */
  __pyx_v_p_res = __pyx_v_res;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":685
 *     cdef uint32_t *p_res =res
 *     cdef int i
 *     for i in range(24):             # <<<<<<<<<<<<<<
 *         p[i] = p1[i]
 *     mat24_perm_to_net(p_p, p_res)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 24; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":686
 *     cdef int i
 *     for i in range(24):
 *         p[i] = p1[i]             # <<<<<<<<<<<<<<
 *     mat24_perm_to_net(p_p, p_res)
 *     return [res[i] for i in range(9)]
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_p1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_p[__pyx_v_i]) = __pyx_t_3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":687
 *     for i in range(24):
 *         p[i] = p1[i]
 *     mat24_perm_to_net(p_p, p_res)             # <<<<<<<<<<<<<<
 *     return [res[i] for i in range(9)]
 * 
 */
  mat24_perm_to_net(__pyx_v_p_p, __pyx_v_p_res);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":688
 *         p[i] = p1[i]
 *     mat24_perm_to_net(p_p, p_res)
 *     return [res[i] for i in range(9)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 9; __pyx_t_1+=1) {
      __pyx_9genexpr17__pyx_v_i = __pyx_t_1;
      __pyx_t_4 = __Pyx_PyInt_From_uint32_t((__pyx_v_res[__pyx_9genexpr17__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":679
 * 
 * 
 * def perm_to_net(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p[24]
 *     cdef uint8_t *p_p = p
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mmgroup.mat24.perm_to_net", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":694
 * 
 * 
 * def op_all_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pm[12]
 *     cdef uint32_t *p_pm = pm
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_133op_all_autpl(PyObject *__pyx_self, PyObject *__pyx_v_m1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_133op_all_autpl = {"op_all_autpl", (PyCFunction)__pyx_pw_7mmgroup_5mat24_133op_all_autpl, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_133op_all_autpl(PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_all_autpl (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_132op_all_autpl(__pyx_self, ((PyObject *)__pyx_v_m1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_132op_all_autpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_m1) {
  uint32_t __pyx_v_pm[12];
  uint32_t *__pyx_v_p_pm;
  uint16_t __pyx_v_res[0x800];
  uint16_t *__pyx_v_p_res;
  int __pyx_v_i;
  int __pyx_9genexpr18__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  uint32_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_all_autpl", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":696
 * def op_all_autpl(m1):
 *     cdef uint32_t pm[12]
 *     cdef uint32_t *p_pm = pm             # <<<<<<<<<<<<<<
 *     cdef uint16_t res[2048]
 *     cdef uint16_t *p_res = res
 */
  __pyx_v_p_pm = __pyx_v_pm;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":698
 *     cdef uint32_t *p_pm = pm
 *     cdef uint16_t res[2048]
 *     cdef uint16_t *p_res = res             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i in range(12):
 */
  __pyx_v_p_res = __pyx_v_res;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":700
 *     cdef uint16_t *p_res = res
 *     cdef int i
 *     for i in range(12):             # <<<<<<<<<<<<<<
 *         pm[i] = m1[i]
 *     mat24_op_all_autpl(p_pm, p_res)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 12; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":701
 *     cdef int i
 *     for i in range(12):
 *         pm[i] = m1[i]             # <<<<<<<<<<<<<<
 *     mat24_op_all_autpl(p_pm, p_res)
 *     return [res[i] for i in range(2048)]
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_m1, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_pm[__pyx_v_i]) = __pyx_t_3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":702
 *     for i in range(12):
 *         pm[i] = m1[i]
 *     mat24_op_all_autpl(p_pm, p_res)             # <<<<<<<<<<<<<<
 *     return [res[i] for i in range(2048)]
 * 
 */
  mat24_op_all_autpl(__pyx_v_p_pm, __pyx_v_p_res);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":703
 *         pm[i] = m1[i]
 *     mat24_op_all_autpl(p_pm, p_res)
 *     return [res[i] for i in range(2048)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_1 = 0; __pyx_t_1 < 0x800; __pyx_t_1+=1) {
      __pyx_9genexpr18__pyx_v_i = __pyx_t_1;
      __pyx_t_4 = __Pyx_PyInt_From_uint16_t((__pyx_v_res[__pyx_9genexpr18__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":694
 * 
 * 
 * def op_all_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pm[12]
 *     cdef uint32_t *p_pm = pm
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mmgroup.mat24.op_all_autpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":707
 * 
 * 
 * def op_all_cocode(c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_135op_all_cocode(PyObject *__pyx_self, PyObject *__pyx_v_c1); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_135op_all_cocode = {"op_all_cocode", (PyCFunction)__pyx_pw_7mmgroup_5mat24_135op_all_cocode, METH_O, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_135op_all_cocode(PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("op_all_cocode (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_134op_all_cocode(__pyx_self, ((PyObject *)__pyx_v_c1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_134op_all_cocode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c1) {
  uint8_t __pyx_v_res[0x800];
  uint8_t *__pyx_v_p_res;
  uint32_t __pyx_v_c1i;
  int __pyx_9genexpr19__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("op_all_cocode", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":709
 * def op_all_cocode(c1):
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res             # <<<<<<<<<<<<<<
 *     cdef uint32_t c1i = c1
 *     cdef int i
 */
  __pyx_v_p_res = __pyx_v_res;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":710
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res
 *     cdef uint32_t c1i = c1             # <<<<<<<<<<<<<<
 *     cdef int i
 *     mat24_op_all_cocode(c1i, p_res)
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_c1); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L1_error)
  __pyx_v_c1i = __pyx_t_1;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":712
 *     cdef uint32_t c1i = c1
 *     cdef int i
 *     mat24_op_all_cocode(c1i, p_res)             # <<<<<<<<<<<<<<
 *     return [res[i] for i in range(2048)]
 * 
 */
  mat24_op_all_cocode(__pyx_v_c1i, __pyx_v_p_res);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":713
 *     cdef int i
 *     mat24_op_all_cocode(c1i, p_res)
 *     return [res[i] for i in range(2048)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    for (__pyx_t_3 = 0; __pyx_t_3 < 0x800; __pyx_t_3+=1) {
      __pyx_9genexpr19__pyx_v_i = __pyx_t_3;
      __pyx_t_4 = __Pyx_PyInt_From_uint8_t((__pyx_v_res[__pyx_9genexpr19__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 713, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 713, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":707
 * 
 * 
 * def op_all_cocode(c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("mmgroup.mat24.op_all_cocode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":725
 * 
 * 
 * cpdef uint32_t Mat24Sub_vtype(uint32_t x):             # <<<<<<<<<<<<<<
 *     """returns type of bit vector in GF(2)**24.
 * 
 */

static PyObject *__pyx_pw_7mmgroup_5mat24_137Mat24Sub_vtype(PyObject *__pyx_self, PyObject *__pyx_arg_x); /*proto*/
static uint32_t __pyx_f_7mmgroup_5mat24_Mat24Sub_vtype(uint32_t __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  uint32_t __pyx_v_w;
  uint32_t __pyx_v_s;
  uint32_t __pyx_v_d;
  uint32_t __pyx_v_y;
  uint32_t __pyx_v_rest;
  uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  long __pyx_t_3;
  __Pyx_RefNannySetupContext("Mat24Sub_vtype", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":745
 *     """
 *     cdef uint32_t w, s, d, y,  rest
 *     w = mat24_bw24(x)             # <<<<<<<<<<<<<<
 *     s = mat24_syndrome(x,0)
 *     d = mat24_bw24(s)
 */
  __pyx_v_w = mat24_bw24(__pyx_v_x);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":746
 *     cdef uint32_t w, s, d, y,  rest
 *     w = mat24_bw24(x)
 *     s = mat24_syndrome(x,0)             # <<<<<<<<<<<<<<
 *     d = mat24_bw24(s)
 *     if d < 4:
 */
  __pyx_v_s = mat24_syndrome(__pyx_v_x, 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":747
 *     w = mat24_bw24(x)
 *     s = mat24_syndrome(x,0)
 *     d = mat24_bw24(s)             # <<<<<<<<<<<<<<
 *     if d < 4:
 *         y =  mat24_bw24(x & s)
 */
  __pyx_v_d = mat24_bw24(__pyx_v_s);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":748
 *     s = mat24_syndrome(x,0)
 *     d = mat24_bw24(s)
 *     if d < 4:             # <<<<<<<<<<<<<<
 *         y =  mat24_bw24(x & s)
 *         if w >= 12: y = d - y
 */
  __pyx_t_1 = ((__pyx_v_d < 4) != 0);
  if (__pyx_t_1) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":749
 *     d = mat24_bw24(s)
 *     if d < 4:
 *         y =  mat24_bw24(x & s)             # <<<<<<<<<<<<<<
 *         if w >= 12: y = d - y
 *     elif w != 12:
 */
    __pyx_v_y = mat24_bw24((__pyx_v_x & __pyx_v_s));

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":750
 *     if d < 4:
 *         y =  mat24_bw24(x & s)
 *         if w >= 12: y = d - y             # <<<<<<<<<<<<<<
 *     elif w != 12:
 *         y = 0
 */
    __pyx_t_1 = ((__pyx_v_w >= 12) != 0);
    if (__pyx_t_1) {
      __pyx_v_y = (__pyx_v_d - __pyx_v_y);
    }

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":748
 *     s = mat24_syndrome(x,0)
 *     d = mat24_bw24(s)
 *     if d < 4:             # <<<<<<<<<<<<<<
 *         y =  mat24_bw24(x & s)
 *         if w >= 12: y = d - y
 */
    goto __pyx_L3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":751
 *         y =  mat24_bw24(x & s)
 *         if w >= 12: y = d - y
 *     elif w != 12:             # <<<<<<<<<<<<<<
 *         y = 0
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_w != 12) != 0);
  if (__pyx_t_1) {

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":752
 *         if w >= 12: y = d - y
 *     elif w != 12:
 *         y = 0             # <<<<<<<<<<<<<<
 *     else:
 *         y, rest = 0 + (s == x & s),  x & ~s & 0xffffff
 */
    __pyx_v_y = 0;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":751
 *         y =  mat24_bw24(x & s)
 *         if w >= 12: y = d - y
 *     elif w != 12:             # <<<<<<<<<<<<<<
 *         y = 0
 *     else:
 */
    goto __pyx_L3;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":754
 *         y = 0
 *     else:
 *         y, rest = 0 + (s == x & s),  x & ~s & 0xffffff             # <<<<<<<<<<<<<<
 *         while rest:
 *             s =  mat24_syndrome(x, mat24_lsbit24(rest))
 */
  /*else*/ {
    __pyx_t_2 = (0 + (__pyx_v_s == (__pyx_v_x & __pyx_v_s)));
    __pyx_t_3 = ((__pyx_v_x & (~__pyx_v_s)) & 0xffffff);
    __pyx_v_y = __pyx_t_2;
    __pyx_v_rest = __pyx_t_3;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":755
 *     else:
 *         y, rest = 0 + (s == x & s),  x & ~s & 0xffffff
 *         while rest:             # <<<<<<<<<<<<<<
 *             s =  mat24_syndrome(x, mat24_lsbit24(rest))
 *             y +=  s == x & s
 */
    while (1) {
      __pyx_t_1 = (__pyx_v_rest != 0);
      if (!__pyx_t_1) break;

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":756
 *         y, rest = 0 + (s == x & s),  x & ~s & 0xffffff
 *         while rest:
 *             s =  mat24_syndrome(x, mat24_lsbit24(rest))             # <<<<<<<<<<<<<<
 *             y +=  s == x & s
 *             rest &= ~s
 */
      __pyx_v_s = mat24_syndrome(__pyx_v_x, mat24_lsbit24(__pyx_v_rest));

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":757
 *         while rest:
 *             s =  mat24_syndrome(x, mat24_lsbit24(rest))
 *             y +=  s == x & s             # <<<<<<<<<<<<<<
 *             rest &= ~s
 *     return ( (w << 5 ) + (d << 2) + y )
 */
      __pyx_v_y = (__pyx_v_y + (__pyx_v_s == (__pyx_v_x & __pyx_v_s)));

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":758
 *             s =  mat24_syndrome(x, mat24_lsbit24(rest))
 *             y +=  s == x & s
 *             rest &= ~s             # <<<<<<<<<<<<<<
 *     return ( (w << 5 ) + (d << 2) + y )
 * 
 */
      __pyx_v_rest = (__pyx_v_rest & (~__pyx_v_s));
    }
  }
  __pyx_L3:;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":759
 *             y +=  s == x & s
 *             rest &= ~s
 *     return ( (w << 5 ) + (d << 2) + y )             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (((__pyx_v_w << 5) + (__pyx_v_d << 2)) + __pyx_v_y);
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":725
 * 
 * 
 * cpdef uint32_t Mat24Sub_vtype(uint32_t x):             # <<<<<<<<<<<<<<
 *     """returns type of bit vector in GF(2)**24.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_137Mat24Sub_vtype(PyObject *__pyx_self, PyObject *__pyx_arg_x); /*proto*/
static char __pyx_doc_7mmgroup_5mat24_136Mat24Sub_vtype[] = "returns type of bit vector in GF(2)**24. \n\n    This is either (w,d,y), 0 <= d <= 3,  or (w,4,t)\n\n    Here w is the bit weight of vector v. In the first case, there\n    is a unique syndrome s(x) of x of weight d, 0 <= d <= 3, and\n    the weight of x & s(x) is equal to y in case w <= 12 and to\n    d - y in case w >= 12.\n    In the second case the syndrome may be any of a tetrad of \n    six vectors s_i of length 4 with disjoint support. In case \n    w = 12, the value  t is the number of vectors s_i completely \n    contained in x. For other values of w the value t is zero.\n  \n    It is well known that all vectors of the same type are\n    in one orbit of the automorphism group of the Golay code.\n\n    The tuple is returned in the form  (w << 5 ) + (d << 2) + y .\n    ";
static PyObject *__pyx_pw_7mmgroup_5mat24_137Mat24Sub_vtype(PyObject *__pyx_self, PyObject *__pyx_arg_x) {
  uint32_t __pyx_v_x;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Mat24Sub_vtype (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("mmgroup.mat24.Mat24Sub_vtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7mmgroup_5mat24_136Mat24Sub_vtype(__pyx_self, ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_136Mat24Sub_vtype(CYTHON_UNUSED PyObject *__pyx_self, uint32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Mat24Sub_vtype", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(__pyx_f_7mmgroup_5mat24_Mat24Sub_vtype(__pyx_v_x, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("mmgroup.mat24.Mat24Sub_vtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "src/mmgroup/dev/pxd_files/mat24fast.pyx":763
 * 
 * 
 * def Mat24Sub_count_vtypes():             # <<<<<<<<<<<<<<
 *      cdef uint32_t a[8000]
 *      cdef uint32_t n
 */

/* Python wrapper */
static PyObject *__pyx_pw_7mmgroup_5mat24_139Mat24Sub_count_vtypes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_7mmgroup_5mat24_139Mat24Sub_count_vtypes = {"Mat24Sub_count_vtypes", (PyCFunction)__pyx_pw_7mmgroup_5mat24_139Mat24Sub_count_vtypes, METH_NOARGS, 0};
static PyObject *__pyx_pw_7mmgroup_5mat24_139Mat24Sub_count_vtypes(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Mat24Sub_count_vtypes (wrapper)", 0);
  __pyx_r = __pyx_pf_7mmgroup_5mat24_138Mat24Sub_count_vtypes(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7mmgroup_5mat24_138Mat24Sub_count_vtypes(CYTHON_UNUSED PyObject *__pyx_self) {
  uint32_t __pyx_v_a[0x1F40];
  uint32_t __pyx_v_n;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Mat24Sub_count_vtypes", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":766
 *      cdef uint32_t a[8000]
 *      cdef uint32_t n
 *      for n in range(8000):             # <<<<<<<<<<<<<<
 *          a[n] = 0
 *      for n in range(0x1000000):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x1F40; __pyx_t_1+=1) {
    __pyx_v_n = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":767
 *      cdef uint32_t n
 *      for n in range(8000):
 *          a[n] = 0             # <<<<<<<<<<<<<<
 *      for n in range(0x1000000):
 *          a[Mat24Sub_vtype(n)] += 1
 */
    (__pyx_v_a[__pyx_v_n]) = 0;
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":768
 *      for n in range(8000):
 *          a[n] = 0
 *      for n in range(0x1000000):             # <<<<<<<<<<<<<<
 *          a[Mat24Sub_vtype(n)] += 1
 *      d =  {}
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x1000000; __pyx_t_1+=1) {
    __pyx_v_n = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":769
 *          a[n] = 0
 *      for n in range(0x1000000):
 *          a[Mat24Sub_vtype(n)] += 1             # <<<<<<<<<<<<<<
 *      d =  {}
 *      for n in range(8000):
 */
    __pyx_t_2 = __pyx_f_7mmgroup_5mat24_Mat24Sub_vtype(__pyx_v_n, 0);
    (__pyx_v_a[__pyx_t_2]) = ((__pyx_v_a[__pyx_t_2]) + 1);
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":770
 *      for n in range(0x1000000):
 *          a[Mat24Sub_vtype(n)] += 1
 *      d =  {}             # <<<<<<<<<<<<<<
 *      for n in range(8000):
 *         if a[n] > 0:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_d = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":771
 *          a[Mat24Sub_vtype(n)] += 1
 *      d =  {}
 *      for n in range(8000):             # <<<<<<<<<<<<<<
 *         if a[n] > 0:
 *             d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x1F40; __pyx_t_1+=1) {
    __pyx_v_n = __pyx_t_1;

    /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":772
 *      d =  {}
 *      for n in range(8000):
 *         if a[n] > 0:             # <<<<<<<<<<<<<<
 *             d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]
 *      return d
 */
    __pyx_t_4 = (((__pyx_v_a[__pyx_v_n]) > 0) != 0);
    if (__pyx_t_4) {

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":773
 *      for n in range(8000):
 *         if a[n] > 0:
 *             d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]             # <<<<<<<<<<<<<<
 *      return d
 * 
 */
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t((__pyx_v_a[__pyx_v_n])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_From_long((__pyx_v_n >> 5)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_From_long(((__pyx_v_n >> 2) & 7)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyInt_From_long((__pyx_v_n & 3)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_d, __pyx_t_8, __pyx_t_3) < 0)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":772
 *      d =  {}
 *      for n in range(8000):
 *         if a[n] > 0:             # <<<<<<<<<<<<<<
 *             d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]
 *      return d
 */
    }
  }

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":774
 *         if a[n] > 0:
 *             d[ (n>>5, (n>>2) & 7, n & 3) ] = a[n]
 *      return d             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_d);
  __pyx_r = __pyx_v_d;
  goto __pyx_L0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":763
 * 
 * 
 * def Mat24Sub_count_vtypes():             # <<<<<<<<<<<<<<
 *      cdef uint32_t a[8000]
 *      cdef uint32_t n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("mmgroup.mat24.Mat24Sub_count_vtypes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"Mat24Sub_vtype", (PyCFunction)__pyx_pw_7mmgroup_5mat24_137Mat24Sub_vtype, METH_O, __pyx_doc_7mmgroup_5mat24_136Mat24Sub_vtype},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_mat24(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_mat24},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "mat24",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_Array_in_function_perm_from_map, __pyx_k_Array_in_function_perm_from_map, sizeof(__pyx_k_Array_in_function_perm_from_map), 0, 1, 0, 0},
  {&__pyx_kp_u_Arrays_in_function_perm_from_map, __pyx_k_Arrays_in_function_perm_from_map, sizeof(__pyx_k_Arrays_in_function_perm_from_map), 0, 1, 0, 0},
  {&__pyx_kp_u_Attempt_to_compute_suboctad_of_a, __pyx_k_Attempt_to_compute_suboctad_of_a, sizeof(__pyx_k_Attempt_to_compute_suboctad_of_a), 0, 1, 0, 0},
  {&__pyx_kp_u_Bit_vector_is_not_a_Golay_code_w, __pyx_k_Bit_vector_is_not_a_Golay_code_w, sizeof(__pyx_k_Bit_vector_is_not_a_Golay_code_w), 0, 1, 0, 0},
  {&__pyx_kp_u_Bit_vector_is_not_an_octad, __pyx_k_Bit_vector_is_not_an_octad, sizeof(__pyx_k_Bit_vector_is_not_an_octad), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_complete_a_vector_to_an_o, __pyx_k_Cannot_complete_a_vector_to_an_o, sizeof(__pyx_k_Cannot_complete_a_vector_to_an_o), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_construct_permutation_in, __pyx_k_Cannot_construct_permutation_in, sizeof(__pyx_k_Cannot_construct_permutation_in), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_construct_permutation_in_2, __pyx_k_Cannot_construct_permutation_in_2, sizeof(__pyx_k_Cannot_construct_permutation_in_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Cannot_correct_signs_of_all_Park, __pyx_k_Cannot_correct_signs_of_all_Park, sizeof(__pyx_k_Cannot_correct_signs_of_all_Park), 0, 1, 0, 0},
  {&__pyx_kp_u_Cocode_element_is_not_a_subset_o, __pyx_k_Cocode_element_is_not_a_subset_o, sizeof(__pyx_k_Cocode_element_is_not_a_subset_o), 0, 1, 0, 0},
  {&__pyx_kp_u_Golay_cocode_word_is_not_a_sexte, __pyx_k_Golay_cocode_word_is_not_a_sexte, sizeof(__pyx_k_Golay_cocode_word_is_not_a_sexte), 0, 1, 0, 0},
  {&__pyx_kp_u_Golay_code_syndrome_is_not_uniqu, __pyx_k_Golay_code_syndrome_is_not_uniqu, sizeof(__pyx_k_Golay_code_syndrome_is_not_uniqu), 0, 1, 0, 0},
  {&__pyx_kp_u_Golay_code_word_is_not_an_octad, __pyx_k_Golay_code_word_is_not_an_octad, sizeof(__pyx_k_Golay_code_word_is_not_an_octad), 0, 1, 0, 0},
  {&__pyx_kp_u_Illegal_entry_in_array_in_functi, __pyx_k_Illegal_entry_in_array_in_functi, sizeof(__pyx_k_Illegal_entry_in_array_in_functi), 0, 1, 0, 0},
  {&__pyx_kp_u_Illegal_number_for_a_permutation, __pyx_k_Illegal_number_for_a_permutation, sizeof(__pyx_k_Illegal_number_for_a_permutation), 0, 1, 0, 0},
  {&__pyx_kp_u_Illegal_octad_number, __pyx_k_Illegal_octad_number, sizeof(__pyx_k_Illegal_octad_number), 0, 1, 0, 0},
  {&__pyx_n_s_MAT24_ORDER, __pyx_k_MAT24_ORDER, sizeof(__pyx_k_MAT24_ORDER), 0, 0, 1, 1},
  {&__pyx_n_s_Mat24Sub_count_vtypes, __pyx_k_Mat24Sub_count_vtypes, sizeof(__pyx_k_Mat24Sub_count_vtypes), 0, 0, 1, 1},
  {&__pyx_kp_u_Octad_suboctad_mismatch, __pyx_k_Octad_suboctad_mismatch, sizeof(__pyx_k_Octad_suboctad_mismatch), 0, 1, 0, 0},
  {&__pyx_kp_u_Permutation_in_group_Mat24_must, __pyx_k_Permutation_in_group_Mat24_must, sizeof(__pyx_k_Permutation_in_group_Mat24_must), 0, 1, 0, 0},
  {&__pyx_kp_u_Permutation_is_not_in_group_Mat2, __pyx_k_Permutation_is_not_in_group_Mat2, sizeof(__pyx_k_Permutation_is_not_in_group_Mat2), 0, 1, 0, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u_Vector_is_not_an_umbral_heptad, __pyx_k_Vector_is_not_an_umbral_heptad, sizeof(__pyx_k_Vector_is_not_an_umbral_heptad), 0, 1, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_acc, __pyx_k_acc, sizeof(__pyx_k_acc), 0, 0, 1, 1},
  {&__pyx_n_s_autpl_set_qform, __pyx_k_autpl_set_qform, sizeof(__pyx_k_autpl_set_qform), 0, 0, 1, 1},
  {&__pyx_n_s_autpl_to_cocode, __pyx_k_autpl_to_cocode, sizeof(__pyx_k_autpl_to_cocode), 0, 0, 1, 1},
  {&__pyx_n_s_autpl_to_perm, __pyx_k_autpl_to_perm, sizeof(__pyx_k_autpl_to_perm), 0, 0, 1, 1},
  {&__pyx_n_s_basis, __pyx_k_basis, sizeof(__pyx_k_basis), 0, 0, 1, 1},
  {&__pyx_n_s_bw24, __pyx_k_bw24, sizeof(__pyx_k_bw24), 0, 0, 1, 1},
  {&__pyx_n_s_c1, __pyx_k_c1, sizeof(__pyx_k_c1), 0, 0, 1, 1},
  {&__pyx_n_s_c1i, __pyx_k_c1i, sizeof(__pyx_k_c1i), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cn, __pyx_k_cn, sizeof(__pyx_k_cn), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_as_subdodecad, __pyx_k_cocode_as_subdodecad, sizeof(__pyx_k_cocode_as_subdodecad), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_syndrome, __pyx_k_cocode_syndrome, sizeof(__pyx_k_cocode_syndrome), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_to_autpl, __pyx_k_cocode_to_autpl, sizeof(__pyx_k_cocode_to_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_to_bit_list, __pyx_k_cocode_to_bit_list, sizeof(__pyx_k_cocode_to_bit_list), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_to_sextet, __pyx_k_cocode_to_sextet, sizeof(__pyx_k_cocode_to_sextet), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_to_suboctad, __pyx_k_cocode_to_suboctad, sizeof(__pyx_k_cocode_to_suboctad), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_to_vect, __pyx_k_cocode_to_vect, sizeof(__pyx_k_cocode_to_vect), 0, 0, 1, 1},
  {&__pyx_n_s_cocode_weight, __pyx_k_cocode_weight, sizeof(__pyx_k_cocode_weight), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_d1, __pyx_k_d1, sizeof(__pyx_k_d1), 0, 0, 1, 1},
  {&__pyx_n_s_d1a, __pyx_k_d1a, sizeof(__pyx_k_d1a), 0, 0, 1, 1},
  {&__pyx_n_s_d2, __pyx_k_d2, sizeof(__pyx_k_d2), 0, 0, 1, 1},
  {&__pyx_n_s_d2a, __pyx_k_d2a, sizeof(__pyx_k_d2a), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_entry, __pyx_k_entry, sizeof(__pyx_k_entry), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_extract_b24, __pyx_k_extract_b24, sizeof(__pyx_k_extract_b24), 0, 0, 1, 1},
  {&__pyx_n_s_gcode_to_bit_list, __pyx_k_gcode_to_bit_list, sizeof(__pyx_k_gcode_to_bit_list), 0, 0, 1, 1},
  {&__pyx_n_s_gcode_to_octad, __pyx_k_gcode_to_octad, sizeof(__pyx_k_gcode_to_octad), 0, 0, 1, 1},
  {&__pyx_n_s_gcode_to_vect, __pyx_k_gcode_to_vect, sizeof(__pyx_k_gcode_to_vect), 0, 0, 1, 1},
  {&__pyx_n_s_gcode_weight, __pyx_k_gcode_weight, sizeof(__pyx_k_gcode_weight), 0, 0, 1, 1},
  {&__pyx_n_s_h1, __pyx_k_h1, sizeof(__pyx_k_h1), 0, 0, 1, 1},
  {&__pyx_n_s_h1a, __pyx_k_h1a, sizeof(__pyx_k_h1a), 0, 0, 1, 1},
  {&__pyx_n_s_h2, __pyx_k_h2, sizeof(__pyx_k_h2), 0, 0, 1, 1},
  {&__pyx_n_s_h2a, __pyx_k_h2a, sizeof(__pyx_k_h2a), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_inv_autpl, __pyx_k_inv_autpl, sizeof(__pyx_k_inv_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_inv_perm, __pyx_k_inv_perm, sizeof(__pyx_k_inv_perm), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_lsbit24, __pyx_k_lsbit24, sizeof(__pyx_k_lsbit24), 0, 0, 1, 1},
  {&__pyx_n_s_m1, __pyx_k_m1, sizeof(__pyx_k_m1), 0, 0, 1, 1},
  {&__pyx_n_s_m1a, __pyx_k_m1a, sizeof(__pyx_k_m1a), 0, 0, 1, 1},
  {&__pyx_n_s_m2, __pyx_k_m2, sizeof(__pyx_k_m2), 0, 0, 1, 1},
  {&__pyx_n_s_m24num_to_perm, __pyx_k_m24num_to_perm, sizeof(__pyx_k_m24num_to_perm), 0, 0, 1, 1},
  {&__pyx_n_s_m2a, __pyx_k_m2a, sizeof(__pyx_k_m2a), 0, 0, 1, 1},
  {&__pyx_n_s_m3a, __pyx_k_m3a, sizeof(__pyx_k_m3a), 0, 0, 1, 1},
  {&__pyx_n_s_m_io, __pyx_k_m_io, sizeof(__pyx_k_m_io), 0, 0, 1, 1},
  {&__pyx_n_s_m_ioa, __pyx_k_m_ioa, sizeof(__pyx_k_m_ioa), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_matrix_from_mod_omega, __pyx_k_matrix_from_mod_omega, sizeof(__pyx_k_matrix_from_mod_omega), 0, 0, 1, 1},
  {&__pyx_n_s_matrix_to_perm, __pyx_k_matrix_to_perm, sizeof(__pyx_k_matrix_to_perm), 0, 0, 1, 1},
  {&__pyx_n_s_mmgroup_mat24, __pyx_k_mmgroup_mat24, sizeof(__pyx_k_mmgroup_mat24), 0, 0, 1, 1},
  {&__pyx_n_s_mul_autpl, __pyx_k_mul_autpl, sizeof(__pyx_k_mul_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_mul_perm, __pyx_k_mul_perm, sizeof(__pyx_k_mul_perm), 0, 0, 1, 1},
  {&__pyx_n_s_mul_ploop, __pyx_k_mul_ploop, sizeof(__pyx_k_mul_ploop), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_octad_to_gcode, __pyx_k_octad_to_gcode, sizeof(__pyx_k_octad_to_gcode), 0, 0, 1, 1},
  {&__pyx_n_s_octad_to_vect, __pyx_k_octad_to_vect, sizeof(__pyx_k_octad_to_vect), 0, 0, 1, 1},
  {&__pyx_n_s_op_all_autpl, __pyx_k_op_all_autpl, sizeof(__pyx_k_op_all_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_op_all_cocode, __pyx_k_op_all_cocode, sizeof(__pyx_k_op_all_cocode), 0, 0, 1, 1},
  {&__pyx_n_s_op_cocode_perm, __pyx_k_op_cocode_perm, sizeof(__pyx_k_op_cocode_perm), 0, 0, 1, 1},
  {&__pyx_n_s_op_gcode_matrix, __pyx_k_op_gcode_matrix, sizeof(__pyx_k_op_gcode_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_op_gcode_perm, __pyx_k_op_gcode_perm, sizeof(__pyx_k_op_gcode_perm), 0, 0, 1, 1},
  {&__pyx_n_s_op_ploop_autpl, __pyx_k_op_ploop_autpl, sizeof(__pyx_k_op_ploop_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_op_vect_perm, __pyx_k_op_vect_perm, sizeof(__pyx_k_op_vect_perm), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_p1, __pyx_k_p1, sizeof(__pyx_k_p1), 0, 0, 1, 1},
  {&__pyx_n_s_p1a, __pyx_k_p1a, sizeof(__pyx_k_p1a), 0, 0, 1, 1},
  {&__pyx_n_s_p2, __pyx_k_p2, sizeof(__pyx_k_p2), 0, 0, 1, 1},
  {&__pyx_n_s_p_d1a, __pyx_k_p_d1a, sizeof(__pyx_k_p_d1a), 0, 0, 1, 1},
  {&__pyx_n_s_p_d2a, __pyx_k_p_d2a, sizeof(__pyx_k_p_d2a), 0, 0, 1, 1},
  {&__pyx_n_s_p_h1a, __pyx_k_p_h1a, sizeof(__pyx_k_p_h1a), 0, 0, 1, 1},
  {&__pyx_n_s_p_h2a, __pyx_k_p_h2a, sizeof(__pyx_k_p_h2a), 0, 0, 1, 1},
  {&__pyx_n_s_p_io, __pyx_k_p_io, sizeof(__pyx_k_p_io), 0, 0, 1, 1},
  {&__pyx_n_s_p_m1, __pyx_k_p_m1, sizeof(__pyx_k_p_m1), 0, 0, 1, 1},
  {&__pyx_n_s_p_m1a, __pyx_k_p_m1a, sizeof(__pyx_k_p_m1a), 0, 0, 1, 1},
  {&__pyx_n_s_p_m2a, __pyx_k_p_m2a, sizeof(__pyx_k_p_m2a), 0, 0, 1, 1},
  {&__pyx_n_s_p_m3a, __pyx_k_p_m3a, sizeof(__pyx_k_p_m3a), 0, 0, 1, 1},
  {&__pyx_n_s_p_m_ioa, __pyx_k_p_m_ioa, sizeof(__pyx_k_p_m_ioa), 0, 0, 1, 1},
  {&__pyx_n_s_p_p, __pyx_k_p_p, sizeof(__pyx_k_p_p), 0, 0, 1, 1},
  {&__pyx_n_s_p_p1, __pyx_k_p_p1, sizeof(__pyx_k_p_p1), 0, 0, 1, 1},
  {&__pyx_n_s_p_p1a, __pyx_k_p_p1a, sizeof(__pyx_k_p_p1a), 0, 0, 1, 1},
  {&__pyx_n_s_p_pi, __pyx_k_p_pi, sizeof(__pyx_k_p_pi), 0, 0, 1, 1},
  {&__pyx_n_s_p_pm, __pyx_k_p_pm, sizeof(__pyx_k_p_pm), 0, 0, 1, 1},
  {&__pyx_n_s_p_pp1, __pyx_k_p_pp1, sizeof(__pyx_k_p_pp1), 0, 0, 1, 1},
  {&__pyx_n_s_p_pp2, __pyx_k_p_pp2, sizeof(__pyx_k_p_pp2), 0, 0, 1, 1},
  {&__pyx_n_s_p_pp3, __pyx_k_p_pp3, sizeof(__pyx_k_p_pp3), 0, 0, 1, 1},
  {&__pyx_n_s_p_res, __pyx_k_p_res, sizeof(__pyx_k_p_res), 0, 0, 1, 1},
  {&__pyx_n_s_pa, __pyx_k_pa, sizeof(__pyx_k_pa), 0, 0, 1, 1},
  {&__pyx_n_s_perm_check, __pyx_k_perm_check, sizeof(__pyx_k_perm_check), 0, 0, 1, 1},
  {&__pyx_n_s_perm_complete_heptad, __pyx_k_perm_complete_heptad, sizeof(__pyx_k_perm_complete_heptad), 0, 0, 1, 1},
  {&__pyx_n_s_perm_complete_octad, __pyx_k_perm_complete_octad, sizeof(__pyx_k_perm_complete_octad), 0, 0, 1, 1},
  {&__pyx_n_s_perm_from_dodecads, __pyx_k_perm_from_dodecads, sizeof(__pyx_k_perm_from_dodecads), 0, 0, 1, 1},
  {&__pyx_n_s_perm_from_heptads, __pyx_k_perm_from_heptads, sizeof(__pyx_k_perm_from_heptads), 0, 0, 1, 1},
  {&__pyx_n_s_perm_from_map, __pyx_k_perm_from_map, sizeof(__pyx_k_perm_from_map), 0, 0, 1, 1},
  {&__pyx_n_s_perm_to_autpl, __pyx_k_perm_to_autpl, sizeof(__pyx_k_perm_to_autpl), 0, 0, 1, 1},
  {&__pyx_n_s_perm_to_iautpl, __pyx_k_perm_to_iautpl, sizeof(__pyx_k_perm_to_iautpl), 0, 0, 1, 1},
  {&__pyx_n_s_perm_to_m24num, __pyx_k_perm_to_m24num, sizeof(__pyx_k_perm_to_m24num), 0, 0, 1, 1},
  {&__pyx_n_s_perm_to_matrix, __pyx_k_perm_to_matrix, sizeof(__pyx_k_perm_to_matrix), 0, 0, 1, 1},
  {&__pyx_n_s_perm_to_net, __pyx_k_perm_to_net, sizeof(__pyx_k_perm_to_net), 0, 0, 1, 1},
  {&__pyx_n_s_pi, __pyx_k_pi, sizeof(__pyx_k_pi), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_assoc, __pyx_k_ploop_assoc, sizeof(__pyx_k_ploop_assoc), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_cap, __pyx_k_ploop_cap, sizeof(__pyx_k_ploop_cap), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_cocycle, __pyx_k_ploop_cocycle, sizeof(__pyx_k_ploop_cocycle), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_comm, __pyx_k_ploop_comm, sizeof(__pyx_k_ploop_comm), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_solve, __pyx_k_ploop_solve, sizeof(__pyx_k_ploop_solve), 0, 0, 1, 1},
  {&__pyx_n_s_ploop_theta, __pyx_k_ploop_theta, sizeof(__pyx_k_ploop_theta), 0, 0, 1, 1},
  {&__pyx_n_s_pm, __pyx_k_pm, sizeof(__pyx_k_pm), 0, 0, 1, 1},
  {&__pyx_n_s_pow_ploop, __pyx_k_pow_ploop, sizeof(__pyx_k_pow_ploop), 0, 0, 1, 1},
  {&__pyx_n_s_pp1, __pyx_k_pp1, sizeof(__pyx_k_pp1), 0, 0, 1, 1},
  {&__pyx_n_s_pp2, __pyx_k_pp2, sizeof(__pyx_k_pp2), 0, 0, 1, 1},
  {&__pyx_n_s_pp3, __pyx_k_pp3, sizeof(__pyx_k_pp3), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_recip_basis, __pyx_k_recip_basis, sizeof(__pyx_k_recip_basis), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_result_length, __pyx_k_result_length, sizeof(__pyx_k_result_length), 0, 0, 1, 1},
  {&__pyx_n_s_scalar_prod, __pyx_k_scalar_prod, sizeof(__pyx_k_scalar_prod), 0, 0, 1, 1},
  {&__pyx_n_s_spread_b24, __pyx_k_spread_b24, sizeof(__pyx_k_spread_b24), 0, 0, 1, 1},
  {&__pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_k_src_mmgroup_dev_pxd_files_mat24f, sizeof(__pyx_k_src_mmgroup_dev_pxd_files_mat24f), 0, 0, 1, 0},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_suboctad_scalar_prod, __pyx_k_suboctad_scalar_prod, sizeof(__pyx_k_suboctad_scalar_prod), 0, 0, 1, 1},
  {&__pyx_n_s_suboctad_to_cocode, __pyx_k_suboctad_to_cocode, sizeof(__pyx_k_suboctad_to_cocode), 0, 0, 1, 1},
  {&__pyx_n_s_suboctad_weight, __pyx_k_suboctad_weight, sizeof(__pyx_k_suboctad_weight), 0, 0, 1, 1},
  {&__pyx_n_s_syndrome, __pyx_k_syndrome, sizeof(__pyx_k_syndrome), 0, 0, 1, 1},
  {&__pyx_n_s_syndrome_table, __pyx_k_syndrome_table, sizeof(__pyx_k_syndrome_table), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_u_exp, __pyx_k_u_exp, sizeof(__pyx_k_u_exp), 0, 0, 1, 1},
  {&__pyx_n_s_u_len, __pyx_k_u_len, sizeof(__pyx_k_u_len), 0, 0, 1, 1},
  {&__pyx_n_s_u_m24, __pyx_k_u_m24, sizeof(__pyx_k_u_m24), 0, 0, 1, 1},
  {&__pyx_n_s_u_mask, __pyx_k_u_mask, sizeof(__pyx_k_u_mask), 0, 0, 1, 1},
  {&__pyx_n_s_u_octad, __pyx_k_u_octad, sizeof(__pyx_k_u_octad), 0, 0, 1, 1},
  {&__pyx_n_s_u_single, __pyx_k_u_single, sizeof(__pyx_k_u_single), 0, 0, 1, 1},
  {&__pyx_n_s_u_strict, __pyx_k_u_strict, sizeof(__pyx_k_u_strict), 0, 0, 1, 1},
  {&__pyx_n_s_u_sub, __pyx_k_u_sub, sizeof(__pyx_k_u_sub), 0, 0, 1, 1},
  {&__pyx_n_s_u_sub1, __pyx_k_u_sub1, sizeof(__pyx_k_u_sub1), 0, 0, 1, 1},
  {&__pyx_n_s_u_sub2, __pyx_k_u_sub2, sizeof(__pyx_k_u_sub2), 0, 0, 1, 1},
  {&__pyx_n_s_u_tetrad, __pyx_k_u_tetrad, sizeof(__pyx_k_u_tetrad), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_v1, __pyx_k_v1, sizeof(__pyx_k_v1), 0, 0, 1, 1},
  {&__pyx_n_s_v2, __pyx_k_v2, sizeof(__pyx_k_v2), 0, 0, 1, 1},
  {&__pyx_n_s_v3, __pyx_k_v3, sizeof(__pyx_k_v3), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_bit_list, __pyx_k_vect_to_bit_list, sizeof(__pyx_k_vect_to_bit_list), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_cocode, __pyx_k_vect_to_cocode, sizeof(__pyx_k_vect_to_cocode), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_gcode, __pyx_k_vect_to_gcode, sizeof(__pyx_k_vect_to_gcode), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_list, __pyx_k_vect_to_list, sizeof(__pyx_k_vect_to_list), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_octad, __pyx_k_vect_to_octad, sizeof(__pyx_k_vect_to_octad), 0, 0, 1, 1},
  {&__pyx_n_s_vect_to_vintern, __pyx_k_vect_to_vintern, sizeof(__pyx_k_vect_to_vintern), 0, 0, 1, 1},
  {&__pyx_n_s_vect_type, __pyx_k_vect_type, sizeof(__pyx_k_vect_type), 0, 0, 1, 1},
  {&__pyx_n_s_vintern_to_vect, __pyx_k_vintern_to_vect, sizeof(__pyx_k_vintern_to_vect), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 43, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 110, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":166
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 *     if res == 0xffffffff:
 *         raise ValueError("Golay code syndrome is not unique")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Golay_code_syndrome_is_not_uniqu); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":217
 *     result = mat24_cocode_to_sextet(c1, &data[0])
 *     if result == 0xffffffff:
 *         raise ValueError("Golay cocode word is not a sextet")             # <<<<<<<<<<<<<<
 *     return list(data[:24])
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_Golay_cocode_word_is_not_a_sexte); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":239
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 *     if res & 0xfffff000:
 *         raise ValueError("Attempt to compute suboctad of a non-octad")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_Attempt_to_compute_suboctad_of_a); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":247
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 *     if res & 0xfffff000:
 *         raise ValueError("Octad/suboctad mismatch")             # <<<<<<<<<<<<<<
 *     return res
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_Octad_suboctad_mismatch); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":342
 *     cdef uint32_t res = mat24_perm_complete_heptad(p_p1)
 *     if res != 0:
 *         raise ValueError("Vector is not an umbral heptad")             # <<<<<<<<<<<<<<
 *     for i in range(24): p_io[i]  = p1[i]
 *     return res
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_Vector_is_not_an_umbral_heptad); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":30
 * 
 * 
 * def lsbit24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_lsbit24(v1)
 * 
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_lsbit24, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":34
 * 
 * 
 * def bw24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_bw24(v1)
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_bw24, 34, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 34, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":53
 * 
 * 
 * def vect_to_bit_list(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */
  __pyx_tuple__10 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_v1, __pyx_n_s_a, __pyx_n_s_w); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_bit_list, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 53, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":61
 * 
 * 
 * def vect_to_list(uint32_t v1, uint32_t u_len):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */
  __pyx_tuple__12 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_u_len, __pyx_n_s_a, __pyx_n_s_w); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_list, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 61, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":69
 * 
 * 
 * def extract_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_extract_b24(v1, u_mask)
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_u_mask); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_extract_b24, 69, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 69, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":73
 * 
 * 
 * def spread_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_spread_b24(v1, u_mask)
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_u_mask); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_spread_b24, 73, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 73, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":84
 * 
 * 
 * def vect_to_vintern(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_vintern(v1)
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_vintern, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":89
 * 
 * 
 * def vintern_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return  mat24_vintern_to_vect(v1)
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vintern_to_vect, 89, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 89, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":93
 * 
 * 
 * def vect_to_cocode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_cocode(v1)
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_cocode, 93, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 93, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":97
 * 
 * 
 * def gcode_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_to_vect(v1)
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_gcode_to_vect, 97, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 97, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":101
 * 
 * 
 * def cocode_to_vect(uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_to_vect(c1)
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_n_s_c1, __pyx_n_s_c1); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_to_vect, 101, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 101, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":105
 * 
 * 
 * def vect_to_gcode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 */
  __pyx_tuple__28 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_v1, __pyx_n_s_cn, __pyx_n_s_err); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_gcode, 105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 105, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":114
 * 
 * 
 * def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 */
  __pyx_tuple__30 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_u_strict, __pyx_n_s_cn, __pyx_n_s_err); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_gcode_to_octad, 114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 114, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":123
 * 
 * 
 * def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 */
  __pyx_tuple__32 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_u_strict, __pyx_n_s_v, __pyx_n_s_err); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_to_octad, 123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 123, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":134
 * 
 * 
 * def octad_to_gcode(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 */
  __pyx_tuple__34 = PyTuple_Pack(4, __pyx_n_s_u_octad, __pyx_n_s_u_octad, __pyx_n_s_cn, __pyx_n_s_err); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_octad_to_gcode, 134, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 134, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":143
 * 
 * 
 * def octad_to_vect(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 */
  __pyx_tuple__36 = PyTuple_Pack(4, __pyx_n_s_u_octad, __pyx_n_s_u_octad, __pyx_n_s_cn, __pyx_n_s_err); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_octad_to_vect, 143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 143, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":159
 * 
 * 
 * def syndrome_table(c1):             # <<<<<<<<<<<<<<
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_n_s_c1); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_syndrome_table, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 159, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":162
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 */
  __pyx_tuple__40 = PyTuple_Pack(3, __pyx_n_s_c1, __pyx_n_s_u_tetrad, __pyx_n_s_res); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_syndrome, 162, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 162, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":170
 * 
 * 
 * def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 */
  __pyx_tuple__42 = PyTuple_Pack(3, __pyx_n_s_v1, __pyx_n_s_u_tetrad, __pyx_n_s_res); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_syndrome, 170, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 170, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":179
 * 
 * 
 * def vect_type(v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)
 */
  __pyx_tuple__44 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_res); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_vect_type, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":187
 * 
 * 
 * def gcode_weight(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_weight(v1)
 * 
 */
  __pyx_tuple__46 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_gcode_weight, 187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 187, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":191
 * 
 * 
 * def gcode_to_bit_list(uint32_t  v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t length
 */
  __pyx_tuple__48 = PyTuple_Pack(4, __pyx_n_s_v1, __pyx_n_s_v1, __pyx_n_s_data, __pyx_n_s_length); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_gcode_to_bit_list, 191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 191, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":199
 * 
 * 
 * def cocode_weight(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_weight(c1)
 * 
 */
  __pyx_tuple__50 = PyTuple_Pack(2, __pyx_n_s_c1, __pyx_n_s_c1); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_weight, 199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 199, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":203
 * 
 * 
 * def cocode_to_bit_list(uint32_t  c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[4]
 *     cdef uint32_t length
 */
  __pyx_tuple__52 = PyTuple_Pack(4, __pyx_n_s_c1, __pyx_n_s_u_tetrad, __pyx_n_s_data, __pyx_n_s_length); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_to_bit_list, 203, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 203, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":212
 * 
 * 
 * def cocode_to_sextet(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t result
 */
  __pyx_tuple__54 = PyTuple_Pack(4, __pyx_n_s_c1, __pyx_n_s_c1, __pyx_n_s_data, __pyx_n_s_result); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_to_sextet, 212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 212, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":227
 * 
 * 
 * def scalar_prod(uint32_t v1, uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_scalar_prod(v1, c1)
 * 
 */
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_c1); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_scalar_prod, 227, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 227, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":235
 * 
 * 
 * def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 */
  __pyx_tuple__58 = PyTuple_Pack(3, __pyx_n_s_u_sub, __pyx_n_s_v1, __pyx_n_s_res); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_suboctad_to_cocode, 235, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 235, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":243
 * 
 * 
 * def cocode_to_suboctad(uint32_t c1, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 */
  __pyx_tuple__60 = PyTuple_Pack(3, __pyx_n_s_c1, __pyx_n_s_v1, __pyx_n_s_res); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_to_suboctad, 243, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 243, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":251
 * 
 * 
 * def suboctad_weight(uint32_t u_sub):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_weight(u_sub)
 * 
 */
  __pyx_tuple__62 = PyTuple_Pack(2, __pyx_n_s_u_sub, __pyx_n_s_u_sub); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_suboctad_weight, 251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 251, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":255
 * 
 * 
 * def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_scalar_prod(u_sub1, u_sub2)
 * 
 */
  __pyx_tuple__64 = PyTuple_Pack(2, __pyx_n_s_u_sub1, __pyx_n_s_u_sub2); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_suboctad_scalar_prod, 255, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 255, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":262
 * ###########################################################################
 * 
 * def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 */
  __pyx_tuple__66 = PyTuple_Pack(5, __pyx_n_s_c1, __pyx_n_s_v1, __pyx_n_s_u_single, __pyx_n_s_res, __pyx_n_s_err); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_as_subdodecad, 262, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 262, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":276
 * 
 * 
 * def ploop_theta(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_theta(v1)
 * 
 */
  __pyx_tuple__68 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v1); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_theta, 276, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 276, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":280
 * 
 * 
 * def ploop_cocycle(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cocycle(v1, v2)
 * 
 */
  __pyx_tuple__70 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v2); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_cocycle, 280, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 280, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":285
 * 
 * 
 * def mul_ploop(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_mul_ploop(v1, v2)
 * 
 */
  __pyx_tuple__72 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v2); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_mul_ploop, 285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 285, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":289
 * 
 * 
 * def pow_ploop(uint32_t v1, uint32_t u_exp):             # <<<<<<<<<<<<<<
 *     return mat24_pow_ploop(v1, u_exp)
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_u_exp); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_pow_ploop, 289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 289, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":294
 * 
 * 
 * def ploop_comm(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_comm(v1, v2)
 * 
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v2); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_comm, 294, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 294, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":298
 * 
 * 
 * def ploop_cap(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cap(v1, v2)
 * 
 */
  __pyx_tuple__78 = PyTuple_Pack(2, __pyx_n_s_v1, __pyx_n_s_v2); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_cap, 298, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 298, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":302
 * 
 * 
 * def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_assoc(v1, v2, v3)
 * 
 */
  __pyx_tuple__80 = PyTuple_Pack(3, __pyx_n_s_v1, __pyx_n_s_v2, __pyx_n_s_v3); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_assoc, 302, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 302, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":309
 *      ploop_solve_size = 64
 * 
 * def ploop_solve(a):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pa[ploop_solve_size + 13]
 *     cdef unsigned int i
 */
  __pyx_tuple__82 = PyTuple_Pack(9, __pyx_n_s_a, __pyx_n_s_pa, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_length, __pyx_n_s_start, __pyx_n_s_d, __pyx_n_s_result_length, __pyx_n_s_err); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(1, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_ploop_solve, 309, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 309, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":335
 * 
 * 
 * def perm_complete_heptad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_tuple__84 = PyTuple_Pack(5, __pyx_n_s_p_io, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_res); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_complete_heptad, 335, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 335, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":348
 * 
 * 
 * def perm_check(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__86 = PyTuple_Pack(7, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_i, __pyx_n_s_acc, __pyx_n_s_entry, __pyx_n_s_err); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__86, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_check, 348, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 348, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":366
 * 
 * 
 * def perm_complete_octad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[8]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_tuple__88 = PyTuple_Pack(6, __pyx_n_s_p_io, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_err); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__88, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_complete_octad, 366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 366, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":380
 * 
 * 
 * def perm_from_heptads(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[7]
 *     cdef uint8_t *p_h1a = h1a
 */
  __pyx_tuple__90 = PyTuple_Pack(12, __pyx_n_s_h1, __pyx_n_s_h2, __pyx_n_s_h1a, __pyx_n_s_p_h1a, __pyx_n_s_h2a, __pyx_n_s_p_h2a, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_err, __pyx_n_s_i); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__90, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_from_heptads, 380, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 380, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":398
 * 
 * 
 * def perm_from_map(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[24]
 *     cdef uint8_t *p_h1a = h1a
 */
  __pyx_tuple__92 = PyTuple_Pack(13, __pyx_n_s_h1, __pyx_n_s_h2, __pyx_n_s_h1a, __pyx_n_s_p_h1a, __pyx_n_s_h2a, __pyx_n_s_p_h2a, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_length, __pyx_n_s_err, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_i); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(2, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_from_map, 398, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 398, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":423
 * 
 * 
 * def m24num_to_perm(uint32_t u_m24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_tuple__94 = PyTuple_Pack(7, __pyx_n_s_u_m24, __pyx_n_s_u_m24, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_err, __pyx_n_s_i); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_m24num_to_perm, 423, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 423, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":435
 * 
 * 
 * def perm_to_m24num(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__96 = PyTuple_Pack(4, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_to_m24num, 435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 435, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":443
 * 
 * 
 * def perm_to_matrix(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__98 = PyTuple_Pack(7, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_m1, __pyx_n_s_p_m1, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_to_matrix, 443, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 443, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":454
 * 
 * 
 * def matrix_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__100 = PyTuple_Pack(7, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__100, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_matrix_to_perm, 454, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 454, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":464
 *     return [p1[i] for i in range(24)]
 * 
 * def matrix_from_mod_omega(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__102 = PyTuple_Pack(4, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_matrix_from_mod_omega, 464, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 464, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":479
 * 
 * 
 * def perm_from_dodecads(d1, d2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t d1a[9]
 *     cdef uint8_t *p_d1a = d1a
 */
  __pyx_tuple__104 = PyTuple_Pack(12, __pyx_n_s_d1, __pyx_n_s_d2, __pyx_n_s_d1a, __pyx_n_s_p_d1a, __pyx_n_s_d2a, __pyx_n_s_p_d2a, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_err, __pyx_n_s_i); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_from_dodecads, 479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":504
 * 
 * 
 * def op_vect_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__106 = PyTuple_Pack(5, __pyx_n_s_v1, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_vect_perm, 504, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 504, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":513
 * 
 * 
 * def op_gcode_matrix(uint32_t v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__108 = PyTuple_Pack(5, __pyx_n_s_v1, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_gcode_matrix, 513, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 513, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":521
 * 
 * 
 * def op_gcode_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__110 = PyTuple_Pack(5, __pyx_n_s_v1, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_gcode_perm, 521, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 521, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":531
 * 
 * 
 * def op_cocode_perm(uint32_t c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__112 = PyTuple_Pack(5, __pyx_n_s_c1, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_cocode_perm, 531, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 531, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":540
 * 
 * 
 * def mul_perm(p1, p2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */
  __pyx_tuple__114 = PyTuple_Pack(10, __pyx_n_s_p1, __pyx_n_s_p2, __pyx_n_s_pp1, __pyx_n_s_p_pp1, __pyx_n_s_pp2, __pyx_n_s_p_pp2, __pyx_n_s_pp3, __pyx_n_s_p_pp3, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__114, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_mul_perm, 540, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 540, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":555
 * 
 * 
 * def inv_perm(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */
  __pyx_tuple__116 = PyTuple_Pack(7, __pyx_n_s_p1, __pyx_n_s_pp1, __pyx_n_s_p_pp1, __pyx_n_s_pp2, __pyx_n_s_p_pp2, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_inv_perm, 555, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(0, 555, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":573
 * 
 * 
 * def autpl_set_qform(m_io):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m_ioa[12]
 *     cdef uint32_t *p_m_ioa = m_ioa
 */
  __pyx_tuple__118 = PyTuple_Pack(4, __pyx_n_s_m_io, __pyx_n_s_m_ioa, __pyx_n_s_p_m_ioa, __pyx_n_s_i); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_autpl_set_qform, 573, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 573, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":584
 * 
 * 
 * def perm_to_autpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__120 = PyTuple_Pack(8, __pyx_n_s_c1, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_m1, __pyx_n_s_p_m1, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_to_autpl, 584, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 584, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":595
 * 
 * 
 * def cocode_to_autpl(c1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */
  __pyx_tuple__122 = PyTuple_Pack(4, __pyx_n_s_c1, __pyx_n_s_m1, __pyx_n_s_p_m1, __pyx_n_s_i); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_cocode_to_autpl, 595, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 595, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":603
 * 
 * 
 * def autpl_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__124 = PyTuple_Pack(7, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_p1, __pyx_n_s_p_p1, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_autpl_to_perm, 603, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(0, 603, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":614
 * 
 * 
 * def autpl_to_cocode(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__126 = PyTuple_Pack(4, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_autpl_to_cocode, 614, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(0, 614, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":623
 * 
 * 
 * def op_ploop_autpl(v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__128 = PyTuple_Pack(5, __pyx_n_s_v1, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_i); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_ploop_autpl, 623, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(0, 623, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":631
 * 
 * 
 * def mul_autpl(m1, m2):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__130 = PyTuple_Pack(10, __pyx_n_s_m1, __pyx_n_s_m2, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_m2a, __pyx_n_s_p_m2a, __pyx_n_s_m3a, __pyx_n_s_p_m3a, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_mul_autpl, 631, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(0, 631, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":646
 * 
 * 
 * def inv_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_tuple__132 = PyTuple_Pack(7, __pyx_n_s_m1, __pyx_n_s_m1a, __pyx_n_s_p_m1a, __pyx_n_s_m2a, __pyx_n_s_p_m2a, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_inv_autpl, 646, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":659
 * 
 * 
 * def perm_to_iautpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_tuple__134 = PyTuple_Pack(11, __pyx_n_s_c1, __pyx_n_s_p1, __pyx_n_s_p1a, __pyx_n_s_p_p1a, __pyx_n_s_m1, __pyx_n_s_p_m1, __pyx_n_s_pi, __pyx_n_s_p_pi, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_to_iautpl, 659, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(0, 659, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":679
 * 
 * 
 * def perm_to_net(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p[24]
 *     cdef uint8_t *p_p = p
 */
  __pyx_tuple__136 = PyTuple_Pack(7, __pyx_n_s_p1, __pyx_n_s_p, __pyx_n_s_p_p, __pyx_n_s_res, __pyx_n_s_p_res, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_perm_to_net, 679, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(0, 679, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":694
 * 
 * 
 * def op_all_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pm[12]
 *     cdef uint32_t *p_pm = pm
 */
  __pyx_tuple__138 = PyTuple_Pack(7, __pyx_n_s_m1, __pyx_n_s_pm, __pyx_n_s_p_pm, __pyx_n_s_res, __pyx_n_s_p_res, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_all_autpl, 694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(0, 694, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":707
 * 
 * 
 * def op_all_cocode(c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res
 */
  __pyx_tuple__140 = PyTuple_Pack(6, __pyx_n_s_c1, __pyx_n_s_res, __pyx_n_s_p_res, __pyx_n_s_c1i, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_op_all_cocode, 707, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(0, 707, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":763
 * 
 * 
 * def Mat24Sub_count_vtypes():             # <<<<<<<<<<<<<<
 *      cdef uint32_t a[8000]
 *      cdef uint32_t n
 */
  __pyx_tuple__142 = PyTuple_Pack(3, __pyx_n_s_a, __pyx_n_s_n, __pyx_n_s_d); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_mmgroup_dev_pxd_files_mat24f, __pyx_n_s_Mat24Sub_count_vtypes, 763, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_2047 = PyInt_FromLong(2047); if (unlikely(!__pyx_int_2047)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_244823040 = PyInt_FromLong(244823040L); if (unlikely(!__pyx_int_244823040)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4278190080 = PyInt_FromString((char *)"4278190080", 0, 0); if (unlikely(!__pyx_int_4278190080)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294963200 = PyInt_FromString((char *)"4294963200", 0, 0); if (unlikely(!__pyx_int_4294963200)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initmat24(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initmat24(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_mat24(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_mat24(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_mat24(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'mat24' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_mat24(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("mat24", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_mmgroup__mat24) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "mmgroup.mat24")) {
      if (unlikely(PyDict_SetItemString(modules, "mmgroup.mat24", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":22
 * 
 * 
 * MAT24_ORDER =  244823040             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAT24_ORDER, __pyx_int_244823040) < 0) __PYX_ERR(0, 22, __pyx_L1_error)

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":30
 * 
 * 
 * def lsbit24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_lsbit24(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_1lsbit24, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lsbit24, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":34
 * 
 * 
 * def bw24(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_bw24(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_3bw24, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bw24, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":43
 * 
 * cdef int _i
 * basis =  [MAT24_BASIS[_i] for _i in range(24)]             # <<<<<<<<<<<<<<
 * recip_basis =  [MAT24_RECIP_BASIS[_i] for _i in range(24)]
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 24; __pyx_t_2+=1) {
      __pyx_7genexpr__pyx_v_7mmgroup_5mat24__i = __pyx_t_2;
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t((MAT24_BASIS[__pyx_7genexpr__pyx_v_7mmgroup_5mat24__i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 43, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_basis, __pyx_t_1) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":44
 * cdef int _i
 * basis =  [MAT24_BASIS[_i] for _i in range(24)]
 * recip_basis =  [MAT24_RECIP_BASIS[_i] for _i in range(24)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 24; __pyx_t_2+=1) {
      __pyx_8genexpr1__pyx_v_7mmgroup_5mat24__i = __pyx_t_2;
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t((MAT24_RECIP_BASIS[__pyx_8genexpr1__pyx_v_7mmgroup_5mat24__i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 44, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_recip_basis, __pyx_t_1) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":53
 * 
 * 
 * def vect_to_bit_list(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_5vect_to_bit_list, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_bit_list, __pyx_t_1) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":61
 * 
 * 
 * def vect_to_list(uint32_t v1, uint32_t u_len):             # <<<<<<<<<<<<<<
 *     cdef uint8_t a[24]
 *     cdef uint32_t w
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_7vect_to_list, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_list, __pyx_t_1) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":69
 * 
 * 
 * def extract_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_extract_b24(v1, u_mask)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_9extract_b24, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_extract_b24, __pyx_t_1) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":73
 * 
 * 
 * def spread_b24(uint32_t v1, uint32_t u_mask):             # <<<<<<<<<<<<<<
 *     return mat24_spread_b24(v1, u_mask)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_11spread_b24, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_spread_b24, __pyx_t_1) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":84
 * 
 * 
 * def vect_to_vintern(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_vintern(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_13vect_to_vintern, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_vintern, __pyx_t_1) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":89
 * 
 * 
 * def vintern_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return  mat24_vintern_to_vect(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_15vintern_to_vect, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vintern_to_vect, __pyx_t_1) < 0) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":93
 * 
 * 
 * def vect_to_cocode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_vect_to_cocode(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_17vect_to_cocode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_cocode, __pyx_t_1) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":97
 * 
 * 
 * def gcode_to_vect(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_to_vect(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_19gcode_to_vect, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gcode_to_vect, __pyx_t_1) < 0) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":101
 * 
 * 
 * def cocode_to_vect(uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_to_vect(c1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_21cocode_to_vect, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_to_vect, __pyx_t_1) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":105
 * 
 * 
 * def vect_to_gcode(uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_vect_to_gcode(v1)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_23vect_to_gcode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_gcode, __pyx_t_1) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":114
 * 
 * 
 * def gcode_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_gcode_to_octad(v1, u_strict)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_25gcode_to_octad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gcode_to_octad, __pyx_t_1) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":123
 * 
 * 
 * def vect_to_octad(uint32_t v1, uint32_t u_strict = 1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t v
 *     v = mat24_vect_to_octad(v1, u_strict)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_27vect_to_octad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_to_octad, __pyx_t_1) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":134
 * 
 * 
 * def octad_to_gcode(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_gcode(u_octad)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_29octad_to_gcode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_octad_to_gcode, __pyx_t_1) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":143
 * 
 * 
 * def octad_to_vect(uint32_t u_octad):             # <<<<<<<<<<<<<<
 *     cdef uint32_t cn
 *     cn = mat24_octad_to_vect(u_octad)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_31octad_to_vect, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_octad_to_vect, __pyx_t_1) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":159
 * 
 * 
 * def syndrome_table(c1):             # <<<<<<<<<<<<<<
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_33syndrome_table, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_syndrome_table, __pyx_t_1) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":162
 *     return MAT24_SYNDROME_TABLE[c1 & 0x7ff]
 * 
 * def cocode_syndrome(uint32_t c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = mat24_cocode_syndrome(c1, u_tetrad)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_35cocode_syndrome, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_syndrome, __pyx_t_1) < 0) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":170
 * 
 * 
 * def syndrome(uint32_t v1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_syndrome(v1, u_tetrad)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_37syndrome, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_syndrome, __pyx_t_1) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":179
 * 
 * 
 * def vect_type(v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res = Mat24Sub_vtype(v1)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_39vect_type, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vect_type, __pyx_t_1) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":187
 * 
 * 
 * def gcode_weight(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_gcode_weight(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_41gcode_weight, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gcode_weight, __pyx_t_1) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":191
 * 
 * 
 * def gcode_to_bit_list(uint32_t  v1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t length
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_43gcode_to_bit_list, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gcode_to_bit_list, __pyx_t_1) < 0) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":199
 * 
 * 
 * def cocode_weight(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     return mat24_cocode_weight(c1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_45cocode_weight, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_weight, __pyx_t_1) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":203
 * 
 * 
 * def cocode_to_bit_list(uint32_t  c1,  uint32_t u_tetrad = 24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[4]
 *     cdef uint32_t length
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_47cocode_to_bit_list, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_to_bit_list, __pyx_t_1) < 0) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":212
 * 
 * 
 * def cocode_to_sextet(uint32_t  c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t data[24]
 *     cdef uint32_t result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_49cocode_to_sextet, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_to_sextet, __pyx_t_1) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":227
 * 
 * 
 * def scalar_prod(uint32_t v1, uint32_t c1):             # <<<<<<<<<<<<<<
 *     return mat24_scalar_prod(v1, c1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_51scalar_prod, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_scalar_prod, __pyx_t_1) < 0) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":235
 * 
 * 
 * def suboctad_to_cocode(uint32_t u_sub, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_suboctad_to_cocode(u_sub, v1)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_53suboctad_to_cocode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_suboctad_to_cocode, __pyx_t_1) < 0) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":243
 * 
 * 
 * def cocode_to_suboctad(uint32_t c1, uint32_t v1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_to_suboctad(c1, v1)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_55cocode_to_suboctad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_to_suboctad, __pyx_t_1) < 0) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":251
 * 
 * 
 * def suboctad_weight(uint32_t u_sub):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_weight(u_sub)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_57suboctad_weight, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_suboctad_weight, __pyx_t_1) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":255
 * 
 * 
 * def suboctad_scalar_prod(uint32_t u_sub1, uint32_t u_sub2):             # <<<<<<<<<<<<<<
 *     return mat24_suboctad_scalar_prod(u_sub1, u_sub2)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_59suboctad_scalar_prod, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_suboctad_scalar_prod, __pyx_t_1) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":262
 * ###########################################################################
 * 
 * def cocode_as_subdodecad(uint32_t c1, uint32_t v1, uint32_t u_single = 24):             # <<<<<<<<<<<<<<
 *     cdef uint32_t res
 *     res =  mat24_cocode_as_subdodecad(c1, v1, u_single)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_61cocode_as_subdodecad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_as_subdodecad, __pyx_t_1) < 0) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":276
 * 
 * 
 * def ploop_theta(uint32_t v1):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_theta(v1)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_63ploop_theta, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_theta, __pyx_t_1) < 0) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":280
 * 
 * 
 * def ploop_cocycle(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cocycle(v1, v2)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_65ploop_cocycle, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_cocycle, __pyx_t_1) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":285
 * 
 * 
 * def mul_ploop(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_mul_ploop(v1, v2)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_67mul_ploop, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mul_ploop, __pyx_t_1) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":289
 * 
 * 
 * def pow_ploop(uint32_t v1, uint32_t u_exp):             # <<<<<<<<<<<<<<
 *     return mat24_pow_ploop(v1, u_exp)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_69pow_ploop, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pow_ploop, __pyx_t_1) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":294
 * 
 * 
 * def ploop_comm(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_comm(v1, v2)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_71ploop_comm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_comm, __pyx_t_1) < 0) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":298
 * 
 * 
 * def ploop_cap(uint32_t v1, uint32_t v2):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_cap(v1, v2)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_73ploop_cap, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_cap, __pyx_t_1) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":302
 * 
 * 
 * def ploop_assoc(uint32_t v1, uint32_t v2, uint32_t v3):             # <<<<<<<<<<<<<<
 *     return mat24_ploop_assoc(v1, v2, v3)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_75ploop_assoc, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_assoc, __pyx_t_1) < 0) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":309
 *      ploop_solve_size = 64
 * 
 * def ploop_solve(a):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pa[ploop_solve_size + 13]
 *     cdef unsigned int i
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_77ploop_solve, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ploop_solve, __pyx_t_1) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":335
 * 
 * 
 * def perm_complete_heptad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_79perm_complete_heptad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_complete_heptad, __pyx_t_1) < 0) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":348
 * 
 * 
 * def perm_check(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_81perm_check, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_check, __pyx_t_1) < 0) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":366
 * 
 * 
 * def perm_complete_octad(p_io):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[8]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_83perm_complete_octad, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_complete_octad, __pyx_t_1) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":380
 * 
 * 
 * def perm_from_heptads(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[7]
 *     cdef uint8_t *p_h1a = h1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_85perm_from_heptads, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_from_heptads, __pyx_t_1) < 0) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":398
 * 
 * 
 * def perm_from_map(h1, h2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t h1a[24]
 *     cdef uint8_t *p_h1a = h1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_87perm_from_map, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_from_map, __pyx_t_1) < 0) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":423
 * 
 * 
 * def m24num_to_perm(uint32_t u_m24):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1[24]
 *     cdef uint8_t *p_p1 = p1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_89m24num_to_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_m24num_to_perm, __pyx_t_1) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":435
 * 
 * 
 * def perm_to_m24num(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_91perm_to_m24num, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_to_m24num, __pyx_t_1) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":443
 * 
 * 
 * def perm_to_matrix(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_93perm_to_matrix, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_to_matrix, __pyx_t_1) < 0) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":454
 * 
 * 
 * def matrix_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_95matrix_to_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_matrix_to_perm, __pyx_t_1) < 0) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":464
 *     return [p1[i] for i in range(24)]
 * 
 * def matrix_from_mod_omega(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_97matrix_from_mod_omega, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_matrix_from_mod_omega, __pyx_t_1) < 0) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":479
 * 
 * 
 * def perm_from_dodecads(d1, d2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t d1a[9]
 *     cdef uint8_t *p_d1a = d1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_99perm_from_dodecads, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_from_dodecads, __pyx_t_1) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":504
 * 
 * 
 * def op_vect_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_101op_vect_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_vect_perm, __pyx_t_1) < 0) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":513
 * 
 * 
 * def op_gcode_matrix(uint32_t v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_103op_gcode_matrix, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_gcode_matrix, __pyx_t_1) < 0) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":521
 * 
 * 
 * def op_gcode_perm(uint32_t v1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_105op_gcode_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_gcode_perm, __pyx_t_1) < 0) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":531
 * 
 * 
 * def op_cocode_perm(uint32_t c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_107op_cocode_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_cocode_perm, __pyx_t_1) < 0) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":540
 * 
 * 
 * def mul_perm(p1, p2):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_109mul_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mul_perm, __pyx_t_1) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":555
 * 
 * 
 * def inv_perm(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t pp1[24]
 *     cdef uint8_t *p_pp1 = pp1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_111inv_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inv_perm, __pyx_t_1) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":573
 * 
 * 
 * def autpl_set_qform(m_io):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m_ioa[12]
 *     cdef uint32_t *p_m_ioa = m_ioa
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_113autpl_set_qform, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_autpl_set_qform, __pyx_t_1) < 0) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":584
 * 
 * 
 * def perm_to_autpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_115perm_to_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_to_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":595
 * 
 * 
 * def cocode_to_autpl(c1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1[12]
 *     cdef uint32_t *p_m1 = m1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_117cocode_to_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cocode_to_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":603
 * 
 * 
 * def autpl_to_perm(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_119autpl_to_perm, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_autpl_to_perm, __pyx_t_1) < 0) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":614
 * 
 * 
 * def autpl_to_cocode(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_121autpl_to_cocode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_autpl_to_cocode, __pyx_t_1) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":623
 * 
 * 
 * def op_ploop_autpl(v1, m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_123op_ploop_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_ploop_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":631
 * 
 * 
 * def mul_autpl(m1, m2):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_125mul_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mul_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":646
 * 
 * 
 * def inv_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t m1a[12]
 *     cdef uint32_t *p_m1a = m1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_127inv_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inv_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":659
 * 
 * 
 * def perm_to_iautpl(c1, p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p1a[24]
 *     cdef uint8_t *p_p1a = p1a
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_129perm_to_iautpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_to_iautpl, __pyx_t_1) < 0) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":679
 * 
 * 
 * def perm_to_net(p1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t p[24]
 *     cdef uint8_t *p_p = p
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_131perm_to_net, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_perm_to_net, __pyx_t_1) < 0) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":694
 * 
 * 
 * def op_all_autpl(m1):             # <<<<<<<<<<<<<<
 *     cdef uint32_t pm[12]
 *     cdef uint32_t *p_pm = pm
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_133op_all_autpl, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_all_autpl, __pyx_t_1) < 0) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":707
 * 
 * 
 * def op_all_cocode(c1):             # <<<<<<<<<<<<<<
 *     cdef uint8_t res[2048]
 *     cdef uint8_t *p_res = res
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_135op_all_cocode, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_op_all_cocode, __pyx_t_1) < 0) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":763
 * 
 * 
 * def Mat24Sub_count_vtypes():             # <<<<<<<<<<<<<<
 *      cdef uint32_t a[8000]
 *      cdef uint32_t n
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7mmgroup_5mat24_139Mat24Sub_count_vtypes, NULL, __pyx_n_s_mmgroup_mat24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Mat24Sub_count_vtypes, __pyx_t_1) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "src/mmgroup/dev/pxd_files/mat24fast.pyx":1
 * # cython: language_level=3             # <<<<<<<<<<<<<<
 * 
 * from __future__ import absolute_import, division, print_function
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init mmgroup.mat24", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init mmgroup.mat24");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value) {
    const uint16_t neg_one = (uint16_t) ((uint16_t) 0 - (uint16_t) 1), const_zero = (uint16_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
    const uint8_t neg_one = (uint8_t) ((uint8_t) 0 - (uint8_t) 1), const_zero = (uint8_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
    const uint32_t neg_one = (uint32_t) ((uint32_t) 0 - (uint32_t) 1), const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
    const uint8_t neg_one = (uint8_t) ((uint8_t) 0 - (uint8_t) 1), const_zero = (uint8_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
